Package: bigrquery
Title: An Interface to Google's 'BigQuery' 'API'
Version: 1.6.2
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-4757-117X")),
    person("Jennifer", "Bryan", , "jenny@posit.co", role = "aut",
           comment = c(ORCID = "0000-0002-6983-2759")),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Easily talk to Google's 'BigQuery' database from R.
License: MIT + file LICENSE
URL: https://bigrquery.r-dbi.org, https://github.com/r-dbi/bigrquery
BugReports: https://github.com/r-dbi/bigrquery/issues
Depends: R (>= 4.1)
Imports: bit64, brio, cli, clock, curl, DBI, gargle (>= 1.5.0), httr,
        jsonlite, lifecycle, methods, nanoparquet (>= 0.3.1),
        prettyunits, rlang (>= 1.1.0), tibble
Suggests: bigrquerystorage (>= 1.2.2), blob, covr, dbplyr (>= 2.4.0),
        dplyr (>= 1.1.0), hms, readr, sodium, testthat (>= 3.1.5),
        withr, wk (>= 0.3.2)
LinkingTo: cli, cpp11, rapidjsonr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/testthat/parallel: TRUE
Config/testthat/start-first: bq-table, dplyr, dbi-connection
Encoding: UTF-8
Collate: 'bigrquery-package.R' 'bq-auth.R' 'bq-dataset.R'
        'bq-download.R' 'bq-field.R' 'bq-job.R' 'bq-param.R'
        'bq-parse.R' 'bq-perform.R' 'bq-project.R' 'bq-projects.R'
        'bq-query.R' 'bq-refs.R' 'bq-request.R' 'bq-table.R'
        'bq-test.R' 'camelCase.R' 'connections-page.R' 'cpp11.R'
        'dbi-driver.R' 'dbi-connection.R' 'dbi-result.R' 'dplyr.R'
        'gs-object.R' 'import-standalone-obj-type.R'
        'import-standalone-s3-register.R'
        'import-standalone-types-check.R' 'utils.R' 'zzz.R'
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-12 12:45:27 UTC; hadleywickham
Author: Hadley Wickham [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4757-117X>),
  Jennifer Bryan [aut] (ORCID: <https://orcid.org/0000-0002-6983-2759>),
  Posit Software, PBC [cph, fnd]
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2026-05-12 13:20:09 UTC
Built: R 4.4.3; x86_64-apple-darwin20; 2026-05-12 14:18:26 UTC; unix
Archs: bigrquery.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(as.character,bq_job)
S3method(as.character,gs_object)
S3method(as_bq_dataset,BigQueryConnection)
S3method(as_bq_dataset,bq_dataset)
S3method(as_bq_dataset,character)
S3method(as_bq_dataset,default)
S3method(as_bq_dataset,list)
S3method(as_bq_field,bq_field)
S3method(as_bq_field,list)
S3method(as_bq_fields,"NULL")
S3method(as_bq_fields,bq_fields)
S3method(as_bq_fields,data.frame)
S3method(as_bq_fields,list)
S3method(as_bq_job,bq_job)
S3method(as_bq_job,character)
S3method(as_bq_job,default)
S3method(as_bq_job,list)
S3method(as_bq_table,BigQueryConnection)
S3method(as_bq_table,bq_table)
S3method(as_bq_table,character)
S3method(as_bq_table,default)
S3method(as_bq_table,list)
S3method(as_json,"NULL")
S3method(as_json,bq_field)
S3method(as_json,bq_fields)
S3method(as_json,bq_param_array)
S3method(as_json,bq_param_scalar)
S3method(as_json,bq_params)
S3method(format,bq_field)
S3method(format,bq_fields)
S3method(format,gs_object)
S3method(op_can_download,lazy_base_query)
S3method(op_can_download,lazy_query)
S3method(op_can_download,lazy_select_query)
S3method(op_can_download,tbl_lazy)
S3method(op_rows,lazy_base_query)
S3method(op_rows,lazy_select_query)
S3method(op_rows,tbl_lazy)
S3method(op_table,lazy_base_remote_query)
S3method(op_table,lazy_query)
S3method(op_table,lazy_select_query)
S3method(op_table,tbl_lazy)
S3method(print,bq_bytes)
S3method(print,bq_dataset)
S3method(print,bq_field)
S3method(print,bq_fields)
S3method(print,bq_job)
S3method(print,bq_param)
S3method(print,bq_params)
S3method(print,bq_table)
S3method(print,gs_object)
S3method(toString,bq_table)
export(as_bq_dataset)
export(as_bq_field)
export(as_bq_fields)
export(as_bq_job)
export(as_bq_table)
export(bigquery)
export(bq_auth)
export(bq_auth_configure)
export(bq_authable)
export(bq_dataset)
export(bq_dataset_create)
export(bq_dataset_delete)
export(bq_dataset_exists)
export(bq_dataset_meta)
export(bq_dataset_query)
export(bq_dataset_tables)
export(bq_dataset_update)
export(bq_deauth)
export(bq_field)
export(bq_fields)
export(bq_has_token)
export(bq_job)
export(bq_job_meta)
export(bq_job_show_statistics)
export(bq_job_status)
export(bq_job_wait)
export(bq_oauth_client)
export(bq_param)
export(bq_param_array)
export(bq_param_scalar)
export(bq_perform_copy)
export(bq_perform_extract)
export(bq_perform_load)
export(bq_perform_query)
export(bq_perform_query_dry_run)
export(bq_perform_query_schema)
export(bq_perform_upload)
export(bq_project_datasets)
export(bq_project_jobs)
export(bq_project_query)
export(bq_projects)
export(bq_table)
export(bq_table_copy)
export(bq_table_create)
export(bq_table_delete)
export(bq_table_download)
export(bq_table_exists)
export(bq_table_fields)
export(bq_table_load)
export(bq_table_meta)
export(bq_table_nrow)
export(bq_table_patch)
export(bq_table_save)
export(bq_table_size)
export(bq_table_upload)
export(bq_test_dataset)
export(bq_test_init)
export(bq_test_project)
export(bq_testable)
export(bq_token)
export(bq_user)
export(dbi_driver)
export(gs_test_bucket)
export(gs_test_object)
export(src_bigquery)
exportClasses(BigQueryConnection)
exportClasses(BigQueryDriver)
exportClasses(BigQueryResult)
exportMethods(dbAppendTable)
exportMethods(dbBegin)
exportMethods(dbBind)
exportMethods(dbClearResult)
exportMethods(dbColumnInfo)
exportMethods(dbCommit)
exportMethods(dbConnect)
exportMethods(dbCreateTable)
exportMethods(dbDataType)
exportMethods(dbDisconnect)
exportMethods(dbExecute)
exportMethods(dbExistsTable)
exportMethods(dbFetch)
exportMethods(dbGetInfo)
exportMethods(dbGetRowCount)
exportMethods(dbGetRowsAffected)
exportMethods(dbGetStatement)
exportMethods(dbHasCompleted)
exportMethods(dbIsValid)
exportMethods(dbListFields)
exportMethods(dbListTables)
exportMethods(dbQuoteIdentifier)
exportMethods(dbQuoteLiteral)
exportMethods(dbQuoteString)
exportMethods(dbReadTable)
exportMethods(dbRemoveTable)
exportMethods(dbRollback)
exportMethods(dbSendQuery)
exportMethods(dbWriteTable)
exportMethods(show)
import(DBI)
import(methods)
import(rlang, except = unbox)
importFrom(bit64,integer64)
importFrom(gargle,token_fetch)
importFrom(httr,DELETE)
importFrom(httr,GET)
importFrom(httr,PATCH)
importFrom(httr,POST)
importFrom(httr,PUT)
importFrom(httr,add_headers)
importFrom(httr,config)
importFrom(httr,content)
importFrom(httr,headers)
importFrom(httr,http_status)
importFrom(httr,parse_media)
importFrom(httr,status_code)
importFrom(jsonlite,unbox)
importFrom(lifecycle,deprecated)
importFrom(tibble,tibble)
useDynLib(bigrquery, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# bigrquery

<!-- badges: start -->

[![CRAN
Status](https://www.r-pkg.org/badges/version/bigrquery)](https://cran.r-project.org/package=bigrquery)
[![R-CMD-check](https://github.com/r-dbi/bigrquery/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-dbi/bigrquery/actions/workflows/R-CMD-check.yaml)

[![Codecov test
coverage](https://codecov.io/gh/r-dbi/bigrquery/graph/badge.svg)](https://app.codecov.io/gh/r-dbi/bigrquery)
<!-- badges: end -->

The bigrquery package makes it easy to work with data stored in [Google
BigQuery](https://docs.cloud.google.com/bigquery/docs) by allowing you
to query BigQuery tables and retrieve metadata about your projects,
datasets, tables, and jobs. The bigrquery package provides three levels
of abstraction on top of BigQuery:

- The low-level API provides thin wrappers over the underlying REST API.
  All the low-level functions start with `bq_`, and mostly have the form
  `bq_noun_verb()`. This level of abstraction is most appropriate if
  you’re familiar with the REST API and you want do something not
  supported in the higher-level APIs.

- The [DBI interface](https://r-dbi.org) wraps the low-level API and
  makes working with BigQuery like working with any other database
  system. This is most convenient layer if you want to execute SQL
  queries in BigQuery or upload smaller amounts (i.e. \<100 MB) of data.

- The [dplyr interface](https://dbplyr.tidyverse.org/) lets you treat
  BigQuery tables as if they are in-memory data frames. This is the most
  convenient layer if you don’t want to write SQL, but instead want
  dbplyr to write it for you.

## Installation

The current bigrquery release can be installed from CRAN:

``` r
install.packages("bigrquery")
```

The newest development release can be installed from GitHub:

``` r
#install.packages("pak")
pak::pak("r-dbi/bigrquery")
```

## Usage

### Low-level API

``` r
library(bigrquery)
billing <- bq_test_project() # replace this with your project ID 
sql <- "SELECT year, month, day, weight_pounds FROM `publicdata.samples.natality`"

tb <- bq_project_query(billing, sql)
bq_table_download(tb, n_max = 10)
#> # A tibble: 10 × 4
#>     year month   day weight_pounds
#>    <int> <int> <int>         <dbl>
#>  1  1969     9     5          7.00
#>  2  1969     6    16          8.06
#>  3  1969    12    17          5.94
#>  4  1969     9     1          7.44
#>  5  1969     4    25          7.94
#>  6  1969     3    13          7.31
#>  7  1969     4    17          7.13
#>  8  1969     7    19          7.00
#>  9  1969    12    13          7.75
#> 10  1969     7    22          6.38
```

### DBI

``` r
library(DBI)

con <- dbConnect(
  bigrquery::bigquery(),
  project = "publicdata",
  dataset = "samples",
  billing = billing
)
con 
#> <BigQueryConnection>
#>   Dataset: publicdata.samples
#>   Billing: gargle-169921

dbListTables(con)
#> [1] "github_nested"   "github_timeline" "gsod"            "natality"       
#> [5] "shakespeare"     "trigrams"        "wikipedia"

dbGetQuery(con, sql, n = 10)
#> # A tibble: 10 × 4
#>     year month   day weight_pounds
#>    <int> <int> <int>         <dbl>
#>  1  1969     9     5          7.00
#>  2  1969     6    16          8.06
#>  3  1969    12    17          5.94
#>  4  1969     9     1          7.44
#>  5  1969     4    25          7.94
#>  6  1969     3    13          7.31
#>  7  1969     4    17          7.13
#>  8  1969     7    19          7.00
#>  9  1969    12    13          7.75
#> 10  1969     7    22          6.38
```

### dplyr

``` r
library(dplyr)

natality <- tbl(con, "natality")

natality %>%
  select(year, month, day, weight_pounds) %>% 
  head(10) %>%
  collect()
#> # A tibble: 10 × 4
#>     year month   day weight_pounds
#>    <int> <int> <int>         <dbl>
#>  1  2005     8    NA          7.81
#>  2  2005    10    NA          7.10
#>  3  2005    12    NA          7.31
#>  4  2005     2    NA          3.75
#>  5  2005     1    NA          7.35
#>  6  2005    12    NA          3.19
#>  7  2005     7    NA          7.87
#>  8  2005     4    NA          8.31
#>  9  2005     4    NA          7.56
#> 10  2005    10    NA          7.19
```

## Important details

### BigQuery account

To use bigrquery, you’ll need a BigQuery project. Fortunately, if you
just want to play around with the BigQuery API, it’s easy to start with
Google’s free [public
data](https://docs.cloud.google.com/bigquery/public-data) and the
[BigQuery sandbox](https://docs.cloud.google.com/bigquery/docs/sandbox).
This gives you some fun data to play with along with enough free compute
(1 TB of queries & 10 GB of storage per month) to learn the ropes.

To get started, open <https://console.cloud.google.com/bigquery> and
create a project. Make a note of the “Project ID” as you’ll use this as
the `billing` project whenever you work with free sample data; and as
the `project` when you work with your own data.

### Authentication and authorization

When using bigrquery interactively, you’ll be prompted to [authorize
bigrquery](https://docs.cloud.google.com/bigquery/docs/authorization) in
the browser. You’ll be asked if you want to cache tokens for reuse in
future sessions. For non-interactive usage, it is preferred to use a
service account token, if possible. More places to learn about auth:

- Help for
  [`bigrquery::bq_auth()`](https://bigrquery.r-dbi.org/reference/bq_auth.html).
- [How gargle gets
  tokens](https://gargle.r-lib.org/articles/how-gargle-gets-tokens.html).
  - bigrquery obtains a token with `gargle::token_fetch()`, which
    supports a variety of token flows. This article provides full
    details, such as how to take advantage of Application Default
    Credentials or service accounts on GCE VMs.
- [Non-interactive
  auth](https://gargle.r-lib.org/articles/non-interactive-auth.html).
  Explains how to set up a project when code must run without any user
  interaction.
- [How to get your own API
  credentials](https://gargle.r-lib.org/articles/get-api-credentials.html).
  Instructions for getting your own OAuth client or service account
  token.

Note that bigrquery requests permission to modify your data; but it will
never do so unless you explicitly request it (e.g. by calling
`bq_table_delete()` or `bq_table_upload()`). Our [Privacy
policy](https://tidyverse.org/google_privacy_policy/) provides more
info.

## Useful links

- [SQL
  reference](https://docs.cloud.google.com/bigquery/docs/reference/standard-sql/functions-and-operators)
- [API
  reference](https://docs.cloud.google.com/bigquery/docs/reference/rest)
- [Query/job console](https://console.cloud.google.com/bigquery/)
- [Billing console](https://console.cloud.google.com/)

## Policies

Please note that the ‘bigrquery’ project is released with a [Contributor
Code of Conduct](https://bigrquery.r-dbi.org/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

[Privacy policy](https://tidyverse.org/google_privacy_policy/)

#' A BigQuery data source for dplyr.
#'
#' Create the connection to the database with `DBI::dbConnect()` then
#' use [dplyr::tbl()] to connect to tables within that database. Generally,
#' it's best to provide the fully qualified name of the table (i.e.
#' `project.dataset.table`) but if you supply a default `dataset` in the
#' connection, you can use just the table name. (This, however, will
#' prevent you from making joins across datasets.)
#'
#' @param project project id or name
#' @param dataset dataset name
#' @param billing billing project, if different to `project`
#' @param max_pages (IGNORED) maximum pages returned by a query
#' @export
#' @examples
#' \dontrun{
#' library(dplyr)
#'
#' # To run this example, replace billing with the id of one of your projects
#' # set up for billing
#' con <- DBI::dbConnect(bigquery(), project = bq_test_project())
#'
#' shakespeare <- con %>% tbl(I("publicdata.samples.shakespeare"))
#' shakespeare
#' shakespeare %>%
#'   group_by(word) %>%
#'   summarise(n = sum(word_count, na.rm = TRUE)) %>%
#'   arrange(desc(n))
#' }
src_bigquery <- function(project, dataset, billing = project, max_pages = 10) {
  check_installed("dbplyr")

  con <- DBI::dbConnect(
    bigquery(),
    project = project,
    dataset = dataset,
    billing = billing,
    use_legacy_sql = FALSE
  )

  dbplyr::src_dbi(con)
}

# registered onLoad
tbl.BigQueryConnection <- function(src, from, ...) {
  src <- dbplyr::src_dbi(src, auto_disconnect = FALSE)

  sql <- dbplyr::sql_query_fields(src$con, from)
  dataset <- if (!is.null(src$con@dataset)) as_bq_dataset(src$con)
  schema <- bq_perform_query_schema(
    sql,
    billing = src$con@billing,
    default_dataset = dataset,
    labels = src$con@labels
  )
  vars <- map_chr(schema, "[[", "name")

  if (utils::packageVersion("dbplyr") >= "2.4.0.9000") {
    tbl <- dplyr::tbl(src, from = from, vars = vars)
  } else {
    tbl <- dplyr::tbl(src, from = from, vars = vars, check_from = FALSE)
  }

  # This is ugly, but I don't see a better way of doing this
  if (!inherits(from, "sql")) {
    tb <- as_bq_table(src$con, from)
    tbl$lazy_query$is_view <- tryCatch(
      bq_table_meta(tb, "type")$type == "VIEW",
      bigrquery_notFound = function(e) TRUE
    )
  } else {
    tbl$lazy_query$is_view <- FALSE
  }
  tbl
}

# registered onLoad
dbplyr_edition.BigQueryConnection <- function(con) 2L

# registered onLoad
db_compute.BigQueryConnection <- function(
  con,
  table,
  sql,
  ...,
  overwrite = FALSE,
  temporary = TRUE,
  unique_indexes = list(),
  indexes = list(),
  analyze = TRUE,
  in_transaction = FALSE
) {
  if (is.null(con@dataset)) {
    destination_table <- if (!temporary) as_bq_table(con, table)
    tb <- bq_project_query(
      con@project,
      sql,
      destination_table = destination_table,
      labels = con@labels
    )
  } else {
    ds <- bq_dataset(con@project, con@dataset)
    destination_table <- if (!temporary) as_bq_table(con, table)

    tb <- bq_dataset_query(
      ds,
      query = sql,
      destination_table = destination_table,
      labels = con@labels
    )
  }

  dbplyr::as_table_path(
    I(paste(tb$project, tb$dataset, tb$table, sep = ".")),
    con
  )
}

# registered onLoad
db_copy_to.BigQueryConnection <- function(
  con,
  table,
  values,
  ...,
  overwrite = FALSE,
  types = NULL,
  temporary = TRUE,
  unique_indexes = NULL,
  indexes = NULL,
  analyze = TRUE,
  in_transaction = TRUE
) {
  if (temporary) {
    cli::cli_abort("BigQuery does not support temporary tables")
  }

  tb <- as_bq_table(con, table)
  write <- if (overwrite) "WRITE_TRUNCATE" else "WRITE_EMPTY"
  bq_table_upload(
    tb,
    values,
    fields = types,
    write_disposition = write,
    labels = con@labels
  )

  table
}

# Efficient downloads -----------------------------------------------

# registered onLoad

#' Collect a BigQuery table
#'
#' This collect method is specialised for BigQuery tables, generating the
#' SQL from your dplyr commands, then calling [bq_project_query()]
#' or [bq_dataset_query()] to run the query, then [bq_table_download()]
#' to download the results. Thus the arguments are a combination of the
#' arguments to [dplyr::collect()], `bq_project_query()`/`bq_dataset_query()`,
#' and `bq_table_download()`.
#'
#' @inheritParams dplyr::collect
#' @inheritParams bq_table_download
#' @param n Maximum number of results to retrieve.
#'   The default, `Inf`, will retrieve all rows.
#' @param ... Other arguments passed on to
#'   `bq_project_query()`/`bq_project_query()`
#' @method collect tbl_BigQueryConnection
collect.tbl_BigQueryConnection <- function(
  x,
  ...,
  n = Inf,
  api = c("json", "arrow"),
  page_size = NULL,
  max_connections = 6L
) {
  api <- check_api(api)
  check_number_whole(n, min = 0, allow_infinite = TRUE)
  check_number_whole(max_connections, min = 1)

  con <- dbplyr::remote_con(x)
  billing <- con@billing

  if (op_can_download(x)) {
    lq <- x$lazy_query
    name <- op_table(x, con)
    tb <- as_bq_table(con, name)
    n <- min(op_rows(x$lazy_query), n)
  } else {
    sql <- dbplyr::db_sql_render(con, x)

    if (is.null(con@dataset)) {
      tb <- bq_project_query(
        billing,
        sql,
        quiet = con@quiet,
        labels = con@labels,
        ...
      )
    } else {
      ds <- as_bq_dataset(con)
      tb <- bq_dataset_query(
        ds,
        sql,
        quiet = con@quiet,
        billing = billing,
        labels = con@labels,
        ...
      )
    }
  }

  quiet <- if (n < 100) TRUE else con@quiet
  bigint <- con@bigint %||% "integer"

  if (api == "arrow") {
    out <- bq_table_download(
      tb,
      n_max = n,
      quiet = quiet,
      bigint = bigint,
      billing = billing,
      api = "arrow"
    )
  } else {
    out <- bq_table_download(
      tb,
      n_max = n,
      page_size = page_size,
      quiet = quiet,
      max_connections = max_connections,
      bigint = bigint,
      api = "json"
    )
  }

  dplyr::grouped_df(out, intersect(dbplyr::op_grps(x), names(out)))
}

# Can download directly if only head and

op_can_download <- function(x) UseMethod("op_can_download")
#' @export
op_can_download.tbl_lazy <- function(x) op_can_download(x$lazy_query)
#' @export
op_can_download.lazy_query <- function(x) FALSE
#' @export
op_can_download.lazy_select_query <- function(x) {
  query_is_head_only(x)
}
#' @export
op_can_download.lazy_base_query <- function(x) {
  if (isTRUE(x$is_view)) {
    FALSE
  } else if (inherits(x$x, "sql")) {
    FALSE
  } else {
    dbplyr::is.ident(x$x) ||
      inherits(x$x, "dbplyr_table_ident") ||
      inherits(x$x, "dbplyr_table_path")
  }
}

query_is_head_only <- function(x) {
  if (!inherits(x$x, "lazy_base_remote_query")) {
    return(FALSE)
  }
  if (!op_can_download(x$x)) {
    return(FALSE)
  }

  vars_base <- dbplyr::op_vars(x$x)
  if (!is_select_trivial(x$select, vars_base)) {
    return(FALSE)
  }

  if (!is_empty(x$where)) {
    return(FALSE)
  }
  if (!is_empty(x$order_by)) {
    return(FALSE)
  }
  if (!is_false(x$distinct)) {
    return(FALSE)
  }

  TRUE
}

is_select_trivial <- function(select, vars_prev) {
  identical(select$name, vars_prev) &&
    all(vapply(select$expr, is_symbol, logical(1))) &&
    identical(syms(select$name), select$expr)
}


op_rows <- function(x) UseMethod("op_rows")
#' @export
op_rows.tbl_lazy <- function(x) op_rows(x$lazy_query)
#' @export
op_rows.lazy_base_query <- function(x) Inf
#' @export
op_rows.lazy_select_query <- function(x) {
  min(x$limit, op_rows(x$x))
}

op_table <- function(x, con) UseMethod("op_table")
#' @export
op_table.tbl_lazy <- function(x, con) op_table(x$lazy_query, con)
#' @export
op_table.lazy_query <- function(x, con) NULL
#' @export
op_table.lazy_base_remote_query <- function(x, con) x$x
#' @export
op_table.lazy_select_query <- function(x, con) {
  if (!query_is_head_only(x)) {
    return(NULL)
  }

  op_table(x$x)
}


# exported onLoad
same_src.tbl_BigQueryConnection <- function(x, y) {
  inherits(y, "tbl_BigQueryConnection")
}

# SQL translation -------------------------------------------------------------

# Don't import to avoid build-time dependency
sql_prefix <- function(f, n = NULL) {
  dbplyr::sql_prefix(f = f, n = n)
}

# registered onLoad
sql_join_suffix.BigQueryConnection <- function(con, ...) {
  c("_x", "_y")
}

# registered onLoad
sql_translation.BigQueryConnection <- function(x) {
  dbplyr::sql_variant(
    dbplyr::sql_translator(
      .parent = dbplyr::base_scalar,

      `^` = sql_prefix("POW"),
      `%%` = sql_prefix("MOD"),
      "%||%" = sql_prefix("IFNULL"),

      # Coercion
      as.integer = function(x) dbplyr::build_sql("SAFE_CAST(", x, " AS INT64)"),
      as.logical = function(x) {
        dbplyr::build_sql("SAFE_CAST(", x, " AS BOOLEAN)")
      },
      as.numeric = function(x) {
        dbplyr::build_sql("SAFE_CAST(", x, " AS FLOAT64)")
      },
      as.character = function(x) {
        dbplyr::build_sql("SAFE_CAST(", x, " AS STRING)")
      },

      # Date/time
      Sys.date = sql_prefix("current_date"),
      Sys.time = sql_prefix("current_time"),

      # Regular expressions
      grepl = function(pattern, x) {
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/string_functions#regexp_contains
        dbplyr::build_sql("REGEXP_CONTAINS", list(x, pattern))
      },
      gsub = function(pattern, replace, x) {
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/string_functions#regexp_replace
        dbplyr::build_sql("REGEXP_REPLACE", list(x, pattern, replace))
      },

      # Other scalar functions
      ifelse = sql_prefix("IF"),

      # string
      paste0 = sql_prefix("CONCAT"),

      # stringr equivalents
      str_detect = sql_prefix("REGEXP_CONTAINS", 2),
      str_extract = sql_prefix("REGEXP_EXTRACT", 2),
      str_replace = sql_prefix("REGEXP_REPLACE", 3),

      # Parallel min and max
      pmax = sql_prefix("GREATEST"),
      pmin = sql_prefix("LEAST"),

      runif = function(n = n(), min = 0, max = 1) {
        RAND <- NULL # quiet R CMD check
        dbplyr::sql_runif(RAND(), n = {{ n }}, min = min, max = max)
      },
    ),
    dbplyr::sql_translator(
      .parent = dbplyr::base_agg,

      n = function() dplyr::sql("count(*)"),

      all = sql_prefix("LOGICAL_AND", 1),
      any = sql_prefix("LOGICAL_OR", 1),

      sd = sql_prefix("STDDEV_SAMP"),
      var = sql_prefix("VAR_SAMP"),
      cor = dbplyr::sql_aggregate_2("CORR"),
      cov = dbplyr::sql_aggregate_2("COVAR_SAMP"),

      # Median
      median = function(x, na.rm = TRUE) {
        dbplyr::build_sql("APPROX_QUANTILES(", x, ", 2)[SAFE_ORDINAL(2)]")
      }
    ),
    dbplyr::sql_translator(
      .parent = dbplyr::base_win,

      all = dbplyr::win_absent("LOGICAL_AND"),
      any = dbplyr::win_absent("LOGICAL_OR"),

      sd = dbplyr::win_recycled("STDDEV_SAMP"),
      var = dbplyr::win_recycled("VAR_SAMP"),
      cor = dbplyr::win_absent("CORR"),
      cov = dbplyr::win_absent("COVAR_SAMP"),

      n_distinct = dbplyr::win_absent("n_distinct"),

      median = dbplyr::win_absent("median")
    )
  )
}

simulate_bigrquery <- function(use_legacy_sql = FALSE) {
  new(
    "BigQueryConnection",
    project = "test",
    dataset = "test",
    billing = "test",
    use_legacy_sql = use_legacy_sql,
    page_size = 0L,
    quiet = TRUE
  )
}

test_that("historical API continues to work", {
  src <- src_bigquery(bq_test_project(), "basedata")

  x <- dplyr::tbl(src, "mtcars")

  expect_s3_class(x, "tbl")
  expect_true("cyl" %in% dbplyr::op_vars(x))
})

test_that("can work with literal SQL", {
  con_us <- DBI::dbConnect(
    bigquery(),
    project = "bigquery-public-data",
    dataset = "utility_us",
    billing = bq_test_project()
  )
  x <- dplyr::tbl(con_us, dplyr::sql("SELECT * FROM country_code_iso"))

  expect_s3_class(x, "tbl")
  expect_true("fips_code" %in% dbplyr::op_vars(x))

  # If if has three or more periods (#611)
  y <- dplyr::tbl(
    con_us,
    dplyr::sql("SELECT c.fips_code, c.country_name FROM country_code_iso c")
  )
  expect_s3_class(y, "tbl")
})

test_that("can work with nested table identifier", {
  con1 <- DBI::dbConnect(
    bigquery(),
    project = "bigquery-public-data",
    billing = bq_test_project()
  )
  # As far as I can tell from the BigQuery API there's no way to provide
  # a default project; you can either provide a default dataset + project or
  # nothing
  table_name <- I("bigquery-public-data.utility_us.country_code_iso")
  expect_no_error(dplyr::collect(head(dplyr::tbl(con1, table_name))))

  con2 <- DBI::dbConnect(
    bigquery(),
    project = "bigquery-public-data",
    dataset = "utility_us",
    billing = bq_test_project(),
  )
  expect_no_error(dplyr::collect(head(dplyr::tbl(con2, "country_code_iso"))))
})

test_that("can copy_to", {
  ds <- bq_test_dataset()
  con <- DBI::dbConnect(ds)

  expect_snapshot(dplyr::copy_to(con, mtcars), error = TRUE)
  bq_mtcars <- dplyr::copy_to(con, mtcars, temporary = FALSE)

  expect_s3_class(bq_mtcars, "tbl_BigQueryConnection")
})

test_that("can collect and compute (no dataset)", {
  con <- DBI::dbConnect(bigquery(), project = bq_test_project())
  bq_mtcars <- dplyr::tbl(con, "basedata.mtcars") %>% dplyr::filter(cyl == 4)

  # collect
  x <- dplyr::collect(bq_mtcars)
  expect_equal(dim(x), c(11, 11))

  # compute: temporary
  temp <- dplyr::compute(bq_mtcars)
  expect_equal(dim(dplyr::collect(temp)), c(11, 11))

  # compute: persistent
  name <- paste0("basedata.", random_name())
  if (packageVersion("dbplyr") >= "2.4.0.9000") {
    name <- I(name)
  }
  perm <- dplyr::compute(bq_mtcars, name = name, temporary = FALSE)
  defer(DBI::dbRemoveTable(con, name))

  expect_true(DBI::dbExistsTable(con, name))
  expect_equal(dim(dplyr::collect(perm)), c(11, 11))
})

test_that("can collect and compute (with dataset)", {
  con <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project(),
    dataset = "basedata"
  )
  bq_mtcars <- dplyr::tbl(con, "mtcars") %>% dplyr::filter(cyl == 4)

  # collect
  x <- dplyr::collect(bq_mtcars)
  expect_equal(dim(x), c(11, 11))

  # compute: temporary
  temp <- dplyr::compute(bq_mtcars)
  expect_equal(dim(dplyr::collect(temp)), c(11, 11))

  # compute: persistent
  name <- random_name()
  perm <- dplyr::compute(bq_mtcars, temporary = FALSE, name = name)
  expect_equal(dim(dplyr::collect(perm)), c(11, 11))
  defer(DBI::dbRemoveTable(con, name))

  expect_true(DBI::dbExistsTable(con, name))
})

test_that("collect can identify directly download tables", {
  con <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project(),
    dataset = "basedata"
  )

  bq1 <- dplyr::tbl(con, "mtcars")
  expect_true(op_can_download(bq1))
  expect_equal(op_rows(bq1), Inf)
  expect_equal(format(op_table(bq1)), "`mtcars`")

  bq2 <- head(bq1, 4)
  expect_true(op_can_download(bq2))
  expect_equal(op_rows(bq2), 4)
  expect_equal(format(op_table(bq1)), "`mtcars`")

  bq3 <- head(bq2, 2)
  expect_true(op_can_download(bq3))
  expect_equal(op_rows(bq3), 2)

  bq3 <- dplyr::filter(bq1, cyl == 1)
  expect_false(op_can_download(bq3))

  x <- dplyr::collect(bq1)
  expect_s3_class(x, "tbl")

  bq5 <- dplyr::tbl(con, dplyr::sql("SELECT * FROM mtcars"))
  expect_false(op_can_download(bq5))
  expect_false(op_can_download(head(bq5)))

  view_name <- random_name()
  DBI::dbExecute(
    con,
    paste0("CREATE VIEW ", view_name, " AS SELECT * FROM basedata.mtcars")
  )
  defer(DBI::dbExecute(con, paste0("DROP VIEW ", view_name)))
  bq6 <- dplyr::tbl(con, view_name)
  expect_false(op_can_download(bq6))

  # INFORMATION_SCHEMA.TABLES errors when we attempt to get metadata
  bq7 <- dplyr::tbl(con, "INFORMATION_SCHEMA.PARTITIONS")
  expect_false(op_can_download(bq7))
  expect_false(op_can_download(head(bq7)))
})

test_that("casting uses bigquery types", {
  skip_if_not_installed("dbplyr")
  con <- simulate_bigrquery()

  expect_equal(
    dbplyr::translate_sql(as.integer(x), con = con),
    dbplyr::sql("SAFE_CAST(`x` AS INT64)")
  )
  expect_equal(
    dbplyr::translate_sql(as.numeric(x), con = con),
    dbplyr::sql("SAFE_CAST(`x` AS FLOAT64)")
  )
})

test_that("%||% translates to IFNULL", {
  skip_if_not_installed("dbplyr")

  expect_equal(
    dbplyr::translate_sql(x %||% 2L, con = simulate_bigrquery()),
    dbplyr::sql("IFNULL(`x`, 2)")
  )
})

test_that("suffixes use _", {
  skip_if_not_installed("dbplyr", "1.99")

  expect_equal(dbplyr::sql_join_suffix(simulate_bigrquery()), c("_x", "_y"))
})

test_that("all BigQuery tbls share the same src", {
  con1 <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project()
  )
  con2 <- DBI::dbConnect(
    bigquery(),
    project = "publicdata",
    billing = bq_test_project()
  )

  tbl1 <- dplyr::tbl(con1, I("basedata.mtcars"), vars = "x")
  tbl2 <- dplyr::tbl(con2, I("publicdata.samples.natality"), vars = "x")
  expect_true(dplyr::same_src(tbl1, tbl2))
  expect_false(dplyr::same_src(tbl1, mtcars))
})

test_that("runif is correctly translated", {
  skip_if_not_installed("dbplyr", "2.4.0")
  expect_equal(
    dbplyr::translate_sql(runif(n()), con = simulate_bigrquery()),
    dbplyr::sql("RAND()")
  )
})

test_that("string functions correctly", {
  con <- simulate_bigrquery()

  expect_snapshot({
    dbplyr::translate_sql(grepl("a.c", x), con = con)
    dbplyr::translate_sql(gsub("a.c", "", x), con = con)
  })
})

test_that("median is correctly translated", {
  expect_equal(
    dbplyr::translate_sql(
      median(x),
      con = simulate_bigrquery(),
      window = FALSE
    ),
    dbplyr::sql("APPROX_QUANTILES(`x`, 2)[SAFE_ORDINAL(2)]")
  )
})

test_that("can correctly print a lazy query", {
  con <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project(),
    dataset = "basedata"
  )

  bq_mtcars <- dplyr::tbl(con, "mtcars")

  # not a snapshot test because column order can vary
  expect_no_error(
    expect_output(
      print(bq_mtcars)
    )
  )
})

.onLoad <- function(libname, pkgname) {
  .auth <<- gargle::init_AuthState(
    package = "bigrquery",
    auth_active = TRUE
  )

  if (has_internal_auth() && in_pkgdown()) {
    bq_auth_internal()
  }

  # S3 methods --------------------------------------------------------------
  s3_register("dplyr::tbl", "BigQueryConnection")
  s3_register("dplyr::collect", "tbl_BigQueryConnection")
  s3_register("dplyr::same_src", "tbl_BigQueryConnection")

  s3_register("dbplyr::dbplyr_edition", "BigQueryConnection")
  s3_register("dbplyr::db_compute", "BigQueryConnection")
  s3_register("dbplyr::db_copy_to", "BigQueryConnection")
  s3_register("dbplyr::sql_join_suffix", "BigQueryConnection")
  s3_register("dbplyr::sql_translation", "BigQueryConnection")

  # Default options --------------------------------------------------------
  op <- options()
  defaults <- list(
    bigrquery.quiet = NA,
    bigrquery.page.size = 1e4,
    bigrquery.labels = NULL
  )
  toset <- !(names(defaults) %in% names(op))
  if (any(toset)) {
    options(defaults[toset])
  }

  invisible()
}

PACKAGE_NAME <- utils::packageName()
PACKAGE_VERSION <- utils::packageVersion(PACKAGE_NAME)

#' S3 classes for BigQuery datasets, tables and jobs
#'
#' @description
#' Create references to BigQuery datasets, jobs, and tables. Each class
#' has a constructor function (`bq_dataset()`, `bq_table()`, `bq_job()`)
#' and a coercion function (`as_bq_dataset()`, `as_bq_table()`, `as_bq_job()`).
#' The coercions functions come with methods for strings (which find components
#' by splitting on `.`), and lists (which look for named components like
#' `projectId` or `project_id`).
#'
#' All `bq_table_`, `bq_dataset_` and `bq_job_` functions call the appropriate
#' coercion functions on their first argument, allowing you to flexible specify
#' their inputs.
#'
#' @param project,dataset,table,job,type Individual project, dataset, table,
#'   job identifiers and table type (strings).
#'
#'   For `bq_table()`, you if supply a `bq_dataset` as the first argument,
#'   the 2nd argument will be interpreted as the `table`
#' @param x An object to coerce to a `bq_job`, `bq_dataset`, or `bq_table`.
#'   Built-in methods handle strings and lists.
#' @inheritParams rlang::args_error_context
#' @param ... Other arguments passed on to methods.
#' @seealso [api-job], [api-perform], [api-dataset], and [api-table] for
#'   functions that work with these objects.
#' @examples
#' # Creation ------------------------------------------------
#' samples <- bq_dataset("publicdata", "samples")
#' natality <- bq_table("publicdata", "samples", "natality")
#' natality
#'
#' # Or
#' bq_table(samples, "natality")
#'
#' bq_job("bigrquery-examples", "m0SgFu2ycbbge6jgcvzvflBJ_Wft")
#'
#' # Coercion ------------------------------------------------
#' as_bq_dataset("publicdata.shakespeare")
#' as_bq_table("publicdata.samples.natality")
#'
#' as_bq_table(list(
#'   project_id = "publicdata",
#'   dataset_id = "samples",
#'   table_id = "natality"
#' ))
#'
#' as_bq_job(list(
#'   projectId = "bigrquery-examples",
#'   jobId = "job_m0SgFu2ycbbge6jgcvzvflBJ_Wft",
#'   location = "US"
#' ))
#'
#' @name bq_refs
#' @aliases NULL
NULL

# dataset -----------------------------------------------------------------

#' @rdname bq_refs
#' @export
bq_dataset <- function(project, dataset) {
  check_string(project)
  check_string(dataset)

  structure(
    list(
      project = project,
      dataset = dataset
    ),
    class = "bq_dataset"
  )
}

setOldClass("bq_dataset")

#' @export
print.bq_dataset <- function(x, ...) {
  cat_line("<bq_dataset> ", x$project, ".", x$dataset)
  invisible(x)
}

#' @rdname bq_refs
#' @export
as_bq_dataset <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  UseMethod("as_bq_dataset")
}

#' @export
as_bq_dataset.bq_dataset <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  x
}

#' @export
as_bq_dataset.character <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  x <- bq_from_string(x, n = 2, error_arg = error_arg, error_call = error_call)
  bq_dataset(x[[1]], x[[2]])
}

#' @export
as_bq_dataset.list <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  x <- bq_from_list(
    x,
    names = c("projectId", "datasetId"),
    error_arg = error_arg,
    error_call = error_call
  )
  bq_dataset(x$projectId, x$datasetId)
}

#' @export
as_bq_dataset.default <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  cli::cli_abort(
    "{.arg {error_arg}} must be a string, list, or {.fun bq_dataset}.",
    call = error_call
  )
}


# table -------------------------------------------------------------------

#' @rdname bq_refs
#' @export
bq_table <- function(project, dataset, table = NULL, type = "TABLE") {
  if (inherits(project, "bq_dataset") && is.null(table)) {
    check_string(dataset)
    table <- dataset
    dataset <- project$dataset
    project <- project$project
  } else {
    check_string(project)
    check_string(dataset)
    check_string(table)
  }

  structure(
    list(
      project = project,
      dataset = dataset,
      table = table,
      type = type
    ),
    class = "bq_table"
  )
}

setOldClass("bq_table")

#' @export
print.bq_table <- function(x, ...) {
  cat_line("<bq_table> ", toString(x))
  invisible(x)
}

#' @export
toString.bq_table <- function(x, ...) {
  paste0(x$project, ".", x$dataset, ".", x$table)
}

#' @rdname bq_refs
#' @export
as_bq_table <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  UseMethod("as_bq_table")
}

#' @export
as_bq_table.bq_table <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  x
}

#' @export
as_bq_table.character <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  x <- bq_from_string(x, n = 3, error_arg = error_arg, error_call = error_call)
  bq_table(x[[1]], x[[2]], x[[3]], ...)
}

#' @export
as_bq_table.list <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  x <- bq_from_list(
    x,
    names = c("projectId", "datasetId", "tableId"),
    error_arg = error_arg,
    error_call = error_call
  )
  bq_table(x$projectId, x$datasetId, x$tableId, ...)
}

#' @export
as_bq_table.default <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  cli::cli_abort(
    "{.arg {error_arg}} must be a string, list, or {.fun bq_table}.",
    call = error_call
  )
}


# job ---------------------------------------------------------------------

#' @rdname bq_refs
#' @param location Job location
#' @export
bq_job <- function(project, job, location = "US") {
  structure(
    list(
      project = project,
      job = job,
      location = location
    ),
    class = "bq_job"
  )
}

#' @rdname bq_refs
#' @export
as_bq_job <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  UseMethod("as_bq_job")
}

#' @export
as_bq_job.bq_job <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  x
}

#' @export
as_bq_job.list <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  x <- bq_from_list(
    x,
    c("projectId", "jobId", "location"),
    error_arg = error_arg,
    error_call = error_call
  )
  bq_job(x$projectId, x$jobId, x$location)
}

#' @export
as_bq_job.default <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  cli::cli_abort(
    "{.arg {error_arg}} must be a string, list, or {.fun bq_job}.",
    call = error_call
  )
}

#' @export
as_bq_job.character <- function(
  x,
  ...,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  x <- bq_from_string(x, 3, error_arg = error_arg, error_call = error_call)
  bq_job(x[[1]], x[[2]], x[[3]])
}

#' @export
print.bq_job <- function(x, ...) {
  cat_line("<bq_job> ", as.character(x))
  invisible(x)
}

#' @export
as.character.bq_job <- function(x, ...) {
  paste0(x$project, ".", x$job, ".", x$location)
}

# JSON --------------------------------------------------------------------

datasetReference <- function(x) {
  x <- as_bq_dataset(x)
  list(
    projectId = unbox(x$project),
    datasetId = unbox(x$dataset)
  )
}

tableReference <- function(x) {
  x <- as_bq_table(x)
  list(
    projectId = unbox(x$project),
    datasetId = unbox(x$dataset),
    tableId = unbox(x$table)
  )
}

# Helpers -----------------------------------------------------------------

bq_from_list <- function(
  x,
  names,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  names(x) <- camelCase(names(x))

  if (length(setdiff(names, names(x))) == 0) {
    return(x)
  }

  cli::cli_abort(
    "When {.arg {error_arg}} is a list, it must have components {.and {.str {names}}}.",
    call = error_call
  )
}

bq_from_string <- function(
  x,
  n,
  error_arg = caller_arg(x),
  error_call = caller_env()
) {
  check_string(x, call = error_call)

  pieces <- strsplit(x, ".", fixed = TRUE)[[1]]
  if (length(pieces) != n) {
    cli::cli_abort(
      "When {.arg {error_arg}} is a string, it must contain {n} components separted by {.str .}.",
      call = error_call
    )
  }
  pieces
}

test_that("can connect and disconnect", {
  con <- DBI::dbConnect(bigquery(), project = "p")
  expect_true(dbIsValid(con))

  expect_true(dbDisconnect(con))
  # But still valid, because there's actually no long-lived connection
  expect_true(dbIsValid(con))
})

test_that("can round-trip connection to bq_dataset", {
  ds <- bq_dataset("x", "y")
  con <- DBI::dbConnect(ds)
  expect_equal(as_bq_dataset(con), ds)
})

test_that("useful print with and without dataset", {
  con1 <- DBI::dbConnect(
    bigquery(),
    project = "p",
    dataset = "x",
    billing = "b"
  )
  con2 <- DBI::dbConnect(bigquery(), project = "p")

  expect_snapshot({
    "With dataset"
    con1

    "Without dataset"
    con2
  })
})

test_that("uses BigQuery quoting standards", {
  con <- DBI::dbConnect(bigquery(), project = "p")

  expect_equal(DBI::dbQuoteString(con, "x"), DBI::SQL("'x'"))
  expect_equal(DBI::dbQuoteIdentifier(con, "x"), DBI::SQL("`x`"))

  # Doesn't double-quote
  x <- SQL("x")
  expect_equal(DBI::dbQuoteString(con, x), x)
  expect_equal(DBI::dbQuoteIdentifier(con, x), x)

  # Returns 0-length outputs
  expect_equal(DBI::dbQuoteString(con, character()), SQL(character()))
  expect_equal(DBI::dbQuoteIdentifier(con, character()), SQL(character()))
})

test_that("dbQuoteIdentifier validates inputs", {
  con <- DBI::dbConnect(bigquery(), project = "")
  expect_snapshot(DBI::dbQuoteIdentifier(con, c("x", NA)), error = TRUE)
})

test_that("can retrieve information about public dataset", {
  con <- DBI::dbConnect(
    bigquery(),
    project = "publicdata",
    dataset = "samples",
    billing = bq_test_project()
  )

  fields <- DBI::dbListFields(con, "natality")
  expect_length(fields, 31)

  expect_true("natality" %in% DBI::dbListTables(con))

  df <- DBI::dbReadTable(con, "natality", n_max = 10)
  expect_equal(nrow(df), 10)
})

test_that("can roundtrip a data frame", {
  ds <- bq_test_dataset()
  con <- DBI::dbConnect(ds)

  DBI::dbWriteTable(con, "mtcars", mtcars)
  df <- DBI::dbReadTable(con, "mtcars")

  expect_equal(nrow(df), 32)
  expect_equal(ncol(df), 11)
})

test_that("dbWriteTable errors on unsupported arguments", {
  con <- DBI::dbConnect(bigquery(), project = "")
  df <- data.frame(x = 1)

  expect_snapshot(error = TRUE, {
    DBI::dbWriteTable(con, "x", df, field.types = list())
    DBI::dbWriteTable(con, "x", df, temporary = TRUE)
  })
})

test_that("can execute a query", {
  tb <- bq_test_table()
  con <- DBI::dbConnect(bq_dataset(tb$project, tb$dataset))

  DBI::dbWriteTable(con, tb$table, data.frame(x = 1:4))
  out <- dbExecute(con, sprintf("UPDATE %s SET x = x + 1 WHERE true", tb$table))
  expect_equal(out, 4)

  out <- dbExecute(con, sprintf("DELETE %s WHERE x <= 3", tb$table))
  expect_equal(out, 2)
})

test_that("can use params", {
  con <- DBI::dbConnect(bigquery(), project = bq_test_project())

  # in dbGetQuery
  df <- DBI::dbGetQuery(con, "SELECT @x AS value", params = list(x = 1))
  expect_equal(df, tibble(value = 1))

  # in dbExecute
  tb <- bq_test_table()
  con <- DBI::dbConnect(bq_dataset(tb$project, tb$dataset))
  DBI::dbWriteTable(con, tb$table, data.frame(x = 1:4))

  out <- dbExecute(
    con,
    sprintf("DELETE %s WHERE x <= @x", tb$table),
    params = list(x = 2)
  )
  expect_equal(out, 2)
})


test_that("can't use parameters", {
  con <- DBI::dbConnect(bigquery(), project = bq_test_project())

  # in dbGetQuery
  expect_snapshot(error = TRUE, {
    DBI::dbGetQuery(con, "SELECT @x AS value", parameters = list(x = 1))
    DBI::dbExecute(con, "DELETE table WHERE x <= @x", parameters = list(x = 1))
  })
})


test_that("can use DBI::Id()", {
  ds <- bq_test_dataset()
  con <- DBI::dbConnect(ds)

  df <- data.frame(x = 1:10)
  id <- DBI::Id(table = "mytable")

  expect_no_error(DBI::dbCreateTable(con, id, df))
  expect_no_error(DBI::dbAppendTable(con, id, df))
  expect_no_error(DBI::dbWriteTable(con, id, df, overwrite = TRUE))
  expect_no_error(DBI::dbReadTable(con, id))
  expect_true(DBI::dbExistsTable(con, id))
  expect_no_error(DBI::dbListFields(con, id))
  expect_no_error(DBI::dbRemoveTable(con, id))
})

test_that("can create an empty dataset then append to it", {
  tb <- bq_test_table()
  con <- DBI::dbConnect(bq_dataset(tb$project, tb$dataset))

  df <- data.frame(x = 1, y = 2)
  DBI::dbCreateTable(con, tb$table, df)
  expect_equal(bq_table_nrow(tb), 0)

  # With dbWriteTable
  DBI::dbWriteTable(con, tb$table, df, append = TRUE)
  expect_equal(bq_table_nrow(tb), 1)

  # Or with dbAppend
  DBI::dbAppendTable(con, tb$table, df)
  expect_equal(bq_table_nrow(tb), 2)
})

test_that("dataset is optional", {
  con <- DBI::dbConnect(bigquery(), project = bq_test_project())
  expect_snapshot(DBI::dbListTables(con), error = TRUE)

  df <- DBI::dbReadTable(con, "publicdata.samples.natality", n_max = 10)
  expect_equal(ncol(df), 31)

  expect_snapshot(DBI::dbReadTable(con, "natality", n_max = 10), error = TRUE)
})

test_that("can create bq_table from connection + name", {
  con1 <- DBI::dbConnect(bigquery(), project = "p")
  expect_snapshot(as_bq_table(con1, "x"), error = TRUE)
  expect_equal(as_bq_table(con1, "x.y"), as_bq_table("p.x.y"))
  expect_equal(as_bq_table(con1, "x.y.z"), as_bq_table("x.y.z"))
  expect_snapshot(as_bq_table(con1, "a.b.c.d"), error = TRUE)

  con2 <- DBI::dbConnect(bigquery(), project = "p", dataset = "d")
  expect_equal(as_bq_table(con2, "x"), as_bq_table("p.d.x"))
  expect_equal(as_bq_table(con2, "x.y"), as_bq_table("p.x.y"))
  expect_equal(as_bq_table(con2, "x.y.z"), as_bq_table("x.y.z"))
})

test_that("can create bq_table() from connection + Id", {
  con1 <- DBI::dbConnect(bigquery(), project = "p")
  expect_equal(
    as_bq_table(con1, Id(schema = "x", table = "y")),
    as_bq_table("p.x.y")
  )
  expect_equal(
    as_bq_table(con1, Id(catalog = "q", schema = "x", table = "y")),
    as_bq_table("q.x.y")
  )

  con2 <- DBI::dbConnect(bigquery(), project = "p", dataset = "d")
  expect_equal(as_bq_table(con2, Id(table = "x")), as_bq_table("p.d.x"))
  expect_equal(
    as_bq_table(con2, Id(schema = "x", table = "y")),
    as_bq_table("p.x.y")
  )
  expect_equal(
    as_bq_table(con2, Id(catalog = "q", schema = "x", table = "y")),
    as_bq_table("q.x.y")
  )
})

test_that("as_bq_table checks its input types", {
  con1 <- DBI::dbConnect(bigquery(), project = "p")
  expect_snapshot(as_bq_table(con1, letters), error = TRUE)
})

test_that("the return type of integer columns is set by the bigint argument", {
  x <- c(
    "-2147483648",
    "-2147483647",
    "-1",
    "0",
    "1",
    "2147483647",
    "2147483648"
  )
  sql <- paste0("SELECT * FROM UNNEST ([", paste0(x, collapse = ","), "]) AS x")

  con_integer64 <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project(),
    bigint = "integer64"
  )
  con_character <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project(),
    bigint = "character"
  )

  expect_equal(DBI::dbGetQuery(con_integer64, sql)$x, bit64::as.integer64(x))
  expect_equal(DBI::dbGetQuery(con_character, sql)$x, x)
})

test_that("DBI::sqlData return TRUE, FALSE for logical class", {
  con <- dbConnect(bigquery(), project = bq_test_project())
  expect_equal(
    DBI::dbQuoteLiteral(con, c(TRUE, FALSE, NA)),
    DBI::SQL(c("TRUE", "FALSE", "NULL"))
  )
})

#' Project to use for testing bigrquery
#'
#' @description
#' You'll need to set the `BIGQUERY_TEST_PROJECT` (name of a project) and
#' `BIGQUERY_TEST_BUCKET` (name of bucket) env vars in order to run bigrquery
#' tests locally. I recommend creating a new project because the tests involve
#' both reading and writing in BigQuery and Cloud Storage.
#'
#' The `BIGQUERY_TEST_PROJECT` must have billing enabled for the project. While
#' logged in, via `bq_auth()`, as a user with permission to work in
#' `BIGQUERY_TEST_PROJECT`, run `bq_test_init()` once to perform some setup.
#'
#' @section
#' Testing: In tests, `bq_test_project()` (and hence `bq_test_dataset()`) will
#' automatically skip if auth and a test project are not available.
#'
#' @return `bq_test_project()` returns the name of a project suitable for use in
#'   testing. `bq_test_dataset()` creates a temporary dataset whose lifetime is
#'   tied to the lifetime of the object that it returns.
#' @export
#' @keywords internal
#' @examplesIf bq_testable()
#' ds <- bq_test_dataset()
#' bq_mtcars <- bq_table_upload(bq_table(ds, "mtcars"), mtcars)
#'
#' # dataset and table will be automatically deleted when ds is GC'd
bq_test_project <- function() {
  if (is_testing() && !bq_authable()) {
    testthat::skip("No BigQuery access credentials available")
  }

  env <- Sys.getenv("BIGQUERY_TEST_PROJECT")
  if (!identical(env, "")) {
    return(env)
  }

  if (is_testing()) {
    testthat::skip("BIGQUERY_TEST_PROJECT not set")
  } else {
    cli::cli_abort(
      "{.envvar BIGQUERY_TEST_PROJECT} envvar must to set to a project name."
    )
  }
}

#' @export
#' @rdname bq_test_project
#' @param name Dataset name - used only for testing.
bq_test_init <- function(name = "basedata") {
  proj <- bq_test_project()

  basedata <- bq_dataset(proj, name)
  if (!bq_dataset_exists(basedata)) {
    bq_dataset_create(basedata)
  }

  bq_mtcars <- bq_table(basedata, "mtcars")
  if (!bq_table_exists(bq_mtcars)) {
    job <- bq_table_upload(bq_mtcars, values = datasets::mtcars)
  }
}

#' @export
#' @rdname bq_test_project
bq_test_dataset <- function(name = random_name(), location = "US") {
  ds <- bq_dataset(bq_test_project(), name)
  bq_dataset_create(ds, location = location)

  env <- new.env()
  reg.finalizer(
    env,
    function(e) bq_dataset_delete(ds, delete_contents = TRUE),
    onexit = TRUE
  )
  attr(ds, "env") <- env

  ds
}

bq_test_table <- function() {
  ds <- env_cache(the, "test_dataset", bq_test_dataset())
  bq_table(ds, random_name())
}

#' @export
#' @rdname bq_test_project
bq_testable <- function() {
  bq_authable() && !identical(Sys.getenv("BIGQUERY_TEST_PROJECT"), "")
}

#' @export
#' @rdname bq_test_project
bq_authable <- function() {
  bq_has_token() || (is_interactive() && !is_testing())
}

#' @export
#' @rdname bq_test_project
gs_test_bucket <- function() {
  env <- Sys.getenv("BIGQUERY_TEST_BUCKET")
  if (!identical(env, "")) {
    return(env)
  }

  if (is_testing()) {
    testthat::skip("BIGQUERY_TEST_BUCKET not set")
  } else {
    cli::cli_abort(
      "{.envvar BIGQUERY_TEST_BUCKET} must be set to a bucket name."
    )
  }
}


#' @export
#' @rdname bq_test_project
gs_test_object <- function(name = random_name()) {
  gs_object(gs_test_bucket(), name)
}


random_name <- function(n = 10) {
  paste0("TESTING_", paste(sample(letters, n, replace = TRUE), collapse = ""))
}

is_testing <- function() identical(Sys.getenv("TESTTHAT"), "true")

is_snapshot <- function() identical(Sys.getenv("TESTTHAT_IS_SNAPSHOT"), "true")

skip_if_no_auth <- function() {
  testthat::skip_if_not(bq_has_token(), "Authentication not available")
}

test_that("can control chattiness of bq_job_wait", {
  withr::local_options(cli.progress_show_after = Inf)
  job <- bq_perform_query("SELECT 1 + 1", bq_test_project())

  expect_snapshot({
    bq_job_wait(job, quiet = TRUE)
    bq_job_wait(job, quiet = FALSE)
  })
})

test_that("informative errors on failure", {
  withr::local_options(cli.progress_show_after = Inf)
  ds <- bq_test_dataset()

  tb <- bq_test_table()
  bq_table_create(
    tb,
    fields = list(bq_field("x", "integer"), bq_field("y", "string"))
  )

  expect_snapshot(
    {
      "One error"
      bq_dataset_query(ds, "SELECT 1 +")

      "Multiple errors"
      bq_table_upload(tb, data.frame(x = "x", y = 1:5))
    },
    error = TRUE,
    transform = function(x) gsub("Job (.*?) failed", "Job <bq_job> failed", x)
  )
})

test_that("can retrieve full query results", {
  con <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project(),
    dataset = "basedata",
    bigint = "integer"
  )

  df <- DBI::dbGetQuery(con, "SELECT count(*) as count FROM mtcars")
  expect_equal(df, tibble(count = 32L))
})

test_that("can retrieve without dataset", {
  con <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project(),
    bigint = "integer"
  )
  df <- DBI::dbGetQuery(con, "SELECT count(*) as count FROM `basedata.mtcars`")
  expect_equal(df, tibble(count = 32L))
})

test_that("can retrieve query in pieces and that quiet is respected", {
  con <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project(),
    dataset = "basedata"
  )

  res <- DBI::dbSendQuery(con, "SELECT cyl, mpg FROM mtcars")
  expect_equal(DBI::dbGetRowCount(res), 0L)

  expect_snapshot(DBI::dbFetch(res, NA), error = TRUE)

  res@quiet <- FALSE
  expect_snapshot(df <- DBI::dbFetch(res, 10))

  expect_equal(nrow(df), 10)
  expect_false(DBI::dbHasCompleted(res))
  expect_equal(DBI::dbGetRowCount(res), 10L)

  res@quiet <- TRUE
  expect_snapshot(df <- DBI::dbFetch(res, -1))
  expect_equal(nrow(df), 22)
  expect_true(DBI::dbHasCompleted(res))
})

test_that("dbHasCompleted() is accurate if you fetch past end", {
  con <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project(),
    dataset = "basedata"
  )

  res <- DBI::dbSendQuery(con, "SELECT cyl FROM mtcars LIMIT 5")
  df <- DBI::dbFetch(res, 10)

  expect_equal(DBI::dbGetRowCount(res), 5)
  expect_true(DBI::dbHasCompleted(res))
})

test_that("can get metadata", {
  con <- DBI::dbConnect(
    bigquery(),
    project = bq_test_project(),
    dataset = "basedata"
  )
  sql <- "SELECT cyl, mpg FROM mtcars"
  res <- DBI::dbSendQuery(con, sql)
  expect_snapshot(res)

  col_info <- DBI::dbColumnInfo(res)
  expect_equal(dim(col_info), c(2, 2))
  expect_equal(col_info$name, c("cyl", "mpg"))

  expect_equal(DBI::dbGetStatement(res), sql)
})

test_that("dbExecute returns modified rows", {
  ds <- bq_test_dataset()
  con <- DBI::dbConnect(ds)

  DBI::dbExecute(con, "CREATE TABLE foo (a INT64)")
  expect_equal(DBI::dbExecute(con, "INSERT INTO foo VALUES (1), (2), (3)"), 3)
  expect_equal(DBI::dbExecute(con, "DELETE FROM foo WHERE a >= 2"), 2)
  DBI::dbExecute(con, "DROP TABLE foo")
})

#' Explicitly define query parameters
#'
#' By default, bigrquery will assume vectors of length 1 are scalars,
#' and longer vectors are arrays. If you need to pass a length-1 array,
#' you'll need to explicitly use `bq_param_array()`.
#'
#' @param value vector of parameter values
#' @param name name of the parameter in the query, omitting the `@`
#' @param type BigQuery type of the parameter
#' @keywords internal
#' @export
#' @examples
#' # bq_param() automatically picks scalar vs array based on length
#' bq_param("a")
#' bq_param(c("a", "b", "c"))
#'
#' # use bq_param_array() to create a length-1 array
#' bq_param_array("a")
bq_param <- function(value, type = NULL, name = NULL) {
  if (length(value) > 1) {
    bq_param_array(value, type, name)
  } else {
    bq_param_scalar(value, type, name)
  }
}

#' @rdname bq_param
#' @export
bq_param_scalar <- function(value, type = NULL, name = NULL) {
  if (length(value) != 1) {
    cli::cli_abort("{.arg value} must be length 1, not {length(value)}.")
  }

  if (is.null(type)) {
    type <- data_type(value)
  }
  structure(
    list(value = value, name = name, type = type),
    class = c("bq_param_scalar", "bq_param")
  )
}

#' @rdname bq_param
#' @export
bq_param_array <- function(value, type = NULL, name = NULL) {
  if (length(value) == 0) {
    cli::cli_abort("{.arg value} can't be zero-length.")
  }

  if (is.null(type)) {
    type <- data_type(value)
  }
  structure(
    list(value = value, name = name, type = type),
    class = c("bq_param_array", "bq_param")
  )
}

as_bq_param <- function(x, name) {
  if (inherits(x, "bq_param")) {
    if (!is.null(name) & is.null(x$name)) {
      x$name <- name
    }
    x
  } else {
    bq_param(name = name, x)
  }
}

bq_params <- function(x) {
  structure(x, class = "bq_params")
}

as_bq_params <- function(x) {
  params <- lapply(names(x), function(name) {
    as_bq_param(x[[name]], name)
  })
  bq_params(params)
}

#' @export
as_json.bq_params <- function(x) {
  json <- lapply(x, as_json)
  unname(json)
}


#' @export
as_json.bq_param_scalar <- function(x) {
  list(
    name = x$name,
    parameterType = list(type = unbox(x$type)),
    parameterValue = list(value = unbox(x$value))
  )
}

#' @export
as_json.bq_param_array <- function(x) {
  values <- unname(c(x$value))
  values <- lapply(values, function(x) list(value = unbox(x)))
  list(
    name = x$name,
    parameterType = list(
      type = "ARRAY",
      arrayType = list(type = unbox(x$type))
    ),
    parameterValue = list(arrayValues = values)
  )
}

#' @export
print.bq_param <- function(x, ...) {
  cat(show_json(as_json(x)))
  invisible(x)
}

#' @export
print.bq_params <- function(x, ...) {
  cat(show_json(as_json(x)))
  invisible(x)
}

#' BigQuery field (and fields) class
#'
#' `bq_field()` and `bq_fields()` create; `as_bq_field()` and `as_bq_fields()`
#' coerce from lists.
#'
#' @param name The field name. The name must contain only letters (a-z, A-Z),
#'   numbers (0-9), or underscores (_), and must start with a letter or
#'   underscore. The maximum length is 300 characters.
#' @param type The field data type. Possible values include:
#'   `"STRING"`, `"BYTES"`, `"INTEGER"`, `"FLOAT"`, `"BOOLEAN"`, `"TIMESTAMP"`,
#'   `"DATE"`, `"TIME"`, `"DATETIME"`, `"GEOGRAPHY"`, `"NUMERIC"`,
#'   `"BIGNUMERIC"`, `"JSON"`, `"RECORD"`.
#' @param mode The field mode. Possible values include: `"NULLABLE"`,
#'   `"REQUIRED"`, and `"REPEATED"`.
#' @param fields For a field of type "record", a list of sub-fields.
#' @param description The field description. The maximum length is 1,024
#'   characters.
#' @param x A list of `bg_fields`
#' @seealso `bq_field()` corresponds to a `TableFieldSchema`, see
#' <https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables#TableFieldSchema>
#' for more details.
#' @export
#' @examples
#' bq_field("name", "string")
#'
#' as_bq_fields(list(
#'   list(name = "name", type = "string"),
#'   bq_field("age", "integer")
#' ))
#'
#' # as_bq_fields() can also take a data frame
#' as_bq_fields(mtcars)
bq_field <- function(
  name,
  type,
  mode = "NULLABLE",
  fields = list(),
  description = NULL
) {
  check_string(name)
  check_string(type)
  check_string(mode)
  check_string(description, allow_null = TRUE)

  structure(
    list(
      name = name,
      type = toupper(type),
      mode = toupper(mode),
      fields = as_bq_fields(fields),
      description = description %||% ""
    ),
    class = "bq_field"
  )
}

#' @export
as_json.bq_field <- function(x) {
  list(
    name = unbox(x$name),
    type = unbox(x$type),
    mode = unbox(x$mode),
    fields = as_json(x$fields),
    description = unbox(x$description)
  )
}

#' @export
#' @rdname bq_field
bq_fields <- function(x) {
  structure(x, class = "bq_fields")
}

#' @export
as_json.bq_fields <- function(x) {
  lapply(x, as_json)
}

#' @export
#' @rdname bq_field
as_bq_field <- function(x) UseMethod("as_bq_field")

#' @export
as_bq_field.bq_field <- function(x) x

#' @export
as_bq_field.list <- function(x) {
  bq_field(
    name = x$name,
    type = x$type,
    mode = x$mode %||% "NULLABLE",
    fields = lapply(x$fields, as_bq_field),
    description = x$description %||% ""
  )
}

#' @export
#' @rdname bq_field
as_bq_fields <- function(x) UseMethod("as_bq_fields")

#' @export
as_bq_fields.NULL <- function(x) x

#' @export
as_bq_fields.bq_fields <- function(x) x

#' @export
as_bq_fields.data.frame <- function(x) {
  types <- vapply(x, data_type, character(1))
  fields <- Map(function(type, name) bq_field(name, type), types, names(x))
  bq_fields(unname(fields))
}

#' @export
as_bq_fields.list <- function(x) {
  bq_fields(lapply(x, as_bq_field))
}

#' @export
format.bq_fields <- function(x, ...) {
  if (length(x) == 0) {
    return("")
  }

  fields <- lapply(x, format)
  gsub("\\n\\s+$", "\n", indent(paste0(fields, collapse = "")))
}

#' @export
print.bq_fields <- function(x, ...) {
  cat_line("<bq_fields>\n", format(x, ...))
  invisible(x)
}

#' @export
format.bq_field <- function(x, ...) {
  type <- x$type
  if (x$mode != "NULLABLE") {
    type <- paste0(type, ": ", tolower(x$mode))
  }

  paste0(x$name, " <", type, ">", "\n", format(x$fields))
}

#' @export
print.bq_field <- function(x, ...) {
  cat_line("<bq_field> ", format(x, ...))
  invisible(x)
}

data_type <- function(x) {
  if (is.factor(x)) {
    return("STRING")
  }
  if (inherits(x, "POSIXt")) {
    return("TIMESTAMP")
  }
  if (inherits(x, "hms")) {
    return("TIME")
  }
  if (inherits(x, "wk_wkt")) {
    return("GEOGRAPHY")
  }
  if (inherits(x, "blob")) {
    return("BYTES")
  }
  if (inherits(x, "Date")) {
    return("DATE")
  }

  switch(
    typeof(x),
    character = "STRING",
    logical = "BOOLEAN",
    double = "FLOAT",
    integer = "INTEGER",
    cli::cli_abort("Unsupported type {.str {typeof(x)}}.")
  )
}

#' @section Package options:
#' \describe{
#'   \item{`bigrquery.quiet`}{Verbose output during processing? The default
#'   value, `NA`, turns on verbose output for queries that run longer than
#'   two seconds.  Use `FALSE` for immediate verbose output, `TRUE`
#'   for quiet operation.}
#'
#'   \item{`bigrquery.page.size`}{Default page size for fetching data,
#'   defaults to 1e4.}
#' }
#'
#' @keywords internal
#' @useDynLib bigrquery, .registration = TRUE
#' @rawNamespace import(rlang, except = unbox)
"_PACKAGE"

the <- new_environment()

# We import rlang this way because jsonlite's unbox() is used extensively.

## usethis namespace: start
#' @importFrom bit64 integer64
#' @importFrom jsonlite unbox
#' @importFrom lifecycle deprecated
#' @importFrom tibble tibble
## usethis namespace: end
NULL

#' Download table data
#'
#' @description
#' This function provides two ways to download data from BigQuery, transfering
#' data using either JSON or arrow, depending on the `api` argument. If
#' bigrquerystorage is installed, `api = "arrow"` will be used (because it's
#' so much faster, but see the limitions below), otherwise you can select
#' deliberately by using `api = "json"` or `api = "arrow"`.
#'
#' ## Arrow API
#'
#' The arrow API is much faster, but has heavier dependencies: bigrquerystorage
#' requires the arrow package, which can be tricky to compile on Linux (but you
#' usually should be able to get a binary from
#' [Posit Public Package Manager](https://p3m.dev/).
#'
#' There's one known limitation of `api = "arrow"`: when querying public data,
#' you'll now need to provide a `billing` project.
#'
#' ## JSON API
#'
#' The JSON API retrieves rows in chunks of `page_size`. It is most suitable
#' for results of smaller queries (<100 MB, say). Unfortunately due to
#' limitations in the BigQuery API, you may need to vary this parameter
#' depending on the complexity of the underlying data.
#'
#' The JSON API will convert nested and repeated columns in to list-columns
#' as follows:
#'
#' * Repeated values (arrays) will become a list-column of vectors.
#' * Records will become list-columns of named lists.
#' * Repeated records will become list-columns of data frames.
#'
#' @return Because data retrieval may generate list-columns and the `data.frame`
#'   print method can have problems with list-columns, this method returns
#'   a tibble. If you need a `data.frame`, coerce the results with
#'   [as.data.frame()].
#' @param x A [bq_table]
#' @param n_max Maximum number of results to retrieve. Use `Inf` to retrieve all
#'   rows.
#' @param page_size (JSON only) The number of rows requested per chunk. It is
#'   recommended to leave this unspecified until you have evidence that the
#'  `page_size` selected automatically by `bq_table_download()` is problematic.
#'
#'   When `page_size = NULL` bigrquery determines a conservative, natural chunk
#'   size empirically. If you specify the `page_size`, it is important that each
#'   chunk fits on one page, i.e. that the requested row limit is low enough to
#'   prevent the API from paginating based on response size.
#' @param start_index (JSON only) Starting row index (zero-based).
#' @param max_connections (JSON only) Number of maximum simultaneous
#'   connections to BigQuery servers.
#' @param api Which API to use? The `"json"` API works where ever bigrquery
#'   does, but is slow and can require fiddling with the `page_size` parameter.
#'   The `"arrow"` API is faster and more reliable, but only works if you
#'   have also installed the bigrquerystorage package.
#'
#'   Because the `"arrow"` API is so much faster, it will be used automatically
#'   if the bigrquerystorage package is installed.
#' @inheritParams api-job
#' @param bigint The R type that BigQuery's 64-bit integer types should be
#'   mapped to. The default is `"integer"`, which returns R's `integer` type,
#'   but results in `NA` for values above/below +/- 2147483647. `"integer64"`
#'   returns a [bit64::integer64], which allows the full range of 64 bit
#'   integers.
#' @param billing (Arrow only) Project to bill; defaults to the project of `x`,
#'   and typically only needs to be specified if you're working with public
#'   datasets.
#' @section Google BigQuery API documentation:
#' * [list](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tabledata/list)
#' @export
#' @examplesIf bq_testable()
#' df <- bq_table_download("publicdata.samples.natality", n_max = 35000, billing = bq_test_project())
bq_table_download <-
  function(
    x,
    n_max = Inf,
    page_size = NULL,
    start_index = 0L,
    max_connections = 6L,
    quiet = getOption("bigrquery.quiet", NA),
    bigint = c("integer", "integer64", "numeric", "character"),
    api = c("json", "arrow"),
    billing = x$project
  ) {
    x <- as_bq_table(x)
    check_number_whole(n_max, min = 0, allow_infinite = TRUE)
    check_number_whole(start_index, min = 0)
    check_number_whole(max_connections, min = 1)
    quiet <- check_quiet(quiet)
    bigint <- arg_match(bigint)
    api <- check_api(api)

    if (api == "arrow") {
      check_installed(
        "bigrquerystorage",
        "required to download using arrow API"
      )
      if (!missing(page_size)) {
        cli::cli_warn(
          '{.arg page_size} is ignored when {.code api == "arrow"}',
          call = environment()
        )
      }
      if (!missing(start_index)) {
        cli::cli_warn(
          '{.arg start_index} is ignored when {.code api == "arrow"}',
          call = environment()
        )
      }
      if (!missing(max_connections)) {
        cli::cli_warn(
          '{.arg max_connections} is ignored when {.code api == "arrow"}',
          call = environment()
        )
      }

      return(bigrquerystorage::bqs_table_download(
        x = toString(x),
        parent = billing,
        n_max = n_max,
        quiet = quiet,
        bigint = bigint,
        as_tibble = TRUE
      ))
    }

    params <- set_row_params(
      nrow = bq_table_nrow(x),
      n_max = n_max,
      start_index = start_index
    )
    n_max <- params$n_max
    start_index <- params$start_index

    schema_path <- bq_download_schema(x, tempfile())
    defer(unlink(schema_path))

    if (n_max == 0) {
      table_data <- bq_parse_files(
        schema_path,
        file_paths = character(),
        n = 0,
        quiet = quiet
      )
      return(table_data)
    }

    pool <- curl::new_pool()

    # get first chunk ----
    if (!quiet) {
      cli::cli_inform("Downloading first chunk of data.")
    }

    check_number_whole(page_size, min = 0, allow_null = TRUE)
    chunk_size_from_user <- !is.null(page_size)
    chunk_size <- page_size

    chunk_plan <- bq_download_plan(
      n_max,
      chunk_size = chunk_size,
      n_chunks = 1,
      start_index = start_index
    )
    handle <- bq_download_chunk_handle(
      x,
      begin = chunk_plan$dat$chunk_begin[1],
      max_results = chunk_plan$dat$chunk_rows[1]
    )
    curl::multi_add(
      handle,
      done = bq_download_callback(chunk_plan$dat$path[1]),
      pool = pool
    )
    curl::multi_run(pool = pool)
    path_first_chunk <- chunk_plan$dat$path[1]
    defer(unlink(path_first_chunk))

    chunk_data <- bq_parse_file(schema_path, path_first_chunk)
    n_got <- nrow(chunk_data)

    if (n_got >= n_max) {
      if (!quiet) {
        cli::cli_inform("First chunk includes all requested rows.")
      }
      return(parse_postprocess(chunk_data, bigint = bigint))
    }

    if (chunk_size_from_user && n_got < chunk_size) {
      cli::cli_abort(c(
        "First chunk is incomplete:",
        x = "{big_mark(chunk_size)} rows were requested, but only {big_mark(n_got)} rows were received.",
        i = "Leave `page_size` unspecified or use an even smaller value."
      ))
    }

    # break rest of work into natural chunks ----
    if (!chunk_size_from_user) {
      if (!quiet) {
        cli::cli_inform("Received {big_mark(n_got)} rows in the first chunk.")
      }
      chunk_size <- trunc(0.75 * n_got)
    }

    n_max_new <- n_max - n_got
    start_index_new <- n_got

    chunk_plan <- bq_download_plan(
      n_max_new,
      chunk_size = chunk_size,
      start_index = start_index_new
    )

    if (!quiet) {
      cli::cli_inform(
        "Downloading the remaining {big_mark(chunk_plan$n_max)} rows in {chunk_plan$n_chunks} \\
         chunks of (up to) {big_mark(chunk_plan$chunk_size)} rows."
      )
      progress <- cli::cli_progress_bar(
        format = "Downloading data {cli::pb_bar} {cli::pb_percent} ETA {cli::pb_eta}",
        total = chunk_plan$n_chunks
      )
    } else {
      progress <- NULL
    }

    for (i in seq_len(chunk_plan$n_chunks)) {
      handle <- bq_download_chunk_handle(
        x,
        begin = chunk_plan$dat$chunk_begin[i],
        max_results = chunk_plan$dat$chunk_rows[i]
      )
      curl::multi_add(
        handle,
        done = bq_download_callback(chunk_plan$dat$path[i], progress),
        pool = pool
      )
    }
    curl::multi_run(pool = pool)
    defer(unlink(chunk_plan$dat$path))

    table_data <- bq_parse_files(
      schema_path,
      c(path_first_chunk, chunk_plan$dat$path),
      n = n_max,
      quiet = quiet
    )
    parse_postprocess(table_data, bigint = bigint)
  }

check_api <- function(api = c("json", "arrow"), error_call = caller_env()) {
  if (identical(api, c("json", "arrow"))) {
    if (has_bigrquerystorage()) "arrow" else "json"
  } else {
    arg_match(api, error_call = error_call)
  }
}

# This function is a modified version of
# https://github.com/r-dbi/RPostgres/blob/master/R/PqResult.R
parse_postprocess <- function(df, bigint) {
  df <- col_apply(
    df,
    function(x) identical(attr(x, "bq_type"), "DATE"),
    function(x) clock::date_parse(x)
  )
  df <- col_apply(
    df,
    function(x) identical(attr(x, "bq_type"), "DATETIME"),
    function(x) bq_datetime_parse(x)
  )
  df <- col_apply(
    df,
    function(x) identical(attr(x, "bq_type"), "TIME"),
    function(x) hms::parse_hms(x)
  )

  if (bigint != "integer64") {
    as_bigint <- switch(
      bigint,
      integer = as.integer,
      numeric = as.numeric,
      character = as.character
    )
    df <- col_apply(df, bit64::is.integer64, as_bigint)
  }

  df
}

col_apply <- function(x, p, f) {
  if (is.list(x)) {
    x[] <- lapply(x, col_apply, p = p, f = f)
    x
  } else if (p(x)) {
    f(x)
  } else {
    x
  }
}

bq_datetime_parse <- function(x) {
  # `format` matches clock already.
  # Bigquery DATETIME is documented as microsecond precision.
  # https://docs.cloud.google.com/bigquery/docs/reference/standard-sql/data-types#datetime_type
  x <- clock::year_month_day_parse(x, precision = "microsecond")

  # Manually retain microseconds for the POSIXct
  microseconds <- clock::get_microsecond(x)
  microseconds <- microseconds / 1000000

  # Convert to POSIXct at second precision since that is what clock asserts
  # the precision of a POSIXct is. Can use sys-time since we are going straight
  # to UTC.
  x <- clock::calendar_narrow(x, "second")
  x <- clock::as_sys_time(x)
  x <- as.POSIXct(x, tz = "UTC")

  # Manually add microseconds back on (lossy, floating point issues!)
  x <- x + microseconds

  x
}

set_row_params <- function(nrow, n_max = Inf, start_index = 0L) {
  check_number_whole(n_max, min = 0, allow_infinite = TRUE)
  check_number_whole(start_index, min = 0)

  n_max <- max(min(n_max, nrow - start_index), 0)

  list(n_max = n_max, start_index = start_index)
}

bq_download_plan <- function(
  n_max,
  chunk_size = NULL,
  n_chunks = NULL,
  start_index = 0
) {
  params <- set_chunk_params(n_max, chunk_size, n_chunks)
  list(
    n_max = n_max,
    chunk_size = params$chunk_size,
    n_chunks = params$n_chunks,
    dat = set_chunk_plan(
      n_max,
      params$chunk_size,
      params$n_chunks,
      start_index
    )
  )
}

set_chunk_params <- function(n_max, chunk_size = NULL, n_chunks = NULL) {
  if (is.null(chunk_size) && is.null(n_chunks)) {
    n_chunks <- 1
  }
  n_chunks <- n_chunks %||% Inf
  chunk_size <- pmin(chunk_size %||% ceiling(n_max / n_chunks), n_max)
  n_chunks <- pmin(n_chunks, ceiling(n_max / chunk_size))
  list(chunk_size = chunk_size, n_chunks = n_chunks)
}

set_chunk_plan <- function(n_max, chunk_size, n_chunks, start_index = 0) {
  chunk_begin <- start_index + (seq_len(n_chunks) - 1) * chunk_size
  chunk_end <- pmin(chunk_begin + chunk_size, start_index + n_max)
  chunk_rows <- chunk_end - chunk_begin
  tibble::tibble(
    chunk_begin,
    chunk_rows,
    path = sort(
      tempfile(
        rep_len("bq-download-", length.out = n_chunks),
        fileext = ".json"
      )
    )
  )
}

bq_download_chunk_handle <- function(x, begin = 0L, max_results = 1e4) {
  x <- as_bq_table(x)
  check_number_whole(begin, min = 0)
  check_number_whole(max_results, min = 1, allow_infinite = TRUE)

  # Pre-format query params with forced non-scientific notation, since the BQ
  # API doesn't accept numbers like 1e5. See issue #395 for details.
  query <- list(
    startIndex = format(begin, scientific = FALSE),
    maxResults = format(max_results, scientific = FALSE)
  )

  url <- paste0(
    base_url,
    bq_path(x$project, dataset = x$dataset, table = x$table, data = "")
  )
  url <- httr::modify_url(url, query = prepare_bq_query(query))

  if (bq_has_token()) {
    token <- .auth$get_cred()
    signed <- token$sign("GET", url)
    url <- signed$url
    headers <- signed$headers
  } else {
    headers <- list()
  }

  h <- curl::new_handle(url = url)
  curl::handle_setopt(h, useragent = bq_ua())
  curl::handle_setheaders(h, .list = headers)

  h
}

bq_download_callback <- function(path, progress = NULL, call = caller_env()) {
  force(path)
  force(progress)

  function(result) {
    if (!is.null(progress)) {
      cli::cli_progress_update(id = progress)
    }

    bq_check_response(
      status = result$status_code,
      type = curl::parse_headers_list(result$headers)[["content-type"]],
      content = result$content,
      call = call
    )

    con <- file(path, open = "wb")
    defer(close(con))
    writeBin(result$content, con)
  }
}

# Helpers for testing -----------------------------------------------------

bq_parse_file <- function(fields, data) {
  fields <- brio::read_file(fields)
  data <- brio::read_file(data)

  bq_parse(fields, data)
}

bq_download_schema <- function(x, path) {
  x <- as_bq_table(x)

  url <- bq_path(x$project, x$dataset, x$table)
  query <- list(fields = "schema")

  json <- bq_get(url, query = query, raw = TRUE)
  writeBin(json, path)

  path
}

test_that("same results regardless of page size", {
  skip_if_no_auth()

  tb <- as_bq_table("bigquery-public-data.moon_phases.moon_phases")

  df3 <- bq_table_download(tb, n_max = 30, page_size = 10, api = "json")
  df1 <- bq_table_download(tb, n_max = 30, page_size = 30, api = "json")
  expect_equal(nrow(df1), 30)
  expect_equal(df1, df3)
})

test_that("can retrieve fraction of page size", {
  skip_if_no_auth()

  tb <- as_bq_table("bigquery-public-data.moon_phases.moon_phases")
  df <- bq_table_download(tb, n_max = 15, page_size = 10, api = "json")
  expect_equal(nrow(df), 15)
})

test_that("can retrieve zero rows", {
  skip_if_no_auth()

  tb <- as_bq_table("bigquery-public-data.moon_phases.moon_phases")
  df <- bq_table_download(tb, n_max = 0, api = "json")
  expect_equal(nrow(df), 0)
  expect_named(df, c("phase", "phase_emoji", "peak_datetime"))
})

test_that("can specify large integers in page params", {
  skip_if_no_auth()

  # Use scipen to nudge R to prefer scientific formatting for very small numbers
  # to allow this test to exercise issue #395 with small datasets.
  withr::local_options(list(scipen = -4))

  tb <- as_bq_table("bigquery-public-data.moon_phases.moon_phases")
  df <- bq_table_download(tb, n_max = 100, page_size = 20, api = "json")
  expect_equal(nrow(df), 100)
})

test_that("errors when table is known to be incomplete", {
  # If this snapshot changes, it's likely because the table has been updated

  skip_if_no_auth()

  tb <- as_bq_table("bigquery-public-data.chicago_taxi_trips.taxi_trips")
  expect_snapshot(
    bq_table_download(
      tb,
      n_max = 35000,
      page_size = 35000,
      bigint = "integer64",
      api = "json"
    ),
    transform = function(x) {
      gsub(
        "[0-9,]+ rows were received",
        "{n} rows were received",
        x,
        perl = TRUE
      )
    },
    error = TRUE
  )
})

# api = "arrow" ----------------------------------------------------------------

test_that("check_api respects inputs", {
  expect_equal(check_api("arrow"), "arrow")
  expect_equal(check_api("json"), "json")
})

test_that("uses arrow api if bigrquerystorage installed", {
  local_mocked_bindings(is_installed = function(...) TRUE)
  expect_equal(check_api(), "arrow")

  local_mocked_bindings(is_installed = function(...) FALSE)
  expect_equal(check_api(), "json")
})

test_that("warns if supplying unnused arguments", {
  tb <- bq_project_query(bq_test_project(), "SELECT 1.0", quiet = TRUE)
  expect_snapshot(
    . <- bq_table_download(
      tb,
      api = "arrow",
      page_size = 1,
      start_index = 1,
      max_connections = 1,
      quiet = TRUE
    )
  )
})

test_that("arrow api can convert non-nested types", {
  sql <- "SELECT
    '\U0001f603' as unicode,
    datetime,
    TRUE as logicaltrue,
    FALSE as logicalfalse,
    CAST ('Hi' as BYTES) as bytes,
    CAST (datetime as DATE) as date,
    CAST (datetime as TIME) as time,
    CAST (datetime as TIMESTAMP) as timestamp,
    ST_GEOGFROMTEXT('POINT (30 10)') as geography
    FROM (SELECT DATETIME '2000-01-02 03:04:05.67' as datetime)
  "

  tb <- bq_project_query(bq_test_project(), sql, quiet = TRUE)
  df <- bq_table_download(tb, api = "arrow", quiet = TRUE)

  base <- ISOdatetime(2000, 1, 2, 3, 4, 5.67, tz = "UTC")
  expect_identical(df$unicode, "\U0001f603", ignore_encoding = FALSE)

  expect_equal(df$logicaltrue, TRUE)
  expect_equal(df$logicalfalse, FALSE)

  expect_equal(df$bytes, blob::as_blob(as.raw(c(0x48, 0x69))))

  expect_equal(df$date, as.Date(base))
  expect_equal(df$timestamp, base)
  expect_equal(df$datetime, base)
  expect_equal(df$time, hms::hms(hours = 3, minutes = 4, seconds = 5.67))

  expect_identical(df$geography, wk::wkt("POINT(30 10)"))
})

test_that("arrow api can convert nested types", {
  sql <- "SELECT
    STRUCT(1.0 AS a, 'abc' AS b) as s,
    [1.0, 2.0, 3.0] as a,
    [STRUCT(1.0 as a, 'a' as b), STRUCT(2.0, 'b'), STRUCT(3, 'c')] as aos,
    STRUCT([1.0, 2.0, 3.0] as a, ['a', 'b'] as b) as soa
  "

  tb <- bq_project_query(bq_test_project(), sql, quiet = TRUE)
  df <- bq_table_download(tb, api = "arrow", quiet = TRUE)

  expect_equal(df$s, tibble(a = 1, b = "abc"))
  expect_equal(df$a, list(c(1, 2, 3)))
  expect_equal(df$aos, list(tibble(a = c(1, 2, 3), b = c("a", "b", "c"))))
  expect_equal(df$soa, tibble(a = list(c(1, 2, 3)), b = list(c("a", "b"))))
})

test_that("arrow api respects bigint", {
  x <- c(
    "-2147483648",
    "-2147483647",
    "-1",
    "0",
    "1",
    "2147483647",
    "2147483648"
  )
  sql <- paste0("SELECT * FROM UNNEST ([", paste0(x, collapse = ","), "]) AS x")
  qry <- bq_project_query(bq_test_project(), sql)

  out_int64 <- bq_table_download(
    qry,
    bigint = "integer64",
    api = "arrow",
    quiet = TRUE
  )$x
  expect_identical(out_int64, bit64::as.integer64(x))

  out_dbl <- bq_table_download(
    qry,
    bigint = "numeric",
    api = "arrow",
    quiet = TRUE
  )$x
  expect_identical(out_dbl, as.double(x))

  out_chr <- bq_table_download(
    qry,
    bigint = "character",
    api = "arrow",
    quiet = TRUE
  )$x
  expect_identical(out_chr, x)
})

# helpers around row and chunk params ------------------------------------------

test_that("set_row_params() works ", {
  # n_max > nrow
  expect_equal(
    set_row_params(10, n_max = 15),
    list(n_max = 10, start_index = 0)
  )
  # start_index > nrow
  expect_equal(
    set_row_params(10, start_index = 12),
    list(n_max = 0, start_index = 12)
  )
  expect_equal(
    set_row_params(10, n_max = 5, start_index = 12),
    list(n_max = 0, start_index = 12)
  )
  # n_max > nrow - start_index
  expect_equal(
    set_row_params(10, n_max = 5, start_index = 7),
    list(n_max = 3, start_index = 7)
  )
})

test_that("set_chunk_params() works", {
  # no chunk_size, no n_chunks
  expect_equal(set_chunk_params(5), list(chunk_size = 5, n_chunks = 1))

  # yes chunk_size, no n_chunks
  expect_equal(
    set_chunk_params(10, chunk_size = 3),
    list(chunk_size = 3, n_chunks = 4)
  )
  expect_equal(
    set_chunk_params(10, chunk_size = 10),
    list(chunk_size = 10, n_chunks = 1)
  )
  expect_equal(
    set_chunk_params(10, chunk_size = 9),
    list(chunk_size = 9, n_chunks = 2)
  )
  expect_equal(
    set_chunk_params(10, chunk_size = 15),
    list(chunk_size = 10, n_chunks = 1)
  )

  # no chunk_size, yes n_chunks
  expect_equal(
    set_chunk_params(10, n_chunks = 1),
    list(chunk_size = 10, n_chunks = 1)
  )
  expect_equal(
    set_chunk_params(10, n_chunks = 3),
    list(chunk_size = 4, n_chunks = 3)
  )
  expect_equal(
    set_chunk_params(10, n_chunks = 10),
    list(chunk_size = 1, n_chunks = 10)
  )
  expect_equal(
    set_chunk_params(10, n_chunks = 11),
    list(chunk_size = 1, n_chunks = 10)
  )

  # yes chunk_size, yes n_chunks
  expect_equal(
    set_chunk_params(10, chunk_size = 3, n_chunks = 1),
    list(chunk_size = 3, n_chunks = 1)
  )
  expect_equal(
    set_chunk_params(10, chunk_size = 7, n_chunks = 5),
    list(chunk_size = 7, n_chunks = 2)
  )
})

test_that("set_chunk_plan() works", {
  dat <- set_chunk_plan(5, chunk_size = 5, n_chunks = 1)
  expect_equal(dat$chunk_begin, 0)
  expect_equal(dat$chunk_rows, 5)

  dat <- set_chunk_plan(5, chunk_size = 5, n_chunks = 1, start_index = 3)
  expect_equal(dat$chunk_begin, 3)
  expect_equal(dat$chunk_rows, 5)

  dat <- set_chunk_plan(10, chunk_size = 3, n_chunks = 4)
  expect_equal(dat$chunk_begin, c(0, 3, 6, 9))
  expect_equal(dat$chunk_rows, c(3, 3, 3, 1))

  dat <- set_chunk_plan(10, chunk_size = 3, n_chunks = 4, start_index = 8)
  expect_equal(dat$chunk_begin, 8 + c(0, 3, 6, 9))
  expect_equal(dat$chunk_rows, c(3, 3, 3, 1))
})

# types -------------------------------------------------------------------

test_that("can read utf-8 strings", {
  sql <- "SELECT '\U0001f603' as x"
  tb <- bq_project_query(bq_test_project(), sql)
  df <- bq_table_download(tb)
  x <- df$x[[1]]

  expect_equal(Encoding(x), "UTF-8")
  expect_equal(x, "\U0001f603")
})

test_that("can convert date time types", {
  sql <- "SELECT
    datetime,
    CAST (datetime as DATE) as date,
    CAST (datetime as TIME) as time,
    CAST (datetime as TIMESTAMP) as timestamp
    FROM (SELECT DATETIME '2000-01-02 03:04:05.67' as datetime)
  "

  tb <- bq_project_query(bq_test_project(), sql, quiet = TRUE)
  df <- bq_table_download(tb, api = "json")

  base <- ISOdatetime(2000, 1, 2, 3, 4, 5.67, tz = "UTC")

  expect_equal(df$datetime, base)
  expect_equal(df$timestamp, base)
  expect_equal(df$date, as.Date(base))
  expect_equal(df$time, hms::hms(hours = 3, minutes = 4, seconds = 5.67))
})


test_that("can parse fractional seconds", {
  x <- c(
    "2000-01-02T03:04:05.67",
    "2000-01-02T03:04:05",
    "2000-01-02T03:04:05.123456"
  )
  parsed <- as.numeric(bq_datetime_parse(x))
  expect_equal(parsed - trunc(parsed), c(0.67, 0, 0.123456), tolerance = 1e-6)
})

test_that("correctly parse logical values", {
  query <- "SELECT TRUE as x"
  tb <- bq_project_query(bq_test_project(), query)
  df <- bq_table_download(tb, api = "json")

  expect_true(df$x)
})

test_that("the return type of integer columns is set by the bigint argument", {
  x <- c(
    "-2147483648",
    "-2147483647",
    "-1",
    "0",
    "1",
    "2147483647",
    "2147483648"
  )
  sql <- paste0("SELECT * FROM UNNEST ([", paste0(x, collapse = ","), "]) AS x")
  qry <- bq_project_query(bq_test_project(), sql)

  expect_warning(
    out_int <- bq_table_download(qry, bigint = "integer", api = "json")$x,
    "integer overflow"
  )
  expect_identical(out_int, suppressWarnings(as.integer(x)))

  out_int64 <- bq_table_download(qry, bigint = "integer64", api = "json")$x
  expect_identical(out_int64, bit64::as.integer64(x))

  out_dbl <- bq_table_download(qry, bigint = "numeric", api = "json")$x
  expect_identical(out_dbl, as.double(x))

  out_chr <- bq_table_download(qry, bigint = "character", api = "json")$x
  expect_identical(out_chr, x)
})

test_that("can convert geography type", {
  skip_if_not_installed("wk")
  sql <- "SELECT ST_GEOGFROMTEXT('POINT (30 10)') as geography"
  tb <- bq_project_query(bq_test_project(), sql, quiet = TRUE)
  df <- bq_table_download(tb, api = "json")

  expect_identical(df$geography, wk::wkt("POINT(30 10)"))
})

test_that("can convert bytes type", {
  sql <- "SELECT ST_ASBINARY(ST_GEOGFROMTEXT('POINT (30 10)')) as bytes"
  tb <- bq_project_query(bq_test_project(), sql, quiet = TRUE)
  df <- bq_table_download(tb, api = "json")

  expect_identical(
    df$bytes,
    blob::blob(
      as.raw(c(
        0x01,
        0x01,
        0x00,
        0x00,
        0x00,
        0xff,
        0xff,
        0xff,
        0xff,
        0xff,
        0xff,
        0x3d,
        0x40,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x00,
        0x24,
        0x40
      ))
    )
  )
})

# Individual values -------------------------------------------------------

test_that("can parse atomic vectors", {
  expect_identical(bq_parse_single("x", "string"), "x")
  expect_identical(bq_parse_single("10", "integer"), bit64::as.integer64(10))
  expect_identical(bq_parse_single("10", "float"), 10)
  expect_identical(bq_parse_single("true", "boolean"), TRUE)
  expect_identical(bq_parse_single("false", "boolean"), FALSE)
})

test_that("gracefully handles int64 values", {
  # Largest/smallest 32-bit int
  expect_equal(
    bq_parse_single("2147483647", "integer"),
    bit64::as.integer64(2147483647L)
  )
  expect_equal(
    bq_parse_single("-2147483647", "integer"),
    bit64::as.integer64(-2147483647L)
  )
  # 2147483647 + 1
  expect_equal(
    bq_parse_single("2147483648", "integer"),
    bit64::as.integer64("2147483648")
  )
  expect_equal(
    bq_parse_single("-2147483648", "integer"),
    bit64::as.integer64("-2147483648")
  )
  # Largest/smallest integer64
  expect_equal(
    bq_parse_single("9223372036854775807", "integer"),
    bit64::as.integer64("9223372036854775807")
  )
  expect_equal(
    bq_parse_single("-9223372036854775807", "integer"),
    bit64::as.integer64("-9223372036854775807")
  )
  # Largest/smallest integer64 + 1
  expect_equal(
    bq_parse_single("9223372036854775808", "integer"),
    bit64::as.integer64(NA)
  )
  expect_equal(
    bq_parse_single("-9223372036854775808", "integer"),
    bit64::as.integer64(NA)
  )
})

test_that("can parse NULLs", {
  expect_identical(bq_parse_single(NULL, "string"), NA_character_)
  expect_identical(bq_parse_single(NULL, "integer"), bit64::as.integer64(NA))
  expect_identical(bq_parse_single(NULL, "float"), NA_real_)
  expect_identical(bq_parse_single(NULL, "boolean"), NA)

  na_dtm <- structure(as.POSIXct(NA), tzone = "UTC")
  expect_identical(bq_parse_single(NULL, "timestamp"), na_dtm)
})


test_that("can parse arrays of simple values", {
  x <- vs("1", "2", "3")

  out1 <- bq_parse_single(x, "string", mode = "repeated")
  expect_equal(out1, list(c("1", "2", "3")))

  out2 <- bq_parse_single(x, "integer", mode = "repeated")
  expect_equal(out2, list(bit64::as.integer64(1:3)))
})

test_that("can parse structs of simple values", {
  fields <- list(
    bq_field("x", "integer"),
    bq_field("y", "string")
  )

  x <- f(v("1"), v("a"))
  out <- bq_parse_single(x, "record", fields = fields)

  expect_equal(out, list(list(x = bit64::as.integer64(1L), y = "a")))
})

test_that("can parse structs of arrays", {
  fields <- list(
    bq_field("x", "integer", "repeated"),
    bq_field("y", "string", "repeated")
  )

  x <- f(v(vs("1", "2", "3")), v(vs("a", "b")))
  out <- bq_parse_single(x, "record", fields = fields)

  expect_equal(out, list(list(x = bit64::as.integer64(1:3), y = c("a", "b"))))
})


test_that("can parse arrays of structs", {
  fields <- list(
    bq_field("x", "integer"),
    bq_field("y", "string")
  )

  x <- vs(list(f = vs("1", "a")), list(f = vs("2", "b")))
  out <- bq_parse_single(x, "record", mode = "repeated", fields = fields)

  expect_equal(out, list(tibble(x = bit64::as.integer64(1:2), y = c("a", "b"))))
})


# Complete files ----------------------------------------------------------

replay_query <- function(name, sql) {
  schema_path <- test_path(sprintf("parse-schema-%s.json", name))
  values_path <- test_path(sprintf("parse-values-%s.json", name))

  if (!file.exists(schema_path)) {
    tbl <- bq_project_query(bq_test_project(), sql)
    bq_table_save_schema(tbl, schema_path)
    bq_table_save_values(tbl, values_path)
  }

  out <- bq_parse_file(schema_path, values_path)
  tibble::as_tibble(out)
}

test_that("can parse nested structures", {
  df <- replay_query("struct", "SELECT STRUCT(1 AS a, 'abc' AS b) as x")
  expect_named(df, "x")
  expect_type(df$x, "list")
  expect_equal(df$x[[1]], list(a = bit64::as.integer64(1), b = "abc"))

  df <- replay_query("array", "SELECT [1, 2, 3] as x")
  expect_named(df, "x")
  expect_type(df$x, "list")
  expect_equal(df$x[[1]], bit64::as.integer64(1:3))

  df <- replay_query(
    "array-struct",
    "SELECT [STRUCT(1 as a, 'a' as b), STRUCT(2, 'b'), STRUCT(3, 'c')] as x"
  )
  expect_named(df, "x")
  expect_type(df$x, "list")
  expect_equal(
    df$x[[1]],
    tibble(a = bit64::as.integer64(1:3), b = c("a", "b", "c"))
  )

  df <- replay_query(
    "struct-array",
    "SELECT STRUCT([1, 2, 3] as a, ['a', 'b'] as b) as x"
  )
  expect_named(df, "x")
  expect_type(df$x, "list")
  expect_equal(df$x[[1]], list(a = bit64::as.integer64(1:3), b = c("a", "b")))
})

test_that("can parse empty arrays", {
  tb <- bq_project_query(bq_test_project(), "SELECT ARRAY<INT64>[] as x")
  df <- bq_table_download(tb, api = "json")
  expect_equal(df$x, list(integer(length = 0)))

  tb <- bq_project_query(
    bq_test_project(),
    "SELECT ARRAY<STRUCT<a INT64, b STRING>>[] as x"
  )
  df <- bq_table_download(tb, api = "json")
  expect_equal(
    df$x,
    list(tibble::tibble(a = integer(length = 0), b = character()))
  )
})

test_that("can parse geography", {
  skip_if_not_installed("wk")

  wkt <- wk::wkt("POINT (30 10)")
  expect_identical(bq_parse_single(as.character(wkt), "geography"), wkt)
  expect_identical(
    bq_parse_single(NA_character_, "geography"),
    wk::wkt(NA_character_)
  )
})

test_that("can parse bytes", {
  bytes <- blob::blob(
    as.raw(c(
      0x01,
      0x01,
      0x00,
      0x00,
      0x00,
      0xff,
      0xff,
      0xff,
      0xff,
      0xff,
      0xff,
      0x3d,
      0x40,
      0x00,
      0x00,
      0x00,
      0x00,
      0x00,
      0x00,
      0x24,
      0x40
    ))
  )

  expect_identical(bq_parse_single(bytes[[1]], "bytes"), bytes)
  expect_identical(
    bq_parse_single(NA_character_, "bytes"),
    blob::blob(NULL)
  )
})

as_df <- function(x) {
  class(x) <- "data.frame"
  attr(x, "row.names") <- c(NA_integer_, -length(x[[1]]))

  x
}

check_quiet <- function(x, arg = caller_arg(x), call = caller_env()) {
  check_bool(x, allow_na = TRUE, arg = arg, call = call)

  if (is.na(x)) {
    !(is_interactive() || is_snapshot())
  } else {
    x
  }
}

check_digits <- function(x, arg = caller_arg(x), call = caller_env()) {
  check_number_whole(
    x,
    min = 0,
    max = 22,
    allow_null = TRUE,
    arg = arg,
    call = call
  )
  x %||% 22
}

bq_check_namespace <- function(pkg, bq_type) {
  check_installed(pkg, sprintf("to parse BigQuery '%s' fields.", bq_type))
}

cat_line <- function(...) {
  cat(paste0(..., "\n", collapse = ""))
}

big_mark <- function(x, ...) {
  mark <- if (identical(getOption("OutDec"), ",")) "." else ","
  format(x, big.mark = mark, scientific = FALSE, ...)
}

map_chr <- function(x, f, ...) {
  vapply(x, f, ..., FUN.VALUE = character(1))
}

indent <- function(x, n = 2) {
  space <- paste(rep(" ", n), collapse = "")
  paste0(space, gsub("\n", paste0("\n", space), x, fixed = TRUE))
}

as_json <- function(x) UseMethod("as_json")

#' @export
as_json.NULL <- function(x) NULL

# nocov start
show_json <- function(x) {
  jsonlite::toJSON(x, pretty = TRUE, auto_unbox = TRUE, digits = 22L)
}

#' @export
print.bq_bytes <- function(x, ...) {
  cat_line(prettyunits::pretty_bytes(unclass(x)))
}
# nocov end

defer <- function(expr, env = caller_env(), after = FALSE) {
  thunk <- as.call(list(function() expr))
  do.call(on.exit, list(thunk, TRUE, after), envir = env)
}

in_pkgdown <- function() {
  identical(Sys.getenv("IN_PKGDOWN"), "true")
}

as_query <- function(x, error_arg = caller_arg(x), error_call = caller_env()) {
  if (is(x, "SQL")) {
    x <- x@.Data
  }
  check_string(x, arg = error_arg, call = error_call)
  x
}

has_bigrquerystorage <- function() {
  is_installed("bigrquerystorage")
}

cli_escape <- function(x) {
  x <- gsub("{", "{{", x, fixed = TRUE)
  x <- gsub("}", "}}", x, fixed = TRUE)
  x
}

check_labels <- function(
  labels,
  arg = caller_arg(labels),
  call = caller_env()
) {
  if (is.null(labels)) {
    return(invisible())
  }

  if (!is.list(labels) || any(names2(labels) == "")) {
    stop_input_type(
      labels,
      "a named list",
      allow_null = TRUE,
      arg = arg,
      call = call
    )
  }
  is_string <- vapply(labels, is_string, logical(1))
  if (!all(is_string)) {
    stop_input_type(
      labels,
      "a named list of strings",
      arg = arg,
      call = call
    )
  }

  invisible()
}

cli_escape <- function(x) {
  x <- gsub("{", "{{", x, fixed = TRUE)
  x <- gsub("}", "}}", x, fixed = TRUE)
  x
}

test_that("bq_check_namespace() works", {
  expect_no_error(bq_check_namespace("bigrquery", "FIELD_TYPE"))
  expect_snapshot(
    bq_check_namespace("invalid package name", "FIELD_TYPE"),
    error = TRUE
  )
})

test_that("cli_escape() doubles cli braces", {
  expect_equal(cli_escape("no braces"), "no braces")
  expect_equal(cli_escape("{x}"), "{{x}}")
})

test_that("check_labels() accepts valid labels and NULL-like inputs", {
  expect_no_error(check_labels(NULL))
  expect_no_error(check_labels(list()))
  expect_no_error(check_labels(list(env = "prod")))
  expect_no_error(check_labels(list(env = "prod", team = "data")))
  expect_no_error(check_labels(list(env = "")))
})

test_that("check_labels() errors on invalid inputs", {
  expect_snapshot(error = TRUE, {
    check_labels(c(env = "prod"))
    check_labels(list("no-name"))
    check_labels(list(env = 1))
  })
})

test_that("cli_escape() doubles cli braces", {
  expect_equal(cli_escape("no braces"), "no braces")
  expect_equal(cli_escape("{x}"), "{{x}}")
})

## This file is the interface between bigrquery and the
## auth functionality in gargle.

# Initialization happens in .onLoad
.auth <- NULL

## The roxygen comments for these functions are mostly generated from data
## in this list and template text maintained in gargle.
gargle_lookup_table <- list(
  PACKAGE = "bigrquery",
  YOUR_STUFF = "your BigQuery projects",
  PRODUCT = "Google BigQuery",
  API = "BigQuery API",
  PREFIX = "bq"
)

#' Authorize bigrquery
#'
#' @eval gargle:::PREFIX_auth_description(gargle_lookup_table)
#' @eval gargle:::PREFIX_auth_details(gargle_lookup_table)
#' @eval gargle:::PREFIX_auth_params()
#' @param scopes A character vector of scopes to request.
#'   Pick from those listed at <https://developers.google.com/identity/protocols/oauth2/scopes>.
#'
#' @family auth functions
#' @export
#'
#' @examples
#' \dontrun{
#' ## load/refresh existing credentials, if available
#' ## otherwise, go to browser for authentication and authorization
#' bq_auth()
#'
#' ## force use of a token associated with a specific email
#' bq_auth(email = "jenny@example.com")
#'
#' ## force a menu where you can choose from existing tokens or
#' ## choose to get a new one
#' bq_auth(email = NA)
#'
#' ## use a 'read only' scope, so it's impossible to change data
#' bq_auth(
#'   scopes = "https://www.googleapis.com/auth/devstorage.read_only"
#' )
#'
#' ## use a service account token
#' bq_auth(path = "foofy-83ee9e7c9c48.json")
#' }
#'
#' @importFrom gargle token_fetch
bq_auth <- function(
  email = gargle::gargle_oauth_email(),
  path = NULL,
  scopes = c(
    "https://www.googleapis.com/auth/bigquery",
    "https://www.googleapis.com/auth/cloud-platform"
  ),
  cache = gargle::gargle_oauth_cache(),
  use_oob = gargle::gargle_oob_default(),
  token = NULL
) {
  if (!missing(email) && !missing(path)) {
    cli::cli_warn(c(
      "It is very unusual to provide both {.arg email} and \\
       {.arg path} to {.fun bq_auth}.",
      "They relate to two different auth methods.",
      "The {.arg path} argument is only for a service account token.",
      "If you need to specify your own OAuth client, use \\
      {.fun bq_auth_configure}."
    ))
  }

  # In a BYO token situation, such as `bq_auth(token = drive_token())`, it's
  # easy to not have, e.g., googledrive attached (provides drive_token()).
  # If we don't force here, the error is muffled in token_fetch()'s tryCatch()
  # treatment, which makes it much harder to figure out what's wrong.
  # By forcing here, we expose this mistake early and noisily.
  force(token)

  cred <- token_fetch(
    scopes = scopes,
    client = bq_oauth_client() %||% gargle::tidyverse_client(),
    email = email,
    path = path,
    package = "bigrquery",
    cache = cache,
    use_oob = use_oob,
    token = token
  )
  if (!inherits(cred, "Token2.0")) {
    cli::cli_abort(c(
      "Can't get Google credentials.",
      i = if (!is_interactive()) {
        "Try calling {.fun bq_auth} directly with necessary specifics."
      }
    ))
  }
  .auth$set_cred(cred)
  .auth$set_auth_active(TRUE)

  invisible()
}

has_internal_auth <- function() {
  gargle::secret_has_key("BIGRQUERY_KEY")
}

bq_auth_internal <- function() {
  path <- system.file("secret", "bigrquery-testing.json", package = "bigrquery")
  json <- gargle::secret_decrypt_json(path, "BIGRQUERY_KEY")
  bq_auth(path = json)
}

#' Clear current token
#'
#' Clears any currently stored token. The next time bigrquery needs a token, the
#' token acquisition process starts over, with a fresh call to [bq_auth()] and,
#' therefore, internally, a call to [gargle::token_fetch()]. Unlike some other
#' packages that use gargle, bigrquery is not usable in a de-authorized state.
#' Therefore, calling `bq_deauth()` only clears the token, i.e. it does NOT
#' imply that subsequent requests are made with an API key in lieu of a token.
#'
#' @family auth functions
#' @export
#' @examples
#' \dontrun{
#' bq_deauth()
#' }
bq_deauth <- function() {
  .auth$clear_cred()
  invisible()
}

#' Produce configured token
#'
#' @eval gargle:::PREFIX_token_description(gargle_lookup_table)
#' @eval gargle:::PREFIX_token_return()
#'
#' @family low-level API functions
#' @export
#' @examples
#' \dontrun{
#' bq_token()
#' }
bq_token <- function() {
  if (!bq_has_token()) {
    bq_auth()
  }
  httr::config(token = .auth$cred)
}

#' Is there a token on hand?
#'
#' @eval gargle:::PREFIX_has_token_description(gargle_lookup_table)
#' @eval gargle:::PREFIX_has_token_return()
#'
#' @family low-level API functions
#' @export
#'
#' @examples
#' bq_has_token()
bq_has_token <- function() {
  inherits(.auth$cred, "Token2.0")
}

#' Edit and view auth configuration
#'
#' @eval gargle:::PREFIX_auth_configure_description(gargle_lookup_table, .has_api_key = FALSE)
#' @param client A Google OAuth client, presumably constructed via
#'   [gargle::gargle_oauth_client_from_json()]. Note, however, that it is
#'   preferred to specify the client with JSON, using the `path` argument.
#' @inheritParams gargle::gargle_oauth_client_from_json
#' @eval gargle:::PREFIX_auth_configure_return(gargle_lookup_table, .has_api_key = FALSE)
#'
#' @family auth functions
#' @export
#' @examples
#' # see and store the current user-configured OAuth client (probably `NULL`)
#' (original_client <- bq_oauth_client())
#'
#' # the preferred way to configure your own client is via a JSON file
#' # downloaded from Google Developers Console
#' # this example JSON is indicative, but fake
#' path_to_json <- system.file(
#'   "extdata", "data", "client_secret_123.googleusercontent.com.json",
#'   package = "bigrquery"
#' )
#' bq_auth_configure(path = path_to_json)
#'
#' # confirm the changes
#' bq_oauth_client()
#'
#' # restore original auth config
#' bq_auth_configure(client = original_client)
bq_auth_configure <- function(client, path) {
  check_exclusive(client, path)
  if (!missing(path)) {
    check_string(path)
    client <- gargle::gargle_oauth_client_from_json(path)
  } else {
    if (!is.null(client) && !inherits(client, "gargle_oauth_client")) {
      stop_input_type(client, "a gargle OAuth client", allow_null = TRUE)
    }
  }

  .auth$set_client(client)
  invisible(.auth)
}

#' @export
#' @rdname bq_auth_configure
bq_oauth_client <- function() {
  .auth$client
}

#' Get info on current user
#'
#' @eval gargle:::PREFIX_user_description()
#' @eval gargle:::PREFIX_user_seealso()
#' @eval gargle:::PREFIX_user_return()
#'
#' @export
#' @examples
#' \dontrun{
#' bq_user()
#' }
bq_user <- function() {
  if (bq_has_token()) {
    gargle::token_email(bq_token())
  } else {
    NULL
  }
}

test_that("bq_user() works", {
  skip_if_no_auth()
  expect_match(bq_user(), "@.*[.]iam[.]gserviceaccount[.]com")
})

test_that("useful error if can't auth", {
  local_mocked_bindings(token_fetch = function(...) NULL)

  expect_snapshot(bq_auth(), error = TRUE)
})

test_that("bq_auth_configure checks its inputs", {
  expect_snapshot(error = TRUE, {
    bq_auth_configure(1, 1)
    bq_auth_configure(client = 1)
    bq_auth_configure(path = 1)
  })
})

base_url <- "https://bigquery.googleapis.com/bigquery/v2/"
upload_url <- "https://bigquery.googleapis.com/upload/bigquery/v2/"

prepare_bq_query <- function(query) {
  api_key <- Sys.getenv("BIGRQUERY_API_KEY")
  if (!nzchar(api_key)) {
    return(query)
  }
  query <- query %||% list()
  query[["key"]] <- query[["key"]] %||% api_key
  query
}

bq_path <- function(project, dataset = NULL, table = NULL, ...) {
  check_string(project, allow_null = TRUE)
  check_string(dataset, allow_null = TRUE)
  check_string(table, allow_null = TRUE)

  components <- c(
    projects = project,
    datasets = dataset,
    tables = table,
    ...
  )

  paste0(names(components), "/", components, collapse = "/")
}

bq_ua <- function() {
  paste0(
    "bigrquery/",
    utils::packageVersion("bigrquery"),
    " ",
    "(GPN:RStudio; )",
    " ",
    "gargle/",
    utils::packageVersion("gargle"),
    " ",
    "httr/",
    utils::packageVersion("httr")
  )
}

bq_body <- function(body, ...) {
  user <- toCamel(list(...))
  utils::modifyList(body, user)
}


#' @importFrom httr GET config
bq_get <- function(url, ..., query = NULL, raw = FALSE, token = bq_token()) {
  req <- GET(
    paste0(base_url, url),
    token,
    httr::user_agent(bq_ua()),
    ...,
    query = prepare_bq_query(query)
  )
  process_request(req, raw = raw)
}

bq_exists <- function(url, ..., query = NULL, token = bq_token()) {
  req <- GET(
    paste0(base_url, url),
    token,
    httr::user_agent(bq_ua()),
    ...,
    query = prepare_bq_query(query)
  )
  status_code(req) >= 200 && status_code(req) < 300
}


#' @importFrom httr GET config
bq_get_paginated <- function(
  url,
  ...,
  query = NULL,
  token = bq_token(),
  page_size = 50,
  max_pages = Inf,
  warn = TRUE
) {
  check_number_whole(max_pages, min = 1, allow_infinite = TRUE)
  check_number_whole(page_size, min = 1)

  if (!is.null(query$fields)) {
    query$fields <- paste0(query$fields, ",nextPageToken")
  }

  query <- utils::modifyList(list(maxResults = page_size), query %||% list())
  pages <- list()

  page <- bq_get(url, ..., query = query, token = token)
  i <- 1
  pages[[i]] <- page
  page_token <- page$nextPageToken

  while (!is.null(page_token) && i < max_pages) {
    query$pageToken <- page_token
    page <- bq_get(url, ..., query = query, token = token)

    i <- i + 1
    pages[[i]] <- page
    page_token <- page$nextPageToken
  }

  if (isTRUE(warn) && !is.null(page_token)) {
    warning(
      "Only first ",
      big_mark(max_pages * page_size),
      " results retrieved. ",
      "Adjust with `max_pages` and `page_size` parameters.",
      call. = FALSE
    )
  }

  pages
}


#' @importFrom httr DELETE config
bq_delete <- function(url, ..., query = NULL, token = bq_token()) {
  req <- DELETE(
    paste0(base_url, url),
    token,
    httr::user_agent(bq_ua()),
    ...,
    query = prepare_bq_query(query)
  )
  process_request(req)
}

#' @importFrom httr POST add_headers config
bq_post <- function(url, body, ..., query = NULL, token = bq_token()) {
  json <- jsonlite::toJSON(body, pretty = TRUE, auto_unbox = TRUE)

  req <- POST(
    paste0(base_url, url),
    body = json,
    httr::user_agent(bq_ua()),
    token,
    add_headers("Content-Type" = "application/json"),
    ...,
    query = prepare_bq_query(query)
  )
  invisible(process_request(req))
}

#' @importFrom httr PATCH add_headers config
bq_patch <- function(url, body, ..., query = NULL, token = bq_token()) {
  json <- jsonlite::toJSON(body, pretty = TRUE, auto_unbox = TRUE)
  req <- PATCH(
    paste0(base_url, url),
    body = json,
    httr::user_agent(bq_ua()),
    token,
    add_headers("Content-Type" = "application/json"),
    ...,
    query = prepare_bq_query(query)
  )
  process_request(req)
}

#' @importFrom httr POST PUT add_headers headers config status_code
# https://cloud.google.com/bigquery/docs/reference/api-uploads
bq_upload <- function(
  url,
  metadata,
  media,
  query = list(),
  token = bq_token()
) {
  query <- utils::modifyList(
    list(fields = "jobReference", uploadType = "resumable"),
    query
  )
  config <- add_headers("Content-Type" = metadata[["type"]])

  req <- POST(
    paste0(upload_url, url),
    body = metadata[["content"]],
    httr::user_agent(bq_ua()),
    token,
    config,
    query = query
  )

  if (status_code(req) == 200) {
    config <- add_headers("Content-Type" = media[["type"]])

    req <- PUT(
      headers(req)$location,
      body = media[["content"]],
      httr::user_agent(bq_ua()),
      token,
      config
    )
  }

  process_request(req)
}


#' @importFrom httr http_status content parse_media status_code
process_request <- function(req, raw = FALSE, call = caller_env()) {
  status <- status_code(req)
  # No content -> success
  if (status == 204) {
    return(TRUE)
  }

  type <- req$headers$`Content-type`
  content <- content(req, "raw")

  bq_check_response(
    status = status,
    type = type,
    content = content,
    call = call
  )

  if (raw) {
    content
  } else {
    jsonlite::fromJSON(rawToChar(content), simplifyVector = FALSE)
  }
}

bq_check_response <- function(status, type, content, call = caller_env()) {
  if (status >= 200 && status < 300) {
    return()
  }

  type <- httr::parse_media(type)
  text <- rawToChar(content)

  if (type$complete == "application/json") {
    json <- jsonlite::fromJSON(text, simplifyVector = FALSE)
    gargle_abort(
      reason = json$error$errors[[1L]]$reason,
      message = json$error$message,
      status = status,
      call = call
    )
  } else {
    message <- paste0("HTTP error [", status, "]\n", text)
    gargle_abort(
      reason = NULL,
      message = message,
      status = status,
      call = call
    )
  }
}

gargle_abort <- function(reason, message, status, call = caller_env()) {
  class <- paste0("bigrquery_http_", status)

  if (!is.null(reason)) {
    advice <- NULL
    if (reason == "responseTooLarge") {
      # If message mentions "allowLargeResults", that's the right advice to
      # follow. But other times we get the error when a single page of results
      # is too large (e.g. for tables with many columns). By decreasing page
      # size, a single page of results is now small enough to return.
      if (!any(grepl("allowLargeResults", message, fixed = TRUE))) {
        advice <- "Try decreasing the `page_size` value of `bq_table_download()`"
      }
    } else if (reason == "rateLimitExceeded") {
      # bigrquery pulls results in parallel to increase throughput.
      # "rateLimitExceeded" can happen when those threads are making too many
      # requests within a brief time interval. Can slow threads down by asking
      # for bigger pages.
      advice <- "Try increasing the `page_size` value of `bq_table_download()`"
    }
    message <- c(
      cli_escape(paste0(message, " [", reason, "] ")),
      i = advice
    )
    class <- c(paste0("bigrquery_", reason), class)
  } else {
    message <- cli_escape(message)
  }

  cli::cli_abort(message, class = class, call = call)
}

test_that("api keys are added when present", {
  key <- "my.secret.key"
  withr::local_envvar(list(BIGRQUERY_API_KEY = key))

  expect_equal(prepare_bq_query(NULL), list(key = key))
  expect_equal(
    prepare_bq_query(list(herring_color = "red")),
    list(herring_color = "red", key = key)
  )
})

test_that("explicit api keys override env vars", {
  key <- "my.secret.key"
  withr::local_envvar(list(BIGRQUERY_API_KEY = key))

  expect_equal(
    prepare_bq_query(list(key = "my.other.key")),
    list(key = "my.other.key")
  )
})

test_that("pagination warns if pages left on server", {
  skip_if_no_auth()

  expect_warning(
    bq_get_paginated(
      bq_path("bigquery-public-data", ""),
      query = list(fields = "datasets(datasetReference)"),
      page_size = 10,
      max_pages = 2
    ),
    "Only first 20 results"
  )
})

test_that("error call is forwarded all the way down", {
  skip_if_not(bq_authable())

  expect_snapshot(bq_job_meta("a.b.c"), error = TRUE)
})

gs_object <- function(bucket, object) {
  structure(
    list(bucket = bucket, object = object),
    class = "gs_object"
  )
}

#' @export
as.character.gs_object <- function(x, ...) {
  format(x)
}

#' @export
format.gs_object <- function(x, ...) {
  sprintf("gs://%s/%s", x$bucket, x$object)
}

#' @export
print.gs_object <- function(x, ...) {
  cat_line("<gs_object> ", format(x))
  invisible(x)
}

gs_object_delete <- function(x, token = bq_token()) {
  url <- sprintf(
    "https://storage.googleapis.com/storage/v1/b/%s/o/%s",
    x$bucket,
    x$object
  )
  req <- httr::DELETE(url, token, httr::user_agent(bq_ua()))
  process_request(req)
}

gs_object_exists <- function(x, token = bq_token()) {
  url <- sprintf(
    "https://storage.googleapis.com/storage/v1/b/%s/o/%s",
    x$bucket,
    x$object
  )
  req <- httr::GET(url, token, httr::user_agent(bq_ua()))
  req$status_code != 404
}

#' BigQuery datasets
#'
#' Basic create-read-update-delete verbs for datasets.
#'
#' @param x A [bq_dataset]
#' @inheritParams api-job
#' @inheritParams api-perform
#' @inheritParams bq_projects
#'
#' @section Google BigQuery API documentation:
#' * [get](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/datasets/get)
#' * [insert](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/datasets/insert)
#' * [delete](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/datasets/delete)
#' * [list](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables/list)
#' @examplesIf bq_testable()
#' ds <- bq_dataset(bq_test_project(), "dataset_api")
#' bq_dataset_exists(ds)
#'
#' bq_dataset_create(ds)
#' bq_dataset_exists(ds)
#' str(bq_dataset_meta(ds))
#'
#' bq_dataset_delete(ds)
#' bq_dataset_exists(ds)
#'
#' # Use bq_test_dataset() to create a temporary dataset that will
#' # be automatically deleted
#' ds <- bq_test_dataset()
#' bq_table_create(bq_table(ds, "x1"))
#' bq_table_create(bq_table(ds, "x2"))
#' bq_table_create(bq_table(ds, "x3"))
#' bq_dataset_tables(ds)
#' @name api-dataset
NULL

#' @export
#' @rdname api-dataset
#' @param location Dataset location
bq_dataset_create <- function(x, location = "US", ...) {
  x <- as_bq_dataset(x)

  url <- bq_path(x$project, "")
  body <- list(
    datasetReference = datasetReference(x),
    location = location
  )
  bq_post(url, body = bq_body(body, ...))

  x
}

#' @export
#' @rdname api-dataset
bq_dataset_meta <- function(x, fields = NULL) {
  x <- as_bq_dataset(x)

  url <- bq_path(x$project, x$dataset)
  bq_get(url, query = list(fields = fields))
}

#' @export
#' @rdname api-dataset
bq_dataset_exists <- function(x) {
  x <- as_bq_dataset(x)

  url <- bq_path(x$project, x$dataset)
  bq_exists(url)
}

#' @export
#' @rdname api-dataset
bq_dataset_update <- function(x, ...) {
  x <- as_bq_dataset(x)

  url <- bq_path(x$project, x$dataset)
  body <- list(datasetReference = datasetReference(x))
  bq_patch(url, body = bq_body(body, ...))

  invisible(x)
}

#' @export
#' @rdname api-dataset
#' @param delete_contents If `TRUE`, will recursively delete all tables in
#'   the dataset. Set to `FALSE` by default for safety.
bq_dataset_delete <- function(x, delete_contents = FALSE) {
  x <- as_bq_dataset(x)

  url <- bq_path(x$project, x$dataset)
  query <- list(deleteContents = delete_contents)
  bq_delete(url, query = query)

  invisible(x)
}

#' @export
#' @rdname api-dataset
bq_dataset_tables <- function(
  x,
  page_size = 50,
  max_pages = Inf,
  warn = TRUE,
  ...
) {
  x <- as_bq_dataset(x)
  url <- bq_path(x$project, x$dataset, "")

  data <- bq_get_paginated(
    url,
    query = list(fields = "tables(tableReference,type)"),
    page_size = page_size,
    max_pages = max_pages,
    warn = warn
  )

  tables <- unlist(lapply(data, function(x) x$tables), recursive = FALSE)
  lapply(tables, function(x) as_bq_table(x$tableReference, type = x$type))
}

#' @include dbi-driver.R
NULL

BigQueryConnection <- function(
  project,
  dataset,
  billing,
  page_size = 1e4,
  quiet = NA,
  use_legacy_sql = FALSE,
  bigint = c("integer", "integer64", "numeric", "character"),
  labels = NULL
) {
  connection_capture()

  new(
    "BigQueryConnection",
    project = project,
    dataset = dataset,
    billing = billing,
    page_size = as.integer(page_size),
    quiet = quiet,
    use_legacy_sql = use_legacy_sql,
    bigint = match.arg(bigint),
    labels = labels
  )
}

#' @rdname DBI
#' @export
setClass(
  "BigQueryConnection",
  contains = "DBIConnection",
  slots = list(
    project = "character",
    dataset = "ANY",
    billing = "character",
    use_legacy_sql = "logical",
    page_size = "integer",
    quiet = "logical",
    bigint = "character",
    labels = "ANY"
  )
)

#' @rdname DBI
#' @inheritParams methods::show
#' @export
setMethod(
  "show",
  "BigQueryConnection",
  function(object) {
    cat_line("<BigQueryConnection>")

    if (!is.null(object@dataset)) {
      cat_line("  Dataset: ", object@project, ".", object@dataset)
    }
    cat_line("  Billing: ", object@billing)
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbIsValid
#' @param dbObj An object inheriting from [DBI::DBIObject-class],
#' i.e. [DBI::DBIDriver-class], [DBI::DBIConnection-class], or a [DBI::DBIResult-class].
#' @param conn A [DBI::DBIConnection-class] object, as returned by
#'   [DBI::dbConnect()].
#' @export
setMethod(
  "dbIsValid",
  "BigQueryConnection",
  function(dbObj, ...) {
    TRUE
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbDisconnect
#' @export
setMethod(
  "dbDisconnect",
  "BigQueryConnection",
  function(conn, ...) {
    on_connection_closed(conn)
    invisible(TRUE)
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbSendQuery
#' @param params Named list of parameters match to query parameters.
#'   Parameter `x` will be matched to placeholder `@x`.
#' @export
setMethod(
  "dbSendQuery",
  c("BigQueryConnection", "character"),
  function(conn, statement, ..., params = NULL) {
    check_for_parameters(..., call = quote(dbSendQuery()))
    BigQueryResult(conn, statement, params = params)
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbSendQuery
#' @export
setMethod(
  "dbExecute",
  c("BigQueryConnection", "character"),
  function(conn, statement, ..., params = NULL) {
    ds <- if (!is.null(conn@dataset)) as_bq_dataset(conn)

    check_for_parameters(..., call = quote(dbExecute()))
    job <- bq_perform_query(
      statement,
      billing = conn@billing,
      default_dataset = ds,
      quiet = conn@quiet,
      parameters = params,
      labels = conn@labels,
      ...
    )
    bq_job_wait(job, quiet = conn@quiet)

    meta <- bq_job_meta(job, "statistics(query(numDmlAffectedRows))")
    as.numeric(meta$statistics$query$numDmlAffectedRows %||% 0)
  }
)

check_for_parameters <- function(..., parameters = NULL, call = caller_env()) {
  if (is.null(parameters)) {
    return()
  }
  cli::cli_abort("Use {.arg params} not {.arg parameters}.", call = call)
}

#' @rdname DBI
#' @inheritParams DBI::dbQuoteString
#' @export
setMethod(
  "dbQuoteString",
  c("BigQueryConnection", "character"),
  function(conn, x, ...) {
    if (length(x) == 0) {
      return(SQL(character()))
    }

    out <- encodeString(x, na.encode = FALSE, quote = "'")
    out[is.na(x)] <- "NULL"
    SQL(out, names = names(x))
  }
)

#' @rdname DBI
#' @export
setMethod(
  "dbQuoteString",
  c("BigQueryConnection", "SQL"),
  function(conn, x, ...) {
    x
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbQuoteIdentifier
#' @export
setMethod(
  "dbQuoteIdentifier",
  c("BigQueryConnection", "character"),
  function(conn, x, ...) {
    if (length(x) == 0) {
      return(SQL(character()))
    }

    if (any(is.na(x))) {
      cli::cli_abort("{.arg x} must not contain missing values.")
    }

    if (conn@use_legacy_sql) {
      out <- paste0("[", x, "]")
    } else {
      out <- encodeString(x, quote = "`")
    }

    SQL(out, names = names(x))
  }
)

#' @rdname DBI
#' @export
setMethod(
  "dbQuoteIdentifier",
  c("BigQueryConnection", "SQL"),
  function(conn, x, ...) {
    x
  }
)

#' @rdname DBI
#' @export
setMethod(
  "dbQuoteLiteral",
  c("BigQueryConnection", "logical"),
  function(conn, x, ...) {
    x <- as.character(x)
    x[is.na(x)] <- "NULL"
    SQL(x, names = names(x))
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbDataType
#' @export
setMethod(
  "dbDataType",
  "BigQueryConnection",
  function(dbObj, obj, ...) {
    data_type(obj)
  }
)

dbWriteTable_bq <- function(
  conn,
  name,
  value,
  ...,
  overwrite = FALSE,
  append = FALSE,
  field.types = NULL,
  temporary = FALSE,
  row.names = NA
) {
  check_bool(overwrite)
  check_bool(append)

  if (!is.null(field.types)) {
    cli::cli_abort(
      "{.arg field.types} not supported by bigrquery.",
      call = quote(DBI::dbWriteTable())
    )
  }
  if (!identical(temporary, FALSE)) {
    cli::cli_abort(
      "{.code temporary = TRUE} not supported by bigrquery.",
      call = quote(DBI::dbWriteTable())
    )
  }

  if (append) {
    create_disposition <- "CREATE_NEVER"
    write_disposition <- "WRITE_APPEND"
  } else {
    create_disposition <- "CREATE_IF_NEEDED"
    write_disposition <- if (overwrite) "WRITE_TRUNCATE" else "WRITE_EMPTY"
  }
  tb <- as_bq_table(conn, name)

  bq_table_upload(
    tb,
    value,
    create_disposition = create_disposition,
    write_disposition = write_disposition,
    billing = conn@billing,
    labels = conn@labels,
    ...
  )
  invisible(TRUE)
}

#' @rdname DBI
#' @inheritParams DBI::dbWriteTable
#' @param row.names A logical specifying whether the `row.names` should be
#'   output to the output DBMS table; if `TRUE`, an extra field whose name
#'   will be whatever the R identifier `"row.names"` maps to the DBMS (see
#'   [DBI::make.db.names()]). If `NA` will add rows names if
#'   they are characters, otherwise will ignore.
#' @param overwrite a logical specifying whether to overwrite an existing table
#'   or not. Its default is `FALSE`.
#' @param append a logical specifying whether to append to an existing table
#'   in the DBMS.  Its default is `FALSE`.
#' @param field.types,temporary Ignored. Included for compatibility with
#'   generic.
#' @export
setMethod(
  "dbWriteTable",
  c("BigQueryConnection", "character", "data.frame"),
  dbWriteTable_bq
)

#' @rdname DBI
#' @export
setMethod(
  "dbWriteTable",
  c("BigQueryConnection", "Id", "data.frame"),
  dbWriteTable_bq
)

#' @rdname DBI
#' @export
setMethod(
  "dbWriteTable",
  c("BigQueryConnection", "AsIs", "data.frame"),
  dbWriteTable_bq
)


dbAppendTable_bq <- function(conn, name, value, ..., row.names = NULL) {
  tb <- as_bq_table(conn, name)

  bq_table_upload(
    tb,
    value,
    create_disposition = "CREATE_NEVER",
    write_disposition = "WRITE_APPEND",
    billing = conn@billing,
    labels = conn@labels,
    ...
  )
  on_connection_updated(conn, toString(tb))

  invisible(TRUE)
}

#' @inheritParams DBI::dbAppendTable
#' @rdname DBI
#' @export
setMethod(
  "dbAppendTable",
  c("BigQueryConnection", "character", "data.frame"),
  dbAppendTable_bq
)

#' @rdname DBI
#' @export
setMethod(
  "dbAppendTable",
  c("BigQueryConnection", "Id", "data.frame"),
  dbAppendTable_bq
)

#' @rdname DBI
#' @export
setMethod(
  "dbAppendTable",
  c("BigQueryConnection", "AsIs", "data.frame"),
  dbAppendTable_bq
)

dbCreateTable_bq <- function(
  conn,
  name,
  fields,
  ...,
  row.names = NULL,
  temporary = FALSE
) {
  if (!identical(temporary, FALSE)) {
    cli::cli_abort(
      "{.code temporary = TRUE} not supported by bigrquery.",
      call = quote(DBI::dbCreateTable())
    )
  }

  tb <- as_bq_table(conn, name)
  bq_table_create(tb, fields)
  on_connection_updated(conn, toString(tb))

  invisible(TRUE)
}

#' @inheritParams DBI::dbCreateTable
#' @rdname DBI
#' @export
setMethod("dbCreateTable", "BigQueryConnection", dbCreateTable_bq)

#' @rdname DBI
#' @export
setMethod("dbCreateTable", "BigQueryConnection", dbCreateTable_bq)

dbReadTable_bq <- function(conn, name, ...) {
  tb <- as_bq_table(conn, name)
  bq_table_download(tb, ..., api = "json")
}

#' @rdname DBI
#' @inheritParams DBI::dbReadTable
#' @export
setMethod("dbReadTable", c("BigQueryConnection", "character"), dbReadTable_bq)

#' @rdname DBI
#' @export
setMethod("dbReadTable", c("BigQueryConnection", "Id"), dbReadTable_bq)

#' @rdname DBI
#' @export
setMethod("dbReadTable", c("BigQueryConnection", "AsIs"), dbReadTable_bq)

#' @rdname DBI
#' @inheritParams DBI::dbListTables
#' @export
setMethod(
  "dbListTables",
  "BigQueryConnection",
  function(conn, ...) {
    if (is.null(conn@dataset)) {
      cli::cli_abort("Can't list tables without a connection `dataset`.")
    }
    ds <- bq_dataset(conn@project, conn@dataset)

    tbs <- bq_dataset_tables(ds, ...)
    map_chr(tbs, function(x) x$table)
  }
)

dbExistsTable_bq <- function(conn, name, ...) {
  tb <- as_bq_table(conn, name)
  bq_table_exists(tb)
}
#' @inheritParams DBI::dbExistsTable
#' @rdname DBI
#' @export
setMethod(
  "dbExistsTable",
  c("BigQueryConnection", "character"),
  dbExistsTable_bq
)

#' @rdname DBI
#' @export
setMethod("dbExistsTable", c("BigQueryConnection", "Id"), dbExistsTable_bq)

#' @rdname DBI
#' @export
setMethod("dbExistsTable", c("BigQueryConnection", "AsIs"), dbExistsTable_bq)

dbListFields_bq <- function(conn, name, ...) {
  tb <- as_bq_table(conn, name)
  flds <- bq_table_fields(tb)
  map_chr(flds, function(x) x$name)
}

#' @inheritParams DBI::dbListFields
#' @rdname DBI
#' @export
setMethod("dbListFields", c("BigQueryConnection", "character"), dbListFields_bq)

#' @rdname DBI
#' @export
setMethod("dbListFields", c("BigQueryConnection", "Id"), dbListFields_bq)

#' @rdname DBI
#' @export
setMethod("dbListFields", c("BigQueryConnection", "AsIs"), dbListFields_bq)

dbRemoveTable_bq <- function(conn, name, ...) {
  tb <- as_bq_table(conn, name)
  bq_table_delete(tb)
  on_connection_updated(conn, toString(tb))
  invisible(TRUE)
}

#' @inheritParams DBI::dbRemoveTable
#' @rdname DBI
#' @export
setMethod(
  "dbRemoveTable",
  c("BigQueryConnection", "character"),
  dbRemoveTable_bq
)

#' @rdname DBI
#' @export
setMethod("dbRemoveTable", c("BigQueryConnection", "Id"), dbRemoveTable_bq)

#' @rdname DBI
#' @export
setMethod("dbRemoveTable", c("BigQueryConnection", "AsIs"), dbRemoveTable_bq)

# nocov start
#' @rdname DBI
#' @inheritParams DBI::dbGetInfo
#' @export
setMethod(
  "dbGetInfo",
  "BigQueryConnection",
  function(dbObj, ...) {
    list(
      db.version = NA,
      dbname = paste0(c(dbObj@project, dbObj@dataset), collapse = "."),
      username = NA,
      host = NA,
      port = NA
    )
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbBegin
#' @export
setMethod(
  "dbBegin",
  "BigQueryConnection",
  function(conn, ...) {
    testthat::skip("Not yet implemented: dbBegin(Connection)")
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbCommit
#' @export
setMethod(
  "dbCommit",
  "BigQueryConnection",
  function(conn, ...) {
    testthat::skip("Not yet implemented: dbCommit(Connection)")
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbRollback
#' @export
setMethod(
  "dbRollback",
  "BigQueryConnection",
  function(conn, ...) {
    testthat::skip("Not yet implemented: dbRollback(Connection)")
  }
)
# nocov end

# Convert to bq objects ---------------------------------------------------

#' @export
as_bq_dataset.BigQueryConnection <- function(x, ..., error_arg, error_call) {
  bq_dataset(x@project, x@dataset)
}


#' @export
as_bq_table.BigQueryConnection <- function(x, name, ...) {
  if (inherits(name, "dbplyr_table_path")) {
    # dbplyr 2.5.0
    pieces <- utils::getFromNamespace("table_path_components", "dbplyr")(
      name,
      x
    )[[1]]
  } else if (inherits(name, "dbplyr_table_ident")) {
    # dbplyr 2.4.0
    name <- unclass(name)
    pieces <- c(name$catalog, name$schema, name$table)
    pieces <- pieces[!is.na(pieces)]

    if (length(pieces) == 1) {
      pieces <- strsplit(pieces, ".", fixed = TRUE)[[1]]
    }
  } else if (is(name, "Id")) {
    pieces <- unname(name@name)
  } else if (is.character(name) && length(name) == 1) {
    # Technically incorrect according to the DBI spec, because it should
    # automatically quote the name; but too high risk to change now
    pieces <- strsplit(name, ".", fixed = TRUE)[[1]]
  } else {
    cli::cli_abort("{.arg name} must be a string or a dbplyr_table_ident.")
  }

  if (length(pieces) > 3) {
    cli::cli_abort("{.arg name} ({.str {name}}) must have 1-3 components.")
  }
  if (length(pieces) == 1 && is.null(x@dataset)) {
    cli::cli_abort(
      "{.arg name} ({.str {name}}) must have 2 or 3 components if the connection doesn't have a dataset."
    )
  }

  switch(
    length(pieces),
    bq_table(x@project, x@dataset, pieces[[1]]),
    bq_table(x@project, pieces[[1]], pieces[[2]]),
    bq_table(pieces[[1]], pieces[[2]], pieces[[3]])
  )
}

#' BigQuery project methods
#'
#' Projects have two primary components: datasets and jobs. Unlike other
#' BigQuery objects, is no accompanying `bq_project` S3 class because a project
#' is a simple string.
#'
#' @section Google BigQuery API documentation:
#' * [datasets](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/datasets/list)
#' * [jobs](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/jobs/list)
#'
#' @return
#' * `bq_project_datasets()`: a list of [bq_dataset]s
#' * `bq_project_jobs()`: a list of [bq_job]s.
#'
#' @name api-project
#' @examplesIf bq_testable()
#' bq_project_datasets("bigquery-public-data")
#' bq_project_datasets("githubarchive")
#'
#' bq_project_jobs(bq_test_project(), page_size = 10)
NULL

#' @export
#' @rdname api-project
#' @param x A string giving a project name.
#' @inheritParams bq_projects
bq_project_datasets <- function(
  x,
  page_size = 100,
  max_pages = 1,
  warn = TRUE
) {
  check_string(x)
  check_number_whole(page_size, min = 1)
  check_number_whole(max_pages, min = 1, allow_infinite = TRUE)
  check_bool(warn)

  pages <- bq_get_paginated(
    bq_path(x, ""),
    query = list(fields = "datasets(datasetReference)"),
    page_size = page_size,
    max_pages = max_pages,
    warn = warn
  )

  datasets <- unlist(lapply(pages, function(x) x$datasets), recursive = FALSE)
  lapply(datasets, function(x) {
    ref <- x$datasetReference
    bq_dataset(ref$projectId, ref$datasetId)
  })
}

#' @export
#' @rdname api-project
bq_project_jobs <- function(x, page_size = 100, max_pages = 1, warn = TRUE) {
  check_string(x)
  check_number_whole(page_size, min = 1)
  check_number_whole(max_pages, min = 1, allow_infinite = TRUE)
  check_bool(warn)

  pages <- bq_get_paginated(
    bq_path(x, jobs = ""),
    query = list(fields = "jobs(jobReference)"),
    page_size = page_size,
    max_pages = max_pages,
    warn = warn
  )
  jobs <- unlist(lapply(pages, function(x) x$jobs), recursive = FALSE)
  lapply(jobs, function(x) as_bq_job(x$jobReference))
}

bq_parse_single <- function(value, type, ...) {
  field <- bq_field("", type, ...)
  field_j <- jsonlite::toJSON(as_json(field))
  value_j <- jsonlite::toJSON(value, auto_unbox = TRUE)

  bq_field_init(field_j, value_j)
}

v <- function(x) list(v = x)
vs <- function(...) lapply(list(...), v)
f <- function(...) list(f = list(...))

#' @include dbi-connection.R
NULL

BigQueryResult <- function(conn, sql, params = NULL, ...) {
  ds <- if (!is.null(conn@dataset)) as_bq_dataset(conn)
  job <- bq_perform_query(
    sql,
    billing = conn@billing,
    default_dataset = ds,
    quiet = conn@quiet,
    parameters = params,
    labels = conn@labels,
    ...
  )

  bq_job_wait(job, quiet = conn@quiet)
  meta <- bq_job_meta(
    job,
    paste0(
      "configuration(query(destinationTable)),",
      "statistics(query(statementType,numDmlAffectedRows))"
    )
  )

  tb <- as_bq_table(meta$configuration$query$destinationTable)

  if (meta$statistics$query$statementType == "SELECT") {
    nrow <- bq_table_nrow(tb)
  } else {
    nrow <- 0
  }

  affected <- as.numeric(meta$statistics$query$numDmlAffectedRows %||% 0)

  res <- new(
    "BigQueryResult",
    bq_table = tb,
    statement = sql,
    nrow = nrow,
    affected = affected,
    page_size = conn@page_size,
    quiet = conn@quiet,
    cursor = cursor(nrow),
    bigint = conn@bigint,
    billing = conn@billing
  )
  res
}

#' @rdname DBI
#' @export
setClass(
  "BigQueryResult",
  contains = "DBIResult",
  slots = list(
    bq_table = "bq_table",
    statement = "character",
    nrow = "numeric",
    affected = "numeric",
    page_size = "numeric",
    quiet = "logical",
    cursor = "list",
    bigint = "character",
    billing = "character"
  )
)

#' @rdname DBI
#' @inheritParams methods::show
#' @export
setMethod(
  "show",
  "BigQueryResult",
  function(object) {
    cat_line(
      "<BigQueryResult>\n",
      "  Query: ",
      dbGetStatement(object),
      "\n  Has completed: ",
      dbHasCompleted(object),
      "\n  Rows fetched: ",
      dbGetRowCount(object)
    )
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbIsValid
#' @export
setMethod(
  "dbIsValid",
  "BigQueryResult",
  function(dbObj, ...) TRUE
)

#' @rdname DBI
#' @inheritParams DBI::dbClearResult
#' @param res An object inheriting from [DBI::DBIResult-class].
#' @export
setMethod(
  "dbClearResult",
  "BigQueryResult",
  function(res, ...) {
    invisible(TRUE)
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbFetch
#' @export
setMethod(
  "dbFetch",
  "BigQueryResult",
  function(res, n = -1, ...) {
    check_number_whole(n, min = -1, allow_infinite = TRUE)
    if (n == -1) {
      n <- Inf
    }

    if (has_bigrquerystorage() && n == Inf && res@cursor$cur() == 0) {
      # https://github.com/meztez/bigrquerystorage/issues/48
      n <- res@cursor$left()

      # If possible, download complete dataset using arrow
      data <- bq_table_download(
        res@bq_table,
        n_max = n,
        bigint = res@bigint,
        quiet = res@quiet,
        api = "arrow"
      )
    } else {
      # Otherwise, fall back to slower JSON API
      data <- bq_table_download(
        res@bq_table,
        n_max = n,
        start_index = res@cursor$cur(),
        page_size = res@page_size,
        bigint = res@bigint,
        quiet = res@quiet,
        api = "json"
      )
    }

    res@cursor$adv(nrow(data))

    data
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbHasCompleted
#' @export
setMethod(
  "dbHasCompleted",
  "BigQueryResult",
  function(res, ...) {
    res@cursor$left() == 0
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbGetStatement
#' @export
setMethod(
  "dbGetStatement",
  "BigQueryResult",
  function(res, ...) {
    res@statement
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbColumnInfo
#' @export
setMethod(
  "dbColumnInfo",
  "BigQueryResult",
  function(res, ...) {
    fields <- bq_table_fields(res@bq_table)

    data.frame(
      name = vapply(fields, function(x) x$name, character(1)),
      type = vapply(fields, function(x) x$type, character(1)),
      stringsAsFactors = FALSE
    )
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbGetRowCount
#' @export
setMethod(
  "dbGetRowCount",
  "BigQueryResult",
  function(res, ...) {
    res@cursor$cur()
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbGetRowsAffected
#' @export
setMethod(
  "dbGetRowsAffected",
  "BigQueryResult",
  function(res, ...) {
    res@affected
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbBind
#' @export
setMethod(
  "dbBind",
  "BigQueryResult",
  function(res, params, ...) {
    testthat::skip("Not yet implemented: dbBind(Result)")
  }
)


cursor <- function(nrow) {
  pos <- 0

  list(
    cur = function() pos,
    adv = function(i) {
      pos <<- pos + i
    },
    left = function() nrow - pos
  )
}

#' BigQuery tables
#'
#' Basic create-read-update-delete verbs for tables, as well as functions
#' uploading data (`bq_table_upload()`), saving to/loading from Google
#' Cloud Storage (`bq_table_load()`, `bq_table_save()`), and getting
#' various values from the metadata.
#'
#' @param x A [bq_table], or an object coercible to a `bq_table`.
#' @inheritParams api-job
#' @inheritParams api-perform
#' @inheritParams bq_projects
#' @section Google BigQuery API documentation:
#' * [insert](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables/insert)
#' * [get](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables/get)
#' * [delete](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables/delete)
#' @return
#' * `bq_table_copy()`, `bq_table_create()`, `bq_table_delete()`, `bq_table_upload()`:
#'   an invisible [bq_table]
#' * `bq_table_exists()`: either `TRUE` or `FALSE`.
#' * `bq_table_size()`: the size of the table in bytes
#' * `bq_table_fields()`: a [bq_fields].
#'
#' @examplesIf bq_testable()
#' ds <- bq_test_dataset()
#'
#' bq_mtcars <- bq_table(ds, "mtcars")
#' bq_table_exists(bq_mtcars)
#'
#' bq_table_create(
#'   bq_mtcars,
#'   fields = mtcars,
#'   friendly_name = "Motor Trend Car Road Tests",
#'   description = "The data was extracted from the 1974 Motor Trend US magazine",
#'   labels = list(category = "example")
#' )
#' bq_table_exists(bq_mtcars)
#'
#' bq_table_upload(bq_mtcars, mtcars)
#'
#' bq_table_fields(bq_mtcars)
#' bq_table_size(bq_mtcars)
#' str(bq_table_meta(bq_mtcars))
#'
#' bq_table_delete(bq_mtcars)
#' bq_table_exists(bq_mtcars)
#'
#' my_natality <- bq_table(ds, "mynatality")
#' bq_table_copy("publicdata.samples.natality", my_natality)
#' @name api-table
NULL

#' @export
#' @rdname api-table
#' @param fields A [bq_fields] specification, or something coercible to it
#'   (like a data frame).
bq_table_create <- function(x, fields = NULL, ...) {
  x <- as_bq_table(x)

  url <- bq_path(x$project, x$dataset, "")
  body <- list(
    tableReference = tableReference(x)
  )
  if (!is.null(fields)) {
    fields <- as_bq_fields(fields)
    body$schema <- list(fields = as_json(fields))
  }

  bq_post(url, body = bq_body(body, ...))

  x
}

#' @export
#' @rdname api-table
#' @inheritParams api-job
bq_table_meta <- function(x, fields = NULL) {
  x <- as_bq_table(x)
  url <- bq_path(x$project, x$dataset, x$table)
  bq_get(url, query = list(fields = fields))
}

#' @export
#' @rdname api-table
bq_table_fields <- function(x) {
  meta <- bq_table_meta(x, fields = "schema")
  fields <- meta$schema$fields

  bq_fields(lapply(fields, as_bq_field))
}

#' @export
#' @rdname api-table
bq_table_size <- function(x) {
  meta <- bq_table_meta(x, fields = "numBytes")
  bytes <- as.numeric(meta$numBytes)
  structure(bytes, class = "bq_bytes")
}

#' @export
#' @rdname api-table
bq_table_nrow <- function(x) {
  meta <- bq_table_meta(x, fields = "numRows")
  as.numeric(meta$numRows)
}

#' @export
#' @rdname api-table
bq_table_exists <- function(x) {
  x <- as_bq_table(x)
  url <- bq_path(x$project, x$dataset, x$table)
  bq_exists(url)
}

#' @export
#' @rdname api-table
bq_table_delete <- function(x) {
  x <- as_bq_table(x)
  url <- bq_path(x$project, x$dataset, x$table)
  invisible(bq_delete(url))
}

#' @export
#' @rdname api-table
#' @inheritParams bq_perform_copy
#' @param dest Source and destination [bq_table]s.
bq_table_copy <- function(
  x,
  dest,
  ...,
  quiet = getOption("bigrquery.quiet", NA)
) {
  x <- as_bq_table(x)
  dest <- as_bq_table(dest)

  job <- bq_perform_copy(x, dest, ...)
  bq_job_wait(job, quiet = quiet)

  dest
}

#' @export
#' @rdname api-table
#' @inheritParams api-perform
bq_table_upload <- function(
  x,
  values,
  ...,
  quiet = getOption("bigrquery.quiet", NA),
  json_digits = NULL
) {
  x <- as_bq_table(x)
  json_digits <- check_digits(json_digits)

  job <- bq_perform_upload(x, values, ..., json_digits = json_digits)
  bq_job_wait(job, quiet = quiet)

  invisible(x)
}

#' @export
#' @rdname api-table
bq_table_save <- function(
  x,
  destination_uris,
  ...,
  quiet = getOption("bigrquery.quiet", NA)
) {
  x <- as_bq_table(x)

  job <- bq_perform_extract(x, destination_uris = destination_uris, ...)
  bq_job_wait(job, quiet = quiet)

  invisible(x)
}

#' @export
#' @rdname api-table
bq_table_load <- function(
  x,
  source_uris,
  ...,
  quiet = getOption("bigrquery.quiet", NA)
) {
  x <- as_bq_table(x)

  job <- bq_perform_load(x, source_uris = source_uris, ...)
  bq_job_wait(job, quiet = quiet)

  invisible(x)
}

#' @export
#' @rdname api-table
bq_table_patch <- function(x, fields) {
  x <- as_bq_table(x)

  url <- bq_path(x$project, x$dataset, x$table)
  body <- list(
    tableReference = tableReference(x)
  )
  fields <- as_bq_fields(fields)
  body$schema <- list(fields = as_json(fields))
  bq_patch(url, body)
}

test_that("can create and delete tables", {
  bq_mtcars <- bq_test_table()
  expect_false(bq_table_exists(bq_mtcars))

  bq_table_create(
    bq_mtcars,
    mtcars,
    friendly_name = "Motor Trend Car Road Tests",
    description = "The data was extracted from the 1974 Motor Trend US magazine",
    labels = list(category = "test")
  )
  expect_true(bq_table_exists(bq_mtcars))

  bq_table_delete(bq_mtcars)
  expect_false(bq_table_exists(bq_mtcars))
})

test_that("can retrieve table size information", {
  bq_mtcars <- bq_table(bq_test_project(), "basedata", "mtcars")
  expect_equal(bq_table_nrow(bq_mtcars), 32)
  expect_equal(as.numeric(bq_table_size(bq_mtcars)), 2816)
})

test_that("can create table with schema", {
  tb <- bq_test_table()

  df <- data.frame(x = 1, y = "a")
  bq_table_create(tb, "df", fields = df)

  fields <- bq_table_fields(tb)
  expect_equal(fields, as_bq_fields(df))
})

test_that("can round trip to non-default location", {
  dallas <- bq_test_dataset(location = "us-south1")
  df1 <- tibble(x = 1:10, y = letters[1:10])

  bq_df <- bq_table(dallas, "df")
  bq_table_upload(bq_df, df1)

  df2 <- bq_table_download(bq_df, api = "json")
  df2 <- df2[order(df2$x), names(df1)] # BQ doesn't guarantee order
  rownames(df2) <- NULL

  expect_equal(df1, df2)
})

test_that("can roundtrip via save + load", {
  tb1 <- bq_table(bq_test_project(), "basedata", "mtcars")
  tb2 <- bq_test_table()
  gs <- gs_test_object()

  bq_table_save(tb1, gs)
  defer(gs_object_delete(gs))
  bq_table_load(tb2, gs)

  df <- bq_table_download(tb2, api = "json")
  expect_equal(dim(df), c(32, 11))
})

test_that("can copy table from public dataset", {
  my_natality <- bq_test_table()

  out <- bq_table_copy("publicdata.samples.natality", my_natality)
  expect_equal(out, my_natality)
  expect_true(bq_table_exists(my_natality))
})

# data-types --------------------------------------------------------------

test_that("can round trip atomic vectors", {
  df1 <- tibble(
    lgl = c(FALSE, TRUE, NA),
    int = c(-1, 1, NA),
    dbl = c(-1.5, 1.5, NA),
    chr = c("A", "B", NA)
  )

  bq_df <- bq_test_table()
  bq_table_upload(bq_df, df1)

  df2 <- bq_table_download(bq_df, bigint = "integer", api = "json")
  df2 <- df2[order(df2[[1]]), names(df1)] # BQ doesn't gaurantee order
  rownames(df2) <- NULL

  expect_equal(df1, df2)
})

test_that("can round-trip POSIXt to either TIMESTAMP or DATETIME", {
  df <- tibble(datetime = as.POSIXct("2020-01-01 09:00", tz = "UTC"))

  tb1 <- bq_table_create(
    bq_test_table(),
    bq_fields(list(bq_field("datetime", "TIMESTAMP")))
  )
  bq_table_upload(tb1, df)
  df1 <- bq_table_download(tb1, api = "json")
  expect_equal(df1, df)

  tb2 <- bq_table_create(
    bq_test_table(),
    bq_fields(list(bq_field("datetime", "DATETIME")))
  )
  bq_table_upload(tb2, df)
  df2 <- bq_table_download(tb2, api = "json")
  expect_equal(df2, df)
})

test_that("can round trip data frame with list-cols", {
  tb <- bq_test_table()

  df1 <- tibble::tibble(
    val = 1.5,
    array = list(1L:5L),
    struct = list(list(x = "a", y = 1.5, z = 2L)),
    array_struct = list(tibble::tibble(x = "a", y = 1.5, z = 2L))
  )
  bq_table_upload(tb, df1)

  df2 <- bq_table_download(tb, bigint = "integer", api = "json")
  # restore column order
  df2 <- df2[names(df1)]
  df2$struct[[1]] <- df2$struct[[1]][c("x", "y", "z")]
  df2$array_struct[[1]] <- df2$array_struct[[1]][c("x", "y", "z")]

  # Converting to dataframe to avoid getting the error:
  # Can't join on 'array' x 'array' because of incompatible types (list / list)
  df1 <- as.data.frame(df1)
  df2 <- as.data.frame(df2)
  expect_equal(df1, df2)
})

test_that("can create table field description", {
  partition_table <- bq_test_table()

  bq_table_create(
    partition_table,
    fields = bq_fields(list(bq_field(
      "id",
      "integer",
      description = "Key field"
    )))
  )

  meta <- bq_table_meta(partition_table)
  expect_equal(meta$schema$fields[[1]]$description, "Key field")
})

test_that("can patch table with new fields in the schema", {
  tb <- bq_test_table()
  df <- data.frame(id = 1)
  bq_table_create(tb, fields = df)

  df.patch <- data.frame(id = 1, title = "record name")
  bq_table_patch(tb, fields = df.patch)

  tb.meta <- bq_table_meta(tb)
  expect_equal(
    tb.meta$schema$fields[[2]]$name,
    "title"
  )
})

test_that("can round-trip GEOGRAPHY", {
  skip_if_not_installed("wk")

  df <- tibble(geography = wk::wkt("POINT(30 10)"))

  tb1 <- bq_table_create(bq_test_table(), as_bq_fields(df))
  bq_table_upload(tb1, df)
  df1 <- bq_table_download(tb1, api = "json")
  expect_equal(df1, df)
})

test_that("can round-trip BYTES", {
  df <- tibble(x = blob::blob(charToRaw("hi!"), charToRaw("bye")))

  tb1 <- bq_table_create(bq_test_table(), as_bq_fields(df))
  bq_table_upload(tb1, df)
  df1 <- bq_table_download(tb1, api = "json")
  expect_equal(df1, df)
})

camelCase <- function(x) {
  gsub("_(.)", "\\U\\1", x, perl = TRUE)
}

toCamel <- function(x) {
  if (is.list(x)) {
    x[] <- lapply(x, toCamel)
  }

  if (!is.null(names(x))) {
    names(x) <- camelCase(names(x))
  }

  x
}

#' List available projects
#'
#' List all projects that you have access to. You can also work with
#' [public datasets](https://docs.cloud.google.com/bigquery/public-data/),
#' but you will need to provide a `billing` project whenever you perform
#' any non-free operation.
#'
#' @section Google BigQuery API documentation:
#' * [list](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/projects/list)
#' @param page_size Number of items per page.
#' @param max_pages Maximum number of pages to retrieve. Use `Inf` to retrieve
#'   all pages (this may take a long time!)
#' @param warn If `TRUE`, warn when there are unretrieved pages.
#' @return A character vector.
#' @export
#' @examplesIf bq_testable()
#' bq_projects()
bq_projects <- function(page_size = 100, max_pages = 1, warn = TRUE) {
  pages <- bq_get_paginated(
    "projects",
    query = list(fields = "projects(projectReference(projectId))"),
    page_size = page_size,
    max_pages = max_pages
  )
  projects <- unlist(lapply(pages, function(x) x$projects), recursive = FALSE)
  map_chr(projects, function(x) x$projectReference$projectId)
}

#' Submit query to BigQuery
#'
#' These submit a query (using [bq_perform_query()]) and then wait for it
#' complete (with [bq_job_wait()]). All BigQuery queries save their results
#' into a table (temporary or otherwise), so these functions return a [bq_table]
#' which you can then query for more information.
#'
#' @param x Either a project (a string) or a [bq_dataset].
#' @param billing If you query a dataset that you only have read access
#'   for, such as a public dataset, you must also submit a `billing` project.
#' @inheritParams bq_perform_query
#' @inheritParams api-job
#' @param ... Passed on to [bq_perform_query()]
#' @name bq_query
#' @return A [bq_table]
#' @examplesIf bq_testable()
#' # Querying a project requires full name in query
#' tb <- bq_project_query(
#'   bq_test_project(),
#'   "SELECT count(*) FROM publicdata.samples.natality"
#' )
#' bq_table_fields(tb)
#' bq_table_download(tb)
#'
#' # Querying a dataset sets default dataset so you can use bare table name,
#' # but for public data, you'll need to set a project to bill.
#' ds <- bq_dataset("publicdata", "samples")
#' tb <- bq_dataset_query(ds,
#'   query = "SELECT count(*) FROM natality",
#'   billing = bq_test_project()
#' )
#' bq_table_download(tb)
#'
#' tb <- bq_dataset_query(ds,
#'   query = "SELECT count(*) FROM natality WHERE state = @state",
#'   parameters = list(state = "KS"),
#'   billing = bq_test_project()
#' )
#' bq_table_download(tb)
NULL

#' @export
#' @rdname bq_query
bq_project_query <- function(
  x,
  query,
  destination_table = NULL,
  ...,
  quiet = getOption("bigrquery.quiet", NA)
) {
  check_string(x)
  query <- as_query(query)
  if (!is.null(destination_table)) {
    destination_table <- as_bq_table(destination_table)
  }
  check_bool(quiet, allow_na = TRUE)

  job <- bq_perform_query(
    query,
    billing = x,
    destination_table = destination_table,
    ...
  )
  bq_job_wait(job, quiet = quiet)
  bq_job_table(job)
}

#' @export
#' @rdname bq_query
bq_dataset_query <- function(
  x,
  query,
  destination_table = NULL,
  ...,
  billing = NULL,
  quiet = getOption("bigrquery.quiet", NA)
) {
  x <- as_bq_dataset(x)
  query <- as_query(query)
  if (!is.null(destination_table)) {
    destination_table <- as_bq_table(destination_table)
  }
  check_string(billing, allow_null = TRUE)
  check_bool(quiet, allow_na = TRUE)

  job <- bq_perform_query(
    query,
    billing = billing %||% x$project,
    destination_table = destination_table,
    default_dataset = x,
    ...
  )
  bq_job_wait(job, quiet = quiet)
  bq_job_table(job)
}

test_that("bq_project_query inputs are checked", {
  expect_snapshot(error = TRUE, {
    bq_project_query(1)
    bq_project_query("abc", 1)
    bq_project_query("abc", "SELECT *", destination_table = 1)
    bq_project_query("abc", "SELECT *", destination_table = "a")
    bq_project_query("abc", "SELECT *", destination_table = list())
    bq_project_query("abc", "SELECT *", quiet = 1)
  })
})

test_that("bq_dataset_query inputs are checked", {
  expect_snapshot(error = TRUE, {
    bq_dataset_query(1)
    bq_dataset_query("abc")
    bq_dataset_query("abc.def", 1)
    bq_dataset_query("abc.def", "SELECT *", destination_table = 1)
    bq_dataset_query("abc.def", "SELECT *", destination_table = "a")
    bq_dataset_query("abc.def", "SELECT *", destination_table = list())
    bq_dataset_query("abc.def", "SELECT *", billing = 1)
    bq_dataset_query("abc.def", "SELECT *", quiet = 1)
  })
})

# nocov start

# Capture connection expression for pane
connection_capture <- function() {
  if (is.null(getOption("connectionObserver"))) {
    return()
  }

  addTaskCallback(function(expr, ...) {
    tryCatch(
      {
        # notify if this is an assignment we can replay
        if (is_call(expr, c("<-", "=")) && is_call(expr[[3]], "dbConnect")) {
          on_connection_opened(
            eval(expr[[2]]),
            paste(c("library(bigrquery)", deparse(expr)), collapse = "\n")
          )
        }
      },
      error = function(e) {
        warning(
          "Could not notify connection observer. ",
          e$message,
          call. = FALSE
        )
      }
    )

    # always return false so the task callback is run at most once
    FALSE
  })
}

# https://rstudio.github.io/rstudio-extensions/connections-contract.html#connection-closed
on_connection_closed <- function(con) {
  observer <- getOption("connectionObserver")
  if (is.null(observer)) {
    return(invisible(NULL))
  }

  observer$connectionClosed(bq_type, con@project)
}

# https://rstudio.github.io/rstudio-extensions/connections-contract.html#connection-updated
on_connection_updated <- function(con, hint) {
  observer <- getOption("connectionObserver")
  if (is.null(observer)) {
    return(invisible(NULL))
  }

  observer$connectionUpdated(bq_type, con@project, hint = hint)
}

# https://rstudio.github.io/rstudio-extensions/connections-contract.html#connection-opened
on_connection_opened <- function(con, code) {
  observer <- getOption("connectionObserver")
  if (is.null(observer)) {
    return(invisible(NULL))
  }

  observer$connectionOpened(
    type = bq_type,
    displayName = paste0(bq_type, ": ", con@project),
    host = con@project,
    icon = system.file("icons/bigquery-512-color.png", package = "bigrquery"),

    # connection code
    connectCode = code,

    # only action is to close connections pane
    disconnect = function() {},

    listObjectTypes = function() {
      list(
        dataset = list(
          icon = system.file("icons/dataset.png", package = "bigrquery"),
          contains = list(
            table = list(
              icon = system.file("icons/table.png", package = "bigrquery"),
              contains = "data"
            ),
            view = list(
              icon = system.file("icons/view.png", package = "bigrquery"),
              contains = "data"
            )
          )
        )
      )
    },

    # table enumeration code
    listObjects = function(...) {
      list_bigquery_objects(con, ...)
    },

    # column enumeration code
    listColumns = function(
      dataset = con@dataset,
      table = NULL,
      view = NULL,
      ...
    ) {
      x <- bq_table(con@project, dataset, paste0(table, view))
      fields <- bq_table_fields(x)

      tibble::tibble(
        name = vapply(fields, `[[`, character(1), "name"),
        type = vapply(fields, `[[`, character(1), "type")
      )
    },

    # table preview code
    previewObject = function(
      rowLimit,
      dataset = con@dataset,
      table = NULL,
      view = NULL,
      ...
    ) {
      x <- bq_table(con@project, dataset, paste0(table, view))
      bq_table_download(x, n_max = rowLimit)
    },

    # no actions

    # raw connection object
    connectionObject = con
  )
}

list_bigquery_objects <- function(con, dataset = con@dataset, ...) {
  if (is.null(dataset)) {
    # Catching VPC/Permission errors to crash gracefully
    bq_datasets <- tryCatch(
      bq_project_datasets(con@project, warn = FALSE),
      error = function(e) list()
    )
    datasets <- map_chr(bq_datasets, `[[`, "dataset")

    tibble::tibble(type = "dataset", name = datasets)
  } else {
    # Catching VPC/Permission errors to crash gracefully
    ds <- bq_dataset(con@project, dataset)
    bq_tables <- tryCatch(bq_dataset_tables(ds), error = function(e) list())
    tables <- map_chr(bq_tables, `[[`, "table")
    types <- map_chr(bq_tables, `[[`, "type")
    types <- grepl("VIEW$", types) + 1L
    types <- c("table", "view")[types]
    tibble::tibble(type = types, name = tables)
  }
}

bq_type <- "BigQuery"

# nocov end

#' BigQuery jobs: perform a job
#'
#' @description
#' These functions are low-level functions designed to be used by experts.
#' Each of these low-level functions is paired with a high-level function that
#' you should use instead:
#'
#' * `bq_perform_copy()`:    [bq_table_copy()].
#' * `bq_perform_query()`:   [bq_dataset_query()], [bq_project_query()].
#' * `bq_perform_upload()`:  [bq_table_upload()].
#' * `bq_perform_load()`:    [bq_table_load()].
#' * `bq_perform_extract()`: [bq_table_save()].
#'
#' @section Google BigQuery API documentation:
#' * [jobs](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/jobs)
#'
#' Additional information at:
#' * [exporting data](https://docs.cloud.google.com/bigquery/docs/exporting-data)
#' * [loading data](https://docs.cloud.google.com/bigquery/docs/loading-data)
#' * [writing queries](https://docs.cloud.google.com/bigquery/docs/writing-results)
#' * [copying a table](https://docs.cloud.google.com/bigquery/docs/managing-tables#copy-table)
#'
#' @return A [bq_job].
#' @keywords internal
#' @examplesIf bq_testable()
#' ds <- bq_test_dataset()
#' bq_mtcars <- bq_table(ds, "mtcars")
#' job <- bq_perform_upload(bq_mtcars, mtcars)
#' bq_table_exists(bq_mtcars)
#'
#' bq_job_wait(job)
#' bq_table_exists(bq_mtcars)
#' head(bq_table_download(bq_mtcars))
#' @name api-perform
NULL

#' @export
#' @name api-perform
#' @param x A [bq_table]
#' @param destination_uris A character vector of fully-qualified Google Cloud
#'   Storage URIs where the extracted table should be written. Can export
#'   up to 1 Gb of data per file. Use a wild card URI (e.g.
#'   `gs://[YOUR_BUCKET]/file-name-*.json`) to automatically create any
#'   number of files.
#' @param destination_format The exported file format:
#'   * For CSV files, specify "CSV" (Nested and repeated data is not supported).
#'   * For newline-delimited JSON, specify "NEWLINE_DELIMITED_JSON".
#'   * For Avro, specify "AVRO".
#'   * For parquet, specify "PARQUET".
#' @param compression The compression type to use for exported files:
#'   * For CSV files: "GZIP" or "NONE".
#'   * For newline-delimited JSON: "GZIP" or "NONE".
#'   * For Avro: "DEFLATE", "SNAPPY" or "NONE".
#'   * For parquet: "SNAPPY", "GZIP", "ZSTD" or "NONE".
#' @param ... Additional arguments passed on to the underlying API call.
#'   snake_case names are automatically converted to camelCase.
#' @param print_header Whether to print out a header row in the results.
#' @param billing Identifier of project to bill.
#' @param labels A named list of strings used to attach
#'   [BigQuery labels](https://docs.cloud.google.com/bigquery/docs/labels-intro)
#'   to the resulting job, e.g. `list(env = "prod", team = "data")`. This
#'   is most useful for cost allocation and other FinOps reporting.
#'   Defaults to the value of `getOption("bigrquery.labels")`.
bq_perform_extract <- function(
  x,
  destination_uris,
  destination_format = "NEWLINE_DELIMITED_JSON",
  compression = "NONE",
  ...,
  print_header = TRUE,
  billing = x$project,
  labels = getOption("bigrquery.labels")
) {
  x <- as_bq_table(x)
  destination_uris <- as.character(destination_uris) # for gs_object
  check_string(destination_format)
  check_string(compression)
  check_bool(print_header)
  check_string(billing)
  check_labels(labels)

  url <- bq_path(billing, jobs = "")
  body <- list(
    configuration = list(
      extract = list(
        sourceTable = tableReference(x),
        destinationUris = as.list(destination_uris),
        destinationFormat = unbox(destination_format),
        compression = unbox(compression),
        printHeader = unbox(print_header)
      ),
      labels = labels
    )
  )

  res <- bq_post(
    url,
    body = bq_body(body, ...),
    query = list(field = "jobReference")
  )
  as_bq_job(res$jobReference)
}

#' @export
#' @name api-perform
#' @param values Data frame of values to insert.
#' @param source_format The format of the data files:
#'   * For newline-delimited JSON, specify "NEWLINE_DELIMITED_JSON".
#'   * For parquet, specify "PARQUET".
#' @param create_disposition Specifies whether the job is allowed to create
#'   new tables.
#'
#'   The following values are supported:
#'   * "CREATE_IF_NEEDED": If the table does not exist, BigQuery creates the
#'     table.
#'   * "CREATE_NEVER": The table must already exist. If it does not, a
#'     'notFound' error is returned in the job result.
#' @param write_disposition Specifies the action that occurs if the
#'   destination table already exists. The following values are supported:
#'
#'   * "WRITE_TRUNCATE": If the table already exists, BigQuery overwrites the
#'     table data.
#'   * "WRITE_APPEND": If the table already exists, BigQuery appends the data
#'     to the table.
#'   * "WRITE_EMPTY": If the table already exists and contains data, a
#'     'duplicate' error is returned in the job result.
#' @param json_digits Species the number of digits for formatting
#'   numeric values.
bq_perform_upload <- function(
  x,
  values,
  fields = NULL,
  source_format = c("NEWLINE_DELIMITED_JSON", "PARQUET"),
  create_disposition = "CREATE_IF_NEEDED",
  write_disposition = "WRITE_EMPTY",
  ...,
  billing = x$project,
  json_digits = NULL,
  labels = getOption("bigrquery.labels")
) {
  x <- as_bq_table(x)
  if (!is.data.frame(values)) {
    cli::cli_abort("{.arg values} must be a data frame.")
  }
  fields <- as_bq_fields(fields)
  source_format <- arg_match(source_format)
  check_string(create_disposition)
  check_string(write_disposition)
  check_string(billing)
  json_digits <- check_digits(json_digits)
  check_labels(labels)

  load <- list(
    sourceFormat = unbox(source_format),
    destinationTable = tableReference(x),
    createDisposition = unbox(create_disposition),
    writeDisposition = unbox(write_disposition)
  )

  if (!is.null(fields)) {
    load$schema <- list(fields = as_json(fields))
  }
  if (!bq_table_exists(x)) {
    load$autodetect <- unbox(TRUE)
  }

  metadata <- list(
    configuration = list(
      load = load,
      labels = labels
    )
  )
  metadata <- bq_body(metadata, ...)
  metadata <- list(
    "type" = "application/json; charset=UTF-8",
    "content" = jsonlite::toJSON(
      metadata,
      auto_unbox = TRUE,
      pretty = TRUE,
      digits = json_digits
    )
  )

  if (source_format == "NEWLINE_DELIMITED_JSON") {
    media <- list(
      "type" = "application/json; charset=UTF-8",
      "content" = export_json(values, json_digits = json_digits)
    )
  } else {
    # https://docs.cloud.google.com/bigquery/docs/loading-data-cloud-storage-parquet?hl=es-419
    media <- list(
      "type" = "application/vnd.apache.parquet",
      "content" = nanoparquet::write_parquet(values, ":raw:")
    )
  }

  url <- bq_path(billing, jobs = "")
  res <- bq_upload(
    url,
    metadata,
    media,
    query = list(fields = "jobReference")
  )
  as_bq_job(res$jobReference)
}

# https://docs.cloud.google.com/bigquery/docs/loading-data-cloud-storage-json#details_of_loading_json_data
export_json <- function(values, json_digits = NULL) {
  # Eliminate row names
  rownames(values) <- NULL

  # Convert times to canonical format
  is_time <- vapply(values, function(x) inherits(x, "POSIXt"), logical(1))
  values[is_time] <- lapply(values[is_time], function(x) {
    format(x, "%Y-%m-%d %H:%M:%OS6", tz = "UTC")
  })

  # Convert wk_wkt to text
  is_wk <- vapply(values, function(x) inherits(x, "wk_vctr"), logical(1))
  values[is_wk] <- lapply(values[is_wk], as.character)

  # Unbox blobs
  is_blob <- vapply(values, function(x) inherits(x, "blob"), logical(1))
  values[is_blob] <- lapply(values[is_blob], function(x) {
    vapply(x, jsonlite::base64_enc, character(1))
  })

  con <- rawConnection(raw(0), "r+")
  defer(close(con))
  jsonlite::stream_out(
    values,
    con,
    verbose = FALSE,
    na = "null",
    digits = json_digits
  )

  rawToChar(rawConnectionValue(con))
}

#' @export
#' @name api-perform
#' @param source_uris The fully-qualified URIs that point to your data in
#'   Google Cloud.
#'
#'   For Google Cloud Storage URIs: Each URI can contain one
#'   `'*'` wildcard character and it must come after the 'bucket' name.
#'   Size limits related to load jobs apply to external data sources.
#'
#'   For Google Cloud Bigtable URIs: Exactly one URI can be specified and
#'   it has be a fully specified and valid HTTPS URL for a Google Cloud
#'    Bigtable table. For Google Cloud Datastore backups: Exactly one URI
#'  can be specified. Also, the '*' wildcard character is not allowed.
#' @param source_format The format of the data files:
#'   * For CSV files, specify "CSV".
#'   * For datastore backups, specify "DATASTORE_BACKUP".
#'   * For newline-delimited JSON, specify "NEWLINE_DELIMITED_JSON".
#'   * For Avro, specify "AVRO".
#'   * For parquet, specify "PARQUET".
#'   * For orc, specify "ORC".
#' @param fields A [bq_fields] specification, or something coercible to it
#'   (like a data frame). Leave as `NULL` to allow BigQuery to auto-detect
#'   the fields.
#' @param nskip For `source_format = "CSV"`, the number of header rows to skip.
bq_perform_load <- function(
  x,
  source_uris,
  billing = x$project,
  source_format = "NEWLINE_DELIMITED_JSON",
  fields = NULL,
  nskip = 0,
  create_disposition = "CREATE_IF_NEEDED",
  write_disposition = "WRITE_EMPTY",
  ...,
  labels = getOption("bigrquery.labels")
) {
  x <- as_bq_table(x)
  source_uris <- as.character(source_uris)
  check_string(billing)
  check_string(source_format)
  check_number_decimal(nskip, min = 0)
  check_string(create_disposition)
  check_string(write_disposition)
  check_labels(labels)

  load <- list(
    sourceUris = as.list(source_uris),
    sourceFormat = unbox(source_format),
    destinationTable = tableReference(x),
    createDisposition = unbox(create_disposition),
    writeDisposition = unbox(write_disposition)
  )

  if (source_format == "CSV") {
    load$skipLeadingRows <- nskip
  }

  if (!is.null(fields)) {
    fields <- as_bq_fields(fields)
    load$schema <- list(fields = as_json(fields))
  } else {
    load$autodetect <- TRUE
  }

  body <- list(
    configuration = list(
      load = load,
      labels = labels
    )
  )

  url <- bq_path(billing, jobs = "")
  res <- bq_post(
    url,
    body = bq_body(body, ...),
    query = list(fields = "jobReference")
  )
  as_bq_job(res$jobReference)
}

#' @export
#' @rdname api-perform
#' @param query SQL query string.
#' @param parameters Named list of parameters match to query parameters.
#'   Parameter `x` will be matched to placeholder `@x`.
#'
#'   Generally, you can supply R vectors and they will be automatically
#'   converted to the correct type. If you need greater control, you can call
#'   [bq_param_scalar()] or [bq_param_array()] explicitly.
#'
#'   See <https://docs.cloud.google.com/bigquery/docs/parameterized-queries>
#'   for more details.
#' @param destination_table A [bq_table] where results should be stored.
#'   If not supplied, results will be saved to a temporary table that lives
#'   in a special dataset. You must supply this parameter for large
#'   queries (> 128 MB compressed).
#' @param priority Specifies a priority for the query. Possible values include
#'   "INTERACTIVE" and "BATCH". Batch queries do not start immediately,
#'   but are not rate-limited in the same way as interactive queries.
#' @param default_dataset A [bq_dataset] used to automatically qualify table names.
#' @param use_legacy_sql If `TRUE` will use BigQuery's legacy SQL format.
bq_perform_query <- function(
  query,
  billing,
  ...,
  parameters = NULL,
  destination_table = NULL,
  default_dataset = NULL,
  create_disposition = "CREATE_IF_NEEDED",
  write_disposition = "WRITE_EMPTY",
  use_legacy_sql = FALSE,
  priority = "INTERACTIVE",
  labels = getOption("bigrquery.labels")
) {
  query <- as_query(query)
  check_string(billing)

  check_string(create_disposition)
  check_string(write_disposition)
  check_bool(use_legacy_sql)
  check_string(priority)
  check_labels(labels)

  query <- list(
    query = unbox(query),
    useLegacySql = unbox(use_legacy_sql),
    priority = unbox(priority)
  )

  if (length(parameters) > 0) {
    parameters <- as_bq_params(parameters)
    query$queryParameters <- as_json(parameters)
  }

  if (!is.null(destination_table)) {
    query$destinationTable <- tableReference(destination_table)
    query$createDisposition <- unbox(create_disposition)
    query$writeDisposition <- unbox(write_disposition)
    if (use_legacy_sql) {
      query$allowLargeResults <- unbox(TRUE)
    }
  }

  if (!is.null(default_dataset)) {
    query$defaultDataset <- datasetReference(default_dataset)
  }

  url <- bq_path(billing, jobs = "")
  body <- list(
    configuration = list(
      query = query,
      labels = labels
    )
  )

  res <- bq_post(
    url,
    body = bq_body(body, ...),
    query = list(fields = "jobReference")
  )
  as_bq_job(res$jobReference)
}

#' @export
#' @rdname api-perform
bq_perform_query_dry_run <- function(
  query,
  billing,
  ...,
  default_dataset = NULL,
  parameters = NULL,
  use_legacy_sql = FALSE,
  labels = getOption("bigrquery.labels")
) {
  query <- bq_perform_query_data(
    query = query,
    default_dataset = default_dataset,
    parameters = parameters,
    use_legacy_sql = use_legacy_sql
  )
  check_labels(labels)

  url <- bq_path(billing, jobs = "")
  body <- list(
    configuration = list(
      query = query,
      labels = labels,
      dryRun = unbox(TRUE)
    )
  )

  res <- bq_post(
    url,
    body = bq_body(body, ...),
    query = list(fields = "statistics")
  )
  bytes <- as.numeric(res$statistics$query$totalBytesProcessed)
  structure(bytes, class = "bq_bytes")
}

#' @export
#' @rdname api-perform
bq_perform_query_schema <- function(
  query,
  billing,
  ...,
  default_dataset = NULL,
  parameters = NULL,
  labels = getOption("bigrquery.labels")
) {
  query <- bq_perform_query_data(
    query = query,
    default_dataset = default_dataset,
    parameters = parameters,
    use_legacy_sql = FALSE
  )
  check_labels(labels)

  url <- bq_path(billing, jobs = "")
  body <- list(
    configuration = list(
      query = query,
      labels = labels,
      dryRun = unbox(TRUE)
    )
  )

  res <- bq_post(
    url,
    body = bq_body(body, ...),
    query = list(fields = "statistics")
  )
  # https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables#TableSchema
  res$statistics$query$schema$fields
}

bq_perform_query_data <- function(
  query,
  ...,
  default_dataset = NULL,
  parameters = NULL,
  use_legacy_sql = FALSE,
  call = caller_env()
) {
  check_string(query, error_call = call)
  check_bool(use_legacy_sql, error_call = call)

  query <- list(
    query = unbox(query),
    useLegacySql = unbox(use_legacy_sql)
  )
  if (!is.null(parameters)) {
    parameters <- as_bq_params(parameters)
    query$queryParameters <- as_json(parameters)
  }
  if (!is.null(default_dataset)) {
    query$defaultDataset <- datasetReference(default_dataset)
  }

  query
}


#' @export
#' @rdname api-perform
bq_perform_copy <- function(
  src,
  dest,
  create_disposition = "CREATE_IF_NEEDED",
  write_disposition = "WRITE_EMPTY",
  ...,
  billing = NULL,
  labels = getOption("bigrquery.labels")
) {
  billing <- billing %||% dest$project
  url <- bq_path(billing, jobs = "")
  check_labels(labels)

  body <- list(
    configuration = list(
      copy = list(
        sourceTable = tableReference(src),
        destinationTable = tableReference(dest),
        createDisposition = unbox(create_disposition),
        writeDisposition = unbox(write_disposition)
      ),
      labels = labels
    )
  )

  res <- bq_post(
    url,
    body = bq_body(body, ...),
    query = list(fields = "jobReference")
  )
  as_bq_job(res$jobReference)
}

test_that("bq_perform_upload creates job that succeeds", {
  withr::local_options(cli.progress_show_after = Inf)

  bq_mtcars <- bq_test_table()
  job <- bq_perform_upload(bq_mtcars, mtcars)

  expect_s3_class(job, "bq_job")
  expect_snapshot(
    {
      bq_job_wait(job, quiet = FALSE)
      bq_job_wait(job, quiet = FALSE)
    },
    transform = \(x) gsub("[0-9.]+ kB", "<xxx> kB", x)
  )

  expect_true(bq_table_exists(bq_mtcars))
})

test_that("bq_perform_upload preserves numerical precision", {
  tb <- bq_test_table()
  bq_table_upload(tb, data.frame(dbl = pi))
  df <- bq_table_download(tb)
  expect_identical(df$dbl, pi)
})

test_that("bq_perform_upload preserves microsecond precision", {
  tbl <- data.frame(psx = .POSIXct(0.123456))
  attr(tbl$psx, "tzone") <- "UTC"

  ds <- bq_test_table()
  bq_table_upload(ds, tbl)
  tbl2 <- bq_table_download(ds)

  expect_equal(tbl$psx, tbl2$psx)
})

test_that("bq_perform_upload correclty assigns tzone", {
  # the bug only produced problems when the system TZ has a non-zero
  # offset (i.e., not UTC)
  withr::local_envvar(TZ = "America/New_York")

  ds <- bq_test_table()
  df <- data.frame(
    null = .POSIXct(0, tz = NULL),
    utc = .POSIXct(0, tz = "UTC"),
    ny = .POSIXct(0, tz = "America/New_York"),
    nz = .POSIXct(0, tz = "Pacific/Auckland")
  )
  bq_table_upload(ds, df)
  db <- bq_table_download(ds)

  expect_equal(db$utc, .POSIXct(0, tz = "UTC"))
  expect_equal(db$null, .POSIXct(0, tz = "UTC"))
  expect_equal(db$ny, .POSIXct(0, tz = "UTC"))
  expect_equal(db$nz, .POSIXct(0, tz = "UTC"))
})

test_that("bq_perform_copy creates job that succeeds", {
  withr::local_options(cli.progress_show_after = Inf)

  src <- as_bq_table("bigquery-public-data.moon_phases.moon_phases")
  dst <- bq_test_table()

  job <- bq_perform_copy(src, dst)
  expect_s3_class(job, "bq_job")

  # Doesn't return any statistics to show
  expect_snapshot({
    bq_job_wait(job, quiet = FALSE)
  })

  expect_true(bq_table_exists(dst))
})


# Load / extract ----------------------------------------------------------

test_that("can round trip extract + load", {
  ds_public <- bq_dataset("bigquery-public-data", "moon_phases")

  tb <- bq_dataset_query(
    ds_public,
    query = "SELECT COUNT(*) as count FROM moon_phases",
    billing = bq_test_project()
  )

  tmp <- gs_test_object()
  # on.exit(gs_object_delete(tmp))

  job <- bq_perform_extract(tb, tmp)
  bq_job_wait(job)

  tb_ks <- bq_test_table()
  job <- bq_perform_load(tb_ks, tmp)
  bq_job_wait(job)

  df <- bq_table_download(tb_ks)
  expect_equal(nrow(df), 1)
  expect_named(df, "count")
})

# Queries -----------------------------------------------------------------

test_that("bq_perform_query creates job that succeeds", {
  ds <- as_bq_dataset("bigquery-public-data.moon_phases")
  job <- bq_perform_query(
    "SELECT count(*) FROM moon_phases",
    billing = bq_test_project(),
    default_dataset = ds
  )

  expect_s3_class(job, "bq_job")
  expect_snapshot({
    bq_job_wait(job, quiet = FALSE)
  })

  job_tb <- bq_job_table(job)
  expect_true(bq_table_exists(job_tb))
})

test_that("can supply scalar parameters", {
  job <- bq_project_query(
    bq_test_project(),
    "SELECT 1 + @x",
    parameters = list(x = bq_param_scalar(1))
  )
  df <- bq_table_download(job)
  expect_setequal(df[[1]], 2)
})

test_that("can supply array parameters", {
  job <- bq_project_query(
    bq_test_project(),
    "SELECT values FROM UNNEST(@x) values",
    parameters = list(x = bq_param_array(c("a", "b")))
  )
  df <- bq_table_download(job)
  expect_setequal(df$values, c("a", "b"))
})

test_that("can estimate cost and get schema", {
  cost <- bq_perform_query_dry_run(
    "SELECT count(*) FROM bigquery-public-data.moon_phases.moon_phases",
    billing = bq_test_project()
  )
  expect_equal(cost, structure(0, class = "bq_bytes"))

  schema <- bq_perform_query_schema(
    "SELECT * FROM bigquery-public-data.moon_phases.moon_phases",
    billing = bq_test_project()
  )
  names <- vapply(schema, function(x) x$name, character(1))
  expect_equal(names, c("phase", "phase_emoji", "peak_datetime"))

  types <- vapply(schema, function(x) x$type, character(1))
  expect_equal(types, c("STRING", "STRING", "DATETIME"))
})

#' DBI methods
#'
#' Implementations of pure virtual functions defined in the `DBI` package.
#' @name DBI
#' @keywords internal
NULL

#' BigQuery DBI driver
#'
#' Creates a BigQuery DBI driver for use in [DBI::dbConnect()].
#'
#' @export
#' @usage NULL
#' @import methods DBI
#' @examplesIf bq_testable()
#' con <- DBI::dbConnect(
#'   bigquery(),
#'   project = "publicdata",
#'   dataset = "samples",
#'   billing = bq_test_project()
#' )
#' con
#' DBI::dbListTables(con)
#' DBI::dbReadTable(con, "natality", n_max = 10)
#'
#' # Create a temporary dataset to explore
#' ds <- bq_test_dataset()
#' con <- DBI::dbConnect(
#'   bigquery(),
#'   project = ds$project,
#'   dataset = ds$dataset
#' )
#' DBI::dbWriteTable(con, "mtcars", mtcars)
#' DBI::dbReadTable(con, "mtcars")[1:6, ]
#'
#' DBI::dbGetQuery(con, "SELECT count(*) FROM mtcars")
#'
#' res <- DBI::dbSendQuery(con, "SELECT cyl, mpg FROM mtcars")
#' dbColumnInfo(res)
#' dbFetch(res, 10)
#' dbFetch(res, -1)
#' DBI::dbHasCompleted(res)
bigquery <- function() {
  new("BigQueryDriver")
}

#' @export
#' @rdname bigquery
#' @usage NULL
dbi_driver <- function() {
  warning(
    "`dbi_driver()` deprecated; please use `bigquery()` instead`",
    call. = FALSE
  )
  new("BigQueryDriver")
}

#' @rdname DBI
#' @export
setClass("BigQueryDriver", contains = "DBIDriver")

#' @rdname bigquery
#' @inheritParams DBI::dbConnect
#' @param drv The result of `bigquery()`.
#' @param project,dataset Project and dataset identifiers
#' @inheritParams bq_perform_query
#' @inheritParams bq_projects
#' @inheritParams api-job
#' @param bigint The R type that BigQuery's 64-bit integer types should be mapped to.
#'   The default is `"integer"` which returns R's `integer` type but results in `NA` for
#'   values above/below +/- 2147483647. `"integer64"` returns a [bit64::integer64],
#'   which allows the full range of 64 bit integers.
#' @param ... Other arguments for compatibility with generic; currently ignored.
#' @export
setMethod(
  "dbConnect",
  "BigQueryDriver",
  function(
    drv,
    project,
    dataset = NULL,
    billing = project,
    page_size = 1e4,
    quiet = NA,
    use_legacy_sql = FALSE,
    bigint = c("integer", "integer64", "numeric", "character"),
    labels = getOption("bigrquery.labels"),
    ...
  ) {
    check_string(project)
    check_string(dataset, allow_null = TRUE)
    check_string(billing)
    check_number_whole(page_size, min = 1)
    check_bool(quiet, allow_na = TRUE)
    check_bool(use_legacy_sql)
    bigint <- arg_match(bigint)
    check_labels(labels, call = quote(dbConnect()))

    BigQueryConnection(
      project = project,
      dataset = dataset,
      billing = billing,
      page_size = page_size,
      quiet = quiet,
      use_legacy_sql = use_legacy_sql,
      bigint = bigint,
      labels = labels
    )
  }
)


#' @rdname DBI
#' @export
setMethod(
  "dbConnect",
  "bq_dataset",
  function(drv, ...) {
    DBI::dbConnect(
      bigquery(),
      project = drv$project,
      dataset = drv$dataset,
      ...
    )
  }
)


# Included for DBI compatibility ------------------------------------------
# nocov start
#' @rdname DBI
#' @inheritParams methods::show
#' @export
setMethod(
  "show",
  "BigQueryDriver",
  function(object) {
    cat("<BigQueryDriver>\n")
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbGetInfo
#' @export
setMethod(
  "dbGetInfo",
  "BigQueryDriver",
  function(dbObj, ...) {
    list(
      driver.version = PACKAGE_VERSION,
      client.version = NA,
      max.connections = NA
    )
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbIsValid
#' @export
setMethod(
  "dbIsValid",
  "BigQueryDriver",
  function(dbObj, ...) {
    TRUE
  }
)

#' @rdname DBI
#' @inheritParams DBI::dbDataType
#' @export
setMethod(
  "dbDataType",
  "BigQueryDriver",
  function(dbObj, obj, ...) {
    data_type(obj)
  }
)
# nocov end

test_that("driver is always valid", {
  expect_true(DBI::dbIsValid(bigquery()))
})

test_that("dbi_driver is deprecated", {
  expect_warning(dbi_driver(), "deprecated")
})

test_that("connecting yields a BigQueryConnection", {
  con <- dbConnect(bigquery(), project = bq_test_project())
  expect_s4_class(con, "BigQueryConnection")
})

test_that("dbConnect() captures labels", {
  con <- dbConnect(
    bigquery(),
    project = bq_test_project(),
    labels = list(env = "test")
  )
  expect_equal(con@labels, list(env = "test"))
})

test_that("dbConnect() validates labels", {
  expect_snapshot(
    error = TRUE,
    dbConnect(bigquery(), project = bq_test_project(), labels = "oops")
  )
})

test_that("dbConnect() reads bigrquery.labels option", {
  withr::local_options(bigrquery.labels = list(env = "from-option"))
  con <- dbConnect(bigquery(), project = bq_test_project())
  expect_equal(con@labels, list(env = "from-option"))
})

#' BigQuery job: retrieve metadata
#'
#' To perform a job, see [api-perform]. These functions all retrieve metadata
#' (in various forms) about an existing job.
#'
#' @section Google BigQuery API documentation:
#' * [get](https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/jobs/get)
#'
#' @examplesIf bq_testable()
#' jobs <- bq_project_jobs(bq_test_project())
#' jobs[[1]]
#'
#' # Show statistics about job
#' bq_job_show_statistics(jobs[[1]])
#'
#' # Wait for job to complete
#' bq_job_wait(jobs[[1]])
#' @name api-job
NULL

#' @export
#' @name api-job
#' @param x A [bq_job]
#' @param fields An optional field specification for
#'   [partial response](https://docs.cloud.google.com/bigquery/docs/api-performance#partial-response)
bq_job_meta <- function(x, fields = NULL) {
  x <- as_bq_job(x)
  bq_get(
    bq_path(x$project, jobs = x$job),
    query = list(
      location = x$location,
      fields = fields
    )
  )
}

bq_job_table <- function(x) {
  meta <- bq_job_meta(x, "configuration(query(destinationTable))")
  as_bq_table(meta$configuration$query$destinationTable)
}

#' @export
#' @name api-job
bq_job_status <- function(x) {
  bq_job_meta(x, "status")$status
}

#' @export
#' @name api-job
bq_job_show_statistics <- function(x) {
  stats <- bq_job_meta(x, "statistics")$statistics

  if ("load" %in% names(stats)) {
    in_bytes <- as.numeric(stats$load$inputFileBytes)
    out_bytes <- as.numeric(stats$load$outputBytes)
    cli::cli_inform("Input: {prettyunits::pretty_bytes(in_bytes)}")
    cli::cli_inform("Output: {prettyunits::pretty_bytes(out_bytes)}")
  }

  if ("query" %in% names(stats)) {
    bytes <- as.numeric(stats$query$totalBytesBilled)
    cli::cli_inform("Billed: {prettyunits::pretty_bytes(bytes)}")
  }

  invisible(x)
}

#' @param quiet If `FALSE`, displays progress bar; if `TRUE` is silent;
#'   if `NA` picks based on whether or not you're in an interactive context.
#' @param pause amount of time to wait between status requests
#' @export
#' @name api-job
#' @inheritParams rlang::args_error_context
bq_job_wait <- function(
  x,
  quiet = getOption("bigrquery.quiet"),
  pause = 0.5,
  call = caller_env()
) {
  x <- as_bq_job(x)
  quiet <- check_quiet(quiet)
  check_number_decimal(pause)

  if (!quiet) {
    cli::cli_progress_bar(
      format = "Running job {x} {cli::pb_spin} {cli::pb_elapsed}",
      total = NA,
      clear = FALSE
    )
  }

  repeat {
    if (!quiet) {
      cli::cli_progress_update()
    }
    # https://docs.cloud.google.com/bigquery/docs/error-messages
    # Switch to req_retry() when we move to httr2
    status <- tryCatch(
      bq_job_status(x),
      bigrquery_http_503 = function(err) NULL
    )
    if (!quiet) {
      cli::cli_progress_update()
    }

    if (!is.null(status) && status$state == "DONE") {
      break
    }
    Sys.sleep(pause)
  }
  if (!quiet) {
    cli::cli_progress_done()
  }

  errors <- status$errors
  if (length(errors) > 0) {
    if (length(errors) > 1) {
      # First error says to look in errors[]
      errors <- errors[-1]
    }

    bullets <- map_chr(errors, function(x) {
      cli_escape(paste0(x$message, " [", x$reason, "]"))
    })
    bullets <- set_names(bullets, "x")
    cli::cli_abort(c("Job {x} failed", bullets), call = call)
  }

  if (!quiet) {
    cli::cli_inform("Job complete")
    bq_job_show_statistics(x)
  }

  invisible(x)
}

# Generated by cpp11: do not edit by hand

bq_parse <- function(meta_s, data_s) {
  .Call(`_bigrquery_bq_parse`, meta_s, data_s)
}

bq_field_init <- function(json, value) {
  .Call(`_bigrquery_bq_field_init`, json, value)
}

bq_parse_files <- function(schema_path, file_paths, n, quiet) {
  .Call(`_bigrquery_bq_parse_files`, schema_path, file_paths, n, quiet)
}

# dbConnect() validates labels

    Code
      dbConnect(bigquery(), project = bq_test_project(), labels = "oops")
    Condition
      Error in `dbConnect()`:
      ! `labels` must be a named list or `NULL`, not the string "oops".

# bq_check_namespace() works

    Code
      bq_check_namespace("invalid package name", "FIELD_TYPE")
    Condition
      Error in `bq_check_namespace()`:
      ! The package "invalid package name" is required to parse BigQuery 'FIELD_TYPE' fields.

# check_labels() errors on invalid inputs

    Code
      check_labels(c(env = "prod"))
    Condition
      Error:
      ! `c(env = "prod")` must be a named list or `NULL`, not the string "prod".
    Code
      check_labels(list("no-name"))
    Condition
      Error:
      ! `list("no-name")` must be a named list or `NULL`, not a list.
    Code
      check_labels(list(env = 1))
    Condition
      Error:
      ! `list(env = 1)` must be a named list of strings, not a list.

# useful print with and without dataset

    Code
      # With dataset
      con1
    Output
      <BigQueryConnection>
        Dataset: p.x
        Billing: b
    Code
      # Without dataset
      con2
    Output
      <BigQueryConnection>
        Billing: p

# dbQuoteIdentifier validates inputs

    Code
      DBI::dbQuoteIdentifier(con, c("x", NA))
    Condition
      Error in `DBI::dbQuoteIdentifier()`:
      ! `x` must not contain missing values.

# dbWriteTable errors on unsupported arguments

    Code
      DBI::dbWriteTable(con, "x", df, field.types = list())
    Condition
      Error in `DBI::dbWriteTable()`:
      ! `field.types` not supported by bigrquery.
    Code
      DBI::dbWriteTable(con, "x", df, temporary = TRUE)
    Condition
      Error in `DBI::dbWriteTable()`:
      ! `temporary = TRUE` not supported by bigrquery.

# can't use parameters

    Code
      DBI::dbGetQuery(con, "SELECT @x AS value", parameters = list(x = 1))
    Condition
      Error in `dbSendQuery()`:
      ! Use `params` not `parameters`.
    Code
      DBI::dbExecute(con, "DELETE table WHERE x <= @x", parameters = list(x = 1))
    Condition
      Error in `dbExecute()`:
      ! Use `params` not `parameters`.

# dataset is optional

    Code
      DBI::dbListTables(con)
    Condition
      Error in `DBI::dbListTables()`:
      ! Can't list tables without a connection `dataset`.

---

    Code
      DBI::dbReadTable(con, "natality", n_max = 10)
    Condition
      Error in `as_bq_table()`:
      ! `name` ("natality") must have 2 or 3 components if the connection doesn't have a dataset.

# can create bq_table from connection + name

    Code
      as_bq_table(con1, "x")
    Condition
      Error in `as_bq_table()`:
      ! `name` ("x") must have 2 or 3 components if the connection doesn't have a dataset.

---

    Code
      as_bq_table(con1, "a.b.c.d")
    Condition
      Error in `as_bq_table()`:
      ! `name` ("a.b.c.d") must have 1-3 components.

# as_bq_table checks its input types

    Code
      as_bq_table(con1, letters)
    Condition
      Error in `as_bq_table()`:
      ! `name` must be a string or a dbplyr_table_ident.

# bq_perform_upload creates job that succeeds

    Code
      bq_job_wait(job, quiet = FALSE)
    Message
      Job complete
      Input: <xxx> kB
      Output: <xxx> kB
    Code
      bq_job_wait(job, quiet = FALSE)
    Message
      Job complete
      Input: <xxx> kB
      Output: <xxx> kB

# bq_perform_copy creates job that succeeds

    Code
      bq_job_wait(job, quiet = FALSE)
    Message
      Job complete

# bq_perform_query creates job that succeeds

    Code
      bq_job_wait(job, quiet = FALSE)
    Message
      Job complete
      Billed: 0 B

# errors when table is known to be incomplete

    Code
      bq_table_download(tb, n_max = 35000, page_size = 35000, bigint = "integer64",
        api = "json")
    Message
      Downloading first chunk of data.
    Condition
      Error in `bq_table_download()`:
      ! First chunk is incomplete:
      x 35,000 rows were requested, but only {n} rows were received.
      i Leave `page_size` unspecified or use an even smaller value.

# warns if supplying unnused arguments

    Code
      . <- bq_table_download(tb, api = "arrow", page_size = 1, start_index = 1,
        max_connections = 1, quiet = TRUE)
    Condition
      Warning in `bq_table_download()`:
      `page_size` is ignored when `api == "arrow"`
      Warning in `bq_table_download()`:
      `start_index` is ignored when `api == "arrow"`
      Warning in `bq_table_download()`:
      `max_connections` is ignored when `api == "arrow"`

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-auth.R
\name{bq_has_token}
\alias{bq_has_token}
\title{Is there a token on hand?}
\usage{
bq_has_token()
}
\value{
Logical.
}
\description{
Reports whether bigrquery has stored a token, ready for use in downstream
requests.
}
\examples{
bq_has_token()
}
\seealso{
Other low-level API functions:
\code{\link[=bq_token]{bq_token()}}
}
\concept{low-level API functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-perform.R
\name{api-perform}
\alias{api-perform}
\alias{bq_perform_extract}
\alias{bq_perform_upload}
\alias{bq_perform_load}
\alias{bq_perform_query}
\alias{bq_perform_query_dry_run}
\alias{bq_perform_query_schema}
\alias{bq_perform_copy}
\title{BigQuery jobs: perform a job}
\usage{
bq_perform_extract(
  x,
  destination_uris,
  destination_format = "NEWLINE_DELIMITED_JSON",
  compression = "NONE",
  ...,
  print_header = TRUE,
  billing = x$project,
  labels = getOption("bigrquery.labels")
)

bq_perform_upload(
  x,
  values,
  fields = NULL,
  source_format = c("NEWLINE_DELIMITED_JSON", "PARQUET"),
  create_disposition = "CREATE_IF_NEEDED",
  write_disposition = "WRITE_EMPTY",
  ...,
  billing = x$project,
  json_digits = NULL,
  labels = getOption("bigrquery.labels")
)

bq_perform_load(
  x,
  source_uris,
  billing = x$project,
  source_format = "NEWLINE_DELIMITED_JSON",
  fields = NULL,
  nskip = 0,
  create_disposition = "CREATE_IF_NEEDED",
  write_disposition = "WRITE_EMPTY",
  ...,
  labels = getOption("bigrquery.labels")
)

bq_perform_query(
  query,
  billing,
  ...,
  parameters = NULL,
  destination_table = NULL,
  default_dataset = NULL,
  create_disposition = "CREATE_IF_NEEDED",
  write_disposition = "WRITE_EMPTY",
  use_legacy_sql = FALSE,
  priority = "INTERACTIVE",
  labels = getOption("bigrquery.labels")
)

bq_perform_query_dry_run(
  query,
  billing,
  ...,
  default_dataset = NULL,
  parameters = NULL,
  use_legacy_sql = FALSE,
  labels = getOption("bigrquery.labels")
)

bq_perform_query_schema(
  query,
  billing,
  ...,
  default_dataset = NULL,
  parameters = NULL,
  labels = getOption("bigrquery.labels")
)

bq_perform_copy(
  src,
  dest,
  create_disposition = "CREATE_IF_NEEDED",
  write_disposition = "WRITE_EMPTY",
  ...,
  billing = NULL,
  labels = getOption("bigrquery.labels")
)
}
\arguments{
\item{x}{A \link{bq_table}}

\item{destination_uris}{A character vector of fully-qualified Google Cloud
Storage URIs where the extracted table should be written. Can export
up to 1 Gb of data per file. Use a wild card URI (e.g.
\verb{gs://[YOUR_BUCKET]/file-name-*.json}) to automatically create any
number of files.}

\item{destination_format}{The exported file format:
\itemize{
\item For CSV files, specify "CSV" (Nested and repeated data is not supported).
\item For newline-delimited JSON, specify "NEWLINE_DELIMITED_JSON".
\item For Avro, specify "AVRO".
\item For parquet, specify "PARQUET".
}}

\item{compression}{The compression type to use for exported files:
\itemize{
\item For CSV files: "GZIP" or "NONE".
\item For newline-delimited JSON: "GZIP" or "NONE".
\item For Avro: "DEFLATE", "SNAPPY" or "NONE".
\item For parquet: "SNAPPY", "GZIP", "ZSTD" or "NONE".
}}

\item{...}{Additional arguments passed on to the underlying API call.
snake_case names are automatically converted to camelCase.}

\item{print_header}{Whether to print out a header row in the results.}

\item{billing}{Identifier of project to bill.}

\item{labels}{A named list of strings used to attach
\href{https://docs.cloud.google.com/bigquery/docs/labels-intro}{BigQuery labels}
to the resulting job, e.g. \code{list(env = "prod", team = "data")}. This
is most useful for cost allocation and other FinOps reporting.
Defaults to the value of \code{getOption("bigrquery.labels")}.}

\item{values}{Data frame of values to insert.}

\item{fields}{A \link{bq_fields} specification, or something coercible to it
(like a data frame). Leave as \code{NULL} to allow BigQuery to auto-detect
the fields.}

\item{source_format}{The format of the data files:
\itemize{
\item For CSV files, specify "CSV".
\item For datastore backups, specify "DATASTORE_BACKUP".
\item For newline-delimited JSON, specify "NEWLINE_DELIMITED_JSON".
\item For Avro, specify "AVRO".
\item For parquet, specify "PARQUET".
\item For orc, specify "ORC".
}}

\item{create_disposition}{Specifies whether the job is allowed to create
new tables.

The following values are supported:
\itemize{
\item "CREATE_IF_NEEDED": If the table does not exist, BigQuery creates the
table.
\item "CREATE_NEVER": The table must already exist. If it does not, a
'notFound' error is returned in the job result.
}}

\item{write_disposition}{Specifies the action that occurs if the
destination table already exists. The following values are supported:
\itemize{
\item "WRITE_TRUNCATE": If the table already exists, BigQuery overwrites the
table data.
\item "WRITE_APPEND": If the table already exists, BigQuery appends the data
to the table.
\item "WRITE_EMPTY": If the table already exists and contains data, a
'duplicate' error is returned in the job result.
}}

\item{json_digits}{Species the number of digits for formatting
numeric values.}

\item{source_uris}{The fully-qualified URIs that point to your data in
Google Cloud.

For Google Cloud Storage URIs: Each URI can contain one
\code{'*'} wildcard character and it must come after the 'bucket' name.
Size limits related to load jobs apply to external data sources.

For Google Cloud Bigtable URIs: Exactly one URI can be specified and
it has be a fully specified and valid HTTPS URL for a Google Cloud
Bigtable table. For Google Cloud Datastore backups: Exactly one URI
can be specified. Also, the '*' wildcard character is not allowed.}

\item{nskip}{For \code{source_format = "CSV"}, the number of header rows to skip.}

\item{query}{SQL query string.}

\item{parameters}{Named list of parameters match to query parameters.
Parameter \code{x} will be matched to placeholder \verb{@x}.

Generally, you can supply R vectors and they will be automatically
converted to the correct type. If you need greater control, you can call
\code{\link[=bq_param_scalar]{bq_param_scalar()}} or \code{\link[=bq_param_array]{bq_param_array()}} explicitly.

See \url{https://docs.cloud.google.com/bigquery/docs/parameterized-queries}
for more details.}

\item{destination_table}{A \link{bq_table} where results should be stored.
If not supplied, results will be saved to a temporary table that lives
in a special dataset. You must supply this parameter for large
queries (> 128 MB compressed).}

\item{default_dataset}{A \link{bq_dataset} used to automatically qualify table names.}

\item{use_legacy_sql}{If \code{TRUE} will use BigQuery's legacy SQL format.}

\item{priority}{Specifies a priority for the query. Possible values include
"INTERACTIVE" and "BATCH". Batch queries do not start immediately,
but are not rate-limited in the same way as interactive queries.}
}
\value{
A \link{bq_job}.
}
\description{
These functions are low-level functions designed to be used by experts.
Each of these low-level functions is paired with a high-level function that
you should use instead:
\itemize{
\item \code{bq_perform_copy()}:    \code{\link[=bq_table_copy]{bq_table_copy()}}.
\item \code{bq_perform_query()}:   \code{\link[=bq_dataset_query]{bq_dataset_query()}}, \code{\link[=bq_project_query]{bq_project_query()}}.
\item \code{bq_perform_upload()}:  \code{\link[=bq_table_upload]{bq_table_upload()}}.
\item \code{bq_perform_load()}:    \code{\link[=bq_table_load]{bq_table_load()}}.
\item \code{bq_perform_extract()}: \code{\link[=bq_table_save]{bq_table_save()}}.
}
}
\section{Google BigQuery API documentation}{

\itemize{
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/jobs}{jobs}
}

Additional information at:
\itemize{
\item \href{https://docs.cloud.google.com/bigquery/docs/exporting-data}{exporting data}
\item \href{https://docs.cloud.google.com/bigquery/docs/loading-data}{loading data}
\item \href{https://docs.cloud.google.com/bigquery/docs/writing-results}{writing queries}
\item \href{https://docs.cloud.google.com/bigquery/docs/managing-tables#copy-table}{copying a table}
}
}

\examples{
\dontshow{if (bq_testable()) withAutoprint(\{ # examplesIf}
ds <- bq_test_dataset()
bq_mtcars <- bq_table(ds, "mtcars")
job <- bq_perform_upload(bq_mtcars, mtcars)
bq_table_exists(bq_mtcars)

bq_job_wait(job)
bq_table_exists(bq_mtcars)
head(bq_table_download(bq_mtcars))
\dontshow{\}) # examplesIf}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-download.R
\name{bq_table_download}
\alias{bq_table_download}
\title{Download table data}
\usage{
bq_table_download(
  x,
  n_max = Inf,
  page_size = NULL,
  start_index = 0L,
  max_connections = 6L,
  quiet = getOption("bigrquery.quiet", NA),
  bigint = c("integer", "integer64", "numeric", "character"),
  api = c("json", "arrow"),
  billing = x$project
)
}
\arguments{
\item{x}{A \link{bq_table}}

\item{n_max}{Maximum number of results to retrieve. Use \code{Inf} to retrieve all
rows.}

\item{page_size}{(JSON only) The number of rows requested per chunk. It is
recommended to leave this unspecified until you have evidence that the
\code{page_size} selected automatically by \code{bq_table_download()} is problematic.

When \code{page_size = NULL} bigrquery determines a conservative, natural chunk
size empirically. If you specify the \code{page_size}, it is important that each
chunk fits on one page, i.e. that the requested row limit is low enough to
prevent the API from paginating based on response size.}

\item{start_index}{(JSON only) Starting row index (zero-based).}

\item{max_connections}{(JSON only) Number of maximum simultaneous
connections to BigQuery servers.}

\item{quiet}{If \code{FALSE}, displays progress bar; if \code{TRUE} is silent;
if \code{NA} picks based on whether or not you're in an interactive context.}

\item{bigint}{The R type that BigQuery's 64-bit integer types should be
mapped to. The default is \code{"integer"}, which returns R's \code{integer} type,
but results in \code{NA} for values above/below +/- 2147483647. \code{"integer64"}
returns a \link[bit64:integer64]{bit64::integer64}, which allows the full range of 64 bit
integers.}

\item{api}{Which API to use? The \code{"json"} API works where ever bigrquery
does, but is slow and can require fiddling with the \code{page_size} parameter.
The \code{"arrow"} API is faster and more reliable, but only works if you
have also installed the bigrquerystorage package.

Because the \code{"arrow"} API is so much faster, it will be used automatically
if the bigrquerystorage package is installed.}

\item{billing}{(Arrow only) Project to bill; defaults to the project of \code{x},
and typically only needs to be specified if you're working with public
datasets.}
}
\value{
Because data retrieval may generate list-columns and the \code{data.frame}
print method can have problems with list-columns, this method returns
a tibble. If you need a \code{data.frame}, coerce the results with
\code{\link[=as.data.frame]{as.data.frame()}}.
}
\description{
This function provides two ways to download data from BigQuery, transfering
data using either JSON or arrow, depending on the \code{api} argument. If
bigrquerystorage is installed, \code{api = "arrow"} will be used (because it's
so much faster, but see the limitions below), otherwise you can select
deliberately by using \code{api = "json"} or \code{api = "arrow"}.
\subsection{Arrow API}{

The arrow API is much faster, but has heavier dependencies: bigrquerystorage
requires the arrow package, which can be tricky to compile on Linux (but you
usually should be able to get a binary from
\href{https://p3m.dev/}{Posit Public Package Manager}.

There's one known limitation of \code{api = "arrow"}: when querying public data,
you'll now need to provide a \code{billing} project.
}

\subsection{JSON API}{

The JSON API retrieves rows in chunks of \code{page_size}. It is most suitable
for results of smaller queries (<100 MB, say). Unfortunately due to
limitations in the BigQuery API, you may need to vary this parameter
depending on the complexity of the underlying data.

The JSON API will convert nested and repeated columns in to list-columns
as follows:
\itemize{
\item Repeated values (arrays) will become a list-column of vectors.
\item Records will become list-columns of named lists.
\item Repeated records will become list-columns of data frames.
}
}
}
\section{Google BigQuery API documentation}{

\itemize{
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tabledata/list}{list}
}
}

\examples{
\dontshow{if (bq_testable()) withAutoprint(\{ # examplesIf}
df <- bq_table_download("publicdata.samples.natality", n_max = 35000, billing = bq_test_project())
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-job.R
\name{api-job}
\alias{api-job}
\alias{bq_job_meta}
\alias{bq_job_status}
\alias{bq_job_show_statistics}
\alias{bq_job_wait}
\title{BigQuery job: retrieve metadata}
\usage{
bq_job_meta(x, fields = NULL)

bq_job_status(x)

bq_job_show_statistics(x)

bq_job_wait(
  x,
  quiet = getOption("bigrquery.quiet"),
  pause = 0.5,
  call = caller_env()
)
}
\arguments{
\item{x}{A \link{bq_job}}

\item{fields}{An optional field specification for
\href{https://docs.cloud.google.com/bigquery/docs/api-performance#partial-response}{partial response}}

\item{quiet}{If \code{FALSE}, displays progress bar; if \code{TRUE} is silent;
if \code{NA} picks based on whether or not you're in an interactive context.}

\item{pause}{amount of time to wait between status requests}

\item{call}{The execution environment of a currently
running function, e.g. \code{caller_env()}. The function will be
mentioned in error messages as the source of the error. See the
\code{call} argument of \code{\link[rlang:abort]{abort()}} for more information.}
}
\description{
To perform a job, see \link{api-perform}. These functions all retrieve metadata
(in various forms) about an existing job.
}
\section{Google BigQuery API documentation}{

\itemize{
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/jobs/get}{get}
}
}

\examples{
\dontshow{if (bq_testable()) withAutoprint(\{ # examplesIf}
jobs <- bq_project_jobs(bq_test_project())
jobs[[1]]

# Show statistics about job
bq_job_show_statistics(jobs[[1]])

# Wait for job to complete
bq_job_wait(jobs[[1]])
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-table.R
\name{api-table}
\alias{api-table}
\alias{bq_table_create}
\alias{bq_table_meta}
\alias{bq_table_fields}
\alias{bq_table_size}
\alias{bq_table_nrow}
\alias{bq_table_exists}
\alias{bq_table_delete}
\alias{bq_table_copy}
\alias{bq_table_upload}
\alias{bq_table_save}
\alias{bq_table_load}
\alias{bq_table_patch}
\title{BigQuery tables}
\usage{
bq_table_create(x, fields = NULL, ...)

bq_table_meta(x, fields = NULL)

bq_table_fields(x)

bq_table_size(x)

bq_table_nrow(x)

bq_table_exists(x)

bq_table_delete(x)

bq_table_copy(x, dest, ..., quiet = getOption("bigrquery.quiet", NA))

bq_table_upload(
  x,
  values,
  ...,
  quiet = getOption("bigrquery.quiet", NA),
  json_digits = NULL
)

bq_table_save(
  x,
  destination_uris,
  ...,
  quiet = getOption("bigrquery.quiet", NA)
)

bq_table_load(x, source_uris, ..., quiet = getOption("bigrquery.quiet", NA))

bq_table_patch(x, fields)
}
\arguments{
\item{x}{A \link{bq_table}, or an object coercible to a \code{bq_table}.}

\item{fields}{A \link{bq_fields} specification, or something coercible to it
(like a data frame).}

\item{...}{Additional arguments passed on to the underlying API call.
snake_case names are automatically converted to camelCase.}

\item{dest}{Source and destination \link{bq_table}s.}

\item{quiet}{If \code{FALSE}, displays progress bar; if \code{TRUE} is silent;
if \code{NA} picks based on whether or not you're in an interactive context.}

\item{values}{Data frame of values to insert.}

\item{json_digits}{Species the number of digits for formatting
numeric values.}

\item{destination_uris}{A character vector of fully-qualified Google Cloud
Storage URIs where the extracted table should be written. Can export
up to 1 Gb of data per file. Use a wild card URI (e.g.
\verb{gs://[YOUR_BUCKET]/file-name-*.json}) to automatically create any
number of files.}

\item{source_uris}{The fully-qualified URIs that point to your data in
Google Cloud.

For Google Cloud Storage URIs: Each URI can contain one
\code{'*'} wildcard character and it must come after the 'bucket' name.
Size limits related to load jobs apply to external data sources.

For Google Cloud Bigtable URIs: Exactly one URI can be specified and
it has be a fully specified and valid HTTPS URL for a Google Cloud
Bigtable table. For Google Cloud Datastore backups: Exactly one URI
can be specified. Also, the '*' wildcard character is not allowed.}
}
\value{
\itemize{
\item \code{bq_table_copy()}, \code{bq_table_create()}, \code{bq_table_delete()}, \code{bq_table_upload()}:
an invisible \link{bq_table}
\item \code{bq_table_exists()}: either \code{TRUE} or \code{FALSE}.
\item \code{bq_table_size()}: the size of the table in bytes
\item \code{bq_table_fields()}: a \link{bq_fields}.
}
}
\description{
Basic create-read-update-delete verbs for tables, as well as functions
uploading data (\code{bq_table_upload()}), saving to/loading from Google
Cloud Storage (\code{bq_table_load()}, \code{bq_table_save()}), and getting
various values from the metadata.
}
\section{Google BigQuery API documentation}{

\itemize{
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables/insert}{insert}
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables/get}{get}
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables/delete}{delete}
}
}

\examples{
\dontshow{if (bq_testable()) withAutoprint(\{ # examplesIf}
ds <- bq_test_dataset()

bq_mtcars <- bq_table(ds, "mtcars")
bq_table_exists(bq_mtcars)

bq_table_create(
  bq_mtcars,
  fields = mtcars,
  friendly_name = "Motor Trend Car Road Tests",
  description = "The data was extracted from the 1974 Motor Trend US magazine",
  labels = list(category = "example")
)
bq_table_exists(bq_mtcars)

bq_table_upload(bq_mtcars, mtcars)

bq_table_fields(bq_mtcars)
bq_table_size(bq_mtcars)
str(bq_table_meta(bq_mtcars))

bq_table_delete(bq_mtcars)
bq_table_exists(bq_mtcars)

my_natality <- bq_table(ds, "mynatality")
bq_table_copy("publicdata.samples.natality", my_natality)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-refs.R
\name{bq_refs}
\alias{bq_dataset}
\alias{as_bq_dataset}
\alias{bq_table}
\alias{as_bq_table}
\alias{bq_job}
\alias{as_bq_job}
\title{S3 classes for BigQuery datasets, tables and jobs}
\usage{
bq_dataset(project, dataset)

as_bq_dataset(x, ..., error_arg = caller_arg(x), error_call = caller_env())

bq_table(project, dataset, table = NULL, type = "TABLE")

as_bq_table(x, ..., error_arg = caller_arg(x), error_call = caller_env())

bq_job(project, job, location = "US")

as_bq_job(x, ..., error_arg = caller_arg(x), error_call = caller_env())
}
\arguments{
\item{project, dataset, table, job, type}{Individual project, dataset, table,
job identifiers and table type (strings).

For \code{bq_table()}, you if supply a \code{bq_dataset} as the first argument,
the 2nd argument will be interpreted as the \code{table}}

\item{x}{An object to coerce to a \code{bq_job}, \code{bq_dataset}, or \code{bq_table}.
Built-in methods handle strings and lists.}

\item{...}{Other arguments passed on to methods.}

\item{error_arg}{An argument name as a string. This argument
will be mentioned in error messages as the input that is at the
origin of a problem.}

\item{error_call}{The execution environment of a currently
running function, e.g. \code{caller_env()}. The function will be
mentioned in error messages as the source of the error. See the
\code{call} argument of \code{\link[rlang:abort]{abort()}} for more information.}

\item{location}{Job location}
}
\description{
Create references to BigQuery datasets, jobs, and tables. Each class
has a constructor function (\code{bq_dataset()}, \code{bq_table()}, \code{bq_job()})
and a coercion function (\code{as_bq_dataset()}, \code{as_bq_table()}, \code{as_bq_job()}).
The coercions functions come with methods for strings (which find components
by splitting on \code{.}), and lists (which look for named components like
\code{projectId} or \code{project_id}).

All \code{bq_table_}, \code{bq_dataset_} and \code{bq_job_} functions call the appropriate
coercion functions on their first argument, allowing you to flexible specify
their inputs.
}
\examples{
# Creation ------------------------------------------------
samples <- bq_dataset("publicdata", "samples")
natality <- bq_table("publicdata", "samples", "natality")
natality

# Or
bq_table(samples, "natality")

bq_job("bigrquery-examples", "m0SgFu2ycbbge6jgcvzvflBJ_Wft")

# Coercion ------------------------------------------------
as_bq_dataset("publicdata.shakespeare")
as_bq_table("publicdata.samples.natality")

as_bq_table(list(
  project_id = "publicdata",
  dataset_id = "samples",
  table_id = "natality"
))

as_bq_job(list(
  projectId = "bigrquery-examples",
  jobId = "job_m0SgFu2ycbbge6jgcvzvflBJ_Wft",
  location = "US"
))

}
\seealso{
\link{api-job}, \link{api-perform}, \link{api-dataset}, and \link{api-table} for
functions that work with these objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-auth.R
\name{bq_deauth}
\alias{bq_deauth}
\title{Clear current token}
\usage{
bq_deauth()
}
\description{
Clears any currently stored token. The next time bigrquery needs a token, the
token acquisition process starts over, with a fresh call to \code{\link[=bq_auth]{bq_auth()}} and,
therefore, internally, a call to \code{\link[gargle:token_fetch]{gargle::token_fetch()}}. Unlike some other
packages that use gargle, bigrquery is not usable in a de-authorized state.
Therefore, calling \code{bq_deauth()} only clears the token, i.e. it does NOT
imply that subsequent requests are made with an API key in lieu of a token.
}
\examples{
\dontrun{
bq_deauth()
}
}
\seealso{
Other auth functions:
\code{\link[=bq_auth]{bq_auth()}},
\code{\link[=bq_auth_configure]{bq_auth_configure()}}
}
\concept{auth functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bigrquery-package.R
\docType{package}
\name{bigrquery-package}
\alias{bigrquery}
\alias{bigrquery-package}
\title{bigrquery: An Interface to Google's 'BigQuery' 'API'}
\description{
Easily talk to Google's 'BigQuery' database from R.
}
\section{Package options}{

\describe{
\item{\code{bigrquery.quiet}}{Verbose output during processing? The default
value, \code{NA}, turns on verbose output for queries that run longer than
two seconds.  Use \code{FALSE} for immediate verbose output, \code{TRUE}
for quiet operation.}

\item{\code{bigrquery.page.size}}{Default page size for fetching data,
defaults to 1e4.}
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://bigrquery.r-dbi.org}
  \item \url{https://github.com/r-dbi/bigrquery}
  \item Report bugs at \url{https://github.com/r-dbi/bigrquery/issues}
}

}
\author{
\strong{Maintainer}: Hadley Wickham \email{hadley@posit.co} (\href{https://orcid.org/0000-0003-4757-117X}{ORCID})

Authors:
\itemize{
  \item Hadley Wickham \email{hadley@posit.co} (\href{https://orcid.org/0000-0003-4757-117X}{ORCID})
  \item Jennifer Bryan \email{jenny@posit.co} (\href{https://orcid.org/0000-0002-6983-2759}{ORCID})
}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-projects.R
\name{bq_projects}
\alias{bq_projects}
\title{List available projects}
\usage{
bq_projects(page_size = 100, max_pages = 1, warn = TRUE)
}
\arguments{
\item{page_size}{Number of items per page.}

\item{max_pages}{Maximum number of pages to retrieve. Use \code{Inf} to retrieve
all pages (this may take a long time!)}

\item{warn}{If \code{TRUE}, warn when there are unretrieved pages.}
}
\value{
A character vector.
}
\description{
List all projects that you have access to. You can also work with
\href{https://docs.cloud.google.com/bigquery/public-data/}{public datasets},
but you will need to provide a \code{billing} project whenever you perform
any non-free operation.
}
\section{Google BigQuery API documentation}{

\itemize{
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/projects/list}{list}
}
}

\examples{
\dontshow{if (bq_testable()) withAutoprint(\{ # examplesIf}
bq_projects()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-param.R
\name{bq_param}
\alias{bq_param}
\alias{bq_param_scalar}
\alias{bq_param_array}
\title{Explicitly define query parameters}
\usage{
bq_param(value, type = NULL, name = NULL)

bq_param_scalar(value, type = NULL, name = NULL)

bq_param_array(value, type = NULL, name = NULL)
}
\arguments{
\item{value}{vector of parameter values}

\item{type}{BigQuery type of the parameter}

\item{name}{name of the parameter in the query, omitting the \code{@}}
}
\description{
By default, bigrquery will assume vectors of length 1 are scalars,
and longer vectors are arrays. If you need to pass a length-1 array,
you'll need to explicitly use \code{bq_param_array()}.
}
\examples{
# bq_param() automatically picks scalar vs array based on length
bq_param("a")
bq_param(c("a", "b", "c"))

# use bq_param_array() to create a length-1 array
bq_param_array("a")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr.R
\name{collect.tbl_BigQueryConnection}
\alias{collect.tbl_BigQueryConnection}
\title{Collect a BigQuery table}
\usage{
\method{collect}{tbl_BigQueryConnection}(
  x,
  ...,
  n = Inf,
  api = c("json", "arrow"),
  page_size = NULL,
  max_connections = 6L
)
}
\arguments{
\item{x}{A data frame, data frame extension (e.g. a tibble), or a lazy
data frame (e.g. from dbplyr or dtplyr). See \emph{Methods}, below, for more
details.}

\item{...}{Other arguments passed on to
\code{bq_project_query()}/\code{bq_project_query()}}

\item{n}{Maximum number of results to retrieve.
The default, \code{Inf}, will retrieve all rows.}

\item{api}{Which API to use? The \code{"json"} API works where ever bigrquery
does, but is slow and can require fiddling with the \code{page_size} parameter.
The \code{"arrow"} API is faster and more reliable, but only works if you
have also installed the bigrquerystorage package.

Because the \code{"arrow"} API is so much faster, it will be used automatically
if the bigrquerystorage package is installed.}

\item{page_size}{(JSON only) The number of rows requested per chunk. It is
recommended to leave this unspecified until you have evidence that the
\code{page_size} selected automatically by \code{bq_table_download()} is problematic.

When \code{page_size = NULL} bigrquery determines a conservative, natural chunk
size empirically. If you specify the \code{page_size}, it is important that each
chunk fits on one page, i.e. that the requested row limit is low enough to
prevent the API from paginating based on response size.}

\item{max_connections}{(JSON only) Number of maximum simultaneous
connections to BigQuery servers.}
}
\description{
This collect method is specialised for BigQuery tables, generating the
SQL from your dplyr commands, then calling \code{\link[=bq_project_query]{bq_project_query()}}
or \code{\link[=bq_dataset_query]{bq_dataset_query()}} to run the query, then \code{\link[=bq_table_download]{bq_table_download()}}
to download the results. Thus the arguments are a combination of the
arguments to \code{\link[dplyr:collect]{dplyr::collect()}}, \code{bq_project_query()}/\code{bq_dataset_query()},
and \code{bq_table_download()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-dataset.R
\name{api-dataset}
\alias{api-dataset}
\alias{bq_dataset_create}
\alias{bq_dataset_meta}
\alias{bq_dataset_exists}
\alias{bq_dataset_update}
\alias{bq_dataset_delete}
\alias{bq_dataset_tables}
\title{BigQuery datasets}
\usage{
bq_dataset_create(x, location = "US", ...)

bq_dataset_meta(x, fields = NULL)

bq_dataset_exists(x)

bq_dataset_update(x, ...)

bq_dataset_delete(x, delete_contents = FALSE)

bq_dataset_tables(x, page_size = 50, max_pages = Inf, warn = TRUE, ...)
}
\arguments{
\item{x}{A \link{bq_dataset}}

\item{location}{Dataset location}

\item{...}{Additional arguments passed on to the underlying API call.
snake_case names are automatically converted to camelCase.}

\item{fields}{An optional field specification for
\href{https://docs.cloud.google.com/bigquery/docs/api-performance#partial-response}{partial response}}

\item{delete_contents}{If \code{TRUE}, will recursively delete all tables in
the dataset. Set to \code{FALSE} by default for safety.}

\item{page_size}{Number of items per page.}

\item{max_pages}{Maximum number of pages to retrieve. Use \code{Inf} to retrieve
all pages (this may take a long time!)}

\item{warn}{If \code{TRUE}, warn when there are unretrieved pages.}
}
\description{
Basic create-read-update-delete verbs for datasets.
}
\section{Google BigQuery API documentation}{

\itemize{
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/datasets/get}{get}
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/datasets/insert}{insert}
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/datasets/delete}{delete}
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables/list}{list}
}
}

\examples{
\dontshow{if (bq_testable()) withAutoprint(\{ # examplesIf}
ds <- bq_dataset(bq_test_project(), "dataset_api")
bq_dataset_exists(ds)

bq_dataset_create(ds)
bq_dataset_exists(ds)
str(bq_dataset_meta(ds))

bq_dataset_delete(ds)
bq_dataset_exists(ds)

# Use bq_test_dataset() to create a temporary dataset that will
# be automatically deleted
ds <- bq_test_dataset()
bq_table_create(bq_table(ds, "x1"))
bq_table_create(bq_table(ds, "x2"))
bq_table_create(bq_table(ds, "x3"))
bq_dataset_tables(ds)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-query.R
\name{bq_query}
\alias{bq_query}
\alias{bq_project_query}
\alias{bq_dataset_query}
\title{Submit query to BigQuery}
\usage{
bq_project_query(
  x,
  query,
  destination_table = NULL,
  ...,
  quiet = getOption("bigrquery.quiet", NA)
)

bq_dataset_query(
  x,
  query,
  destination_table = NULL,
  ...,
  billing = NULL,
  quiet = getOption("bigrquery.quiet", NA)
)
}
\arguments{
\item{x}{Either a project (a string) or a \link{bq_dataset}.}

\item{query}{SQL query string.}

\item{destination_table}{A \link{bq_table} where results should be stored.
If not supplied, results will be saved to a temporary table that lives
in a special dataset. You must supply this parameter for large
queries (> 128 MB compressed).}

\item{...}{Passed on to \code{\link[=bq_perform_query]{bq_perform_query()}}}

\item{quiet}{If \code{FALSE}, displays progress bar; if \code{TRUE} is silent;
if \code{NA} picks based on whether or not you're in an interactive context.}

\item{billing}{If you query a dataset that you only have read access
for, such as a public dataset, you must also submit a \code{billing} project.}
}
\value{
A \link{bq_table}
}
\description{
These submit a query (using \code{\link[=bq_perform_query]{bq_perform_query()}}) and then wait for it
complete (with \code{\link[=bq_job_wait]{bq_job_wait()}}). All BigQuery queries save their results
into a table (temporary or otherwise), so these functions return a \link{bq_table}
which you can then query for more information.
}
\examples{
\dontshow{if (bq_testable()) withAutoprint(\{ # examplesIf}
# Querying a project requires full name in query
tb <- bq_project_query(
  bq_test_project(),
  "SELECT count(*) FROM publicdata.samples.natality"
)
bq_table_fields(tb)
bq_table_download(tb)

# Querying a dataset sets default dataset so you can use bare table name,
# but for public data, you'll need to set a project to bill.
ds <- bq_dataset("publicdata", "samples")
tb <- bq_dataset_query(ds,
  query = "SELECT count(*) FROM natality",
  billing = bq_test_project()
)
bq_table_download(tb)

tb <- bq_dataset_query(ds,
  query = "SELECT count(*) FROM natality WHERE state = @state",
  parameters = list(state = "KS"),
  billing = bq_test_project()
)
bq_table_download(tb)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-field.R
\name{bq_field}
\alias{bq_field}
\alias{bq_fields}
\alias{as_bq_field}
\alias{as_bq_fields}
\title{BigQuery field (and fields) class}
\usage{
bq_field(name, type, mode = "NULLABLE", fields = list(), description = NULL)

bq_fields(x)

as_bq_field(x)

as_bq_fields(x)
}
\arguments{
\item{name}{The field name. The name must contain only letters (a-z, A-Z),
numbers (0-9), or underscores (_), and must start with a letter or
underscore. The maximum length is 300 characters.}

\item{type}{The field data type. Possible values include:
\code{"STRING"}, \code{"BYTES"}, \code{"INTEGER"}, \code{"FLOAT"}, \code{"BOOLEAN"}, \code{"TIMESTAMP"},
\code{"DATE"}, \code{"TIME"}, \code{"DATETIME"}, \code{"GEOGRAPHY"}, \code{"NUMERIC"},
\code{"BIGNUMERIC"}, \code{"JSON"}, \code{"RECORD"}.}

\item{mode}{The field mode. Possible values include: \code{"NULLABLE"},
\code{"REQUIRED"}, and \code{"REPEATED"}.}

\item{fields}{For a field of type "record", a list of sub-fields.}

\item{description}{The field description. The maximum length is 1,024
characters.}

\item{x}{A list of \code{bg_fields}}
}
\description{
\code{bq_field()} and \code{bq_fields()} create; \code{as_bq_field()} and \code{as_bq_fields()}
coerce from lists.
}
\examples{
bq_field("name", "string")

as_bq_fields(list(
  list(name = "name", type = "string"),
  bq_field("age", "integer")
))

# as_bq_fields() can also take a data frame
as_bq_fields(mtcars)
}
\seealso{
\code{bq_field()} corresponds to a \code{TableFieldSchema}, see
\url{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/tables#TableFieldSchema}
for more details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi-driver.R, R/dbi-connection.R,
%   R/dbi-result.R
\docType{class}
\name{DBI}
\alias{DBI}
\alias{BigQueryDriver-class}
\alias{dbConnect,bq_dataset-method}
\alias{show,BigQueryDriver-method}
\alias{dbGetInfo,BigQueryDriver-method}
\alias{dbIsValid,BigQueryDriver-method}
\alias{dbDataType,BigQueryDriver-method}
\alias{BigQueryConnection-class}
\alias{show,BigQueryConnection-method}
\alias{dbIsValid,BigQueryConnection-method}
\alias{dbDisconnect,BigQueryConnection-method}
\alias{dbSendQuery,BigQueryConnection,character-method}
\alias{dbExecute,BigQueryConnection,character-method}
\alias{dbQuoteString,BigQueryConnection,character-method}
\alias{dbQuoteString,BigQueryConnection,SQL-method}
\alias{dbQuoteIdentifier,BigQueryConnection,character-method}
\alias{dbQuoteIdentifier,BigQueryConnection,SQL-method}
\alias{dbQuoteLiteral,BigQueryConnection,logical-method}
\alias{dbDataType,BigQueryConnection-method}
\alias{dbWriteTable,BigQueryConnection,character,data.frame-method}
\alias{dbWriteTable,BigQueryConnection,Id,data.frame-method}
\alias{dbWriteTable,BigQueryConnection,AsIs,data.frame-method}
\alias{dbAppendTable,BigQueryConnection,character,data.frame-method}
\alias{dbAppendTable,BigQueryConnection,Id,data.frame-method}
\alias{dbAppendTable,BigQueryConnection,AsIs,data.frame-method}
\alias{dbCreateTable,BigQueryConnection-method}
\alias{dbReadTable,BigQueryConnection,character-method}
\alias{dbReadTable,BigQueryConnection,Id-method}
\alias{dbReadTable,BigQueryConnection,AsIs-method}
\alias{dbListTables,BigQueryConnection-method}
\alias{dbExistsTable,BigQueryConnection,character-method}
\alias{dbExistsTable,BigQueryConnection,Id-method}
\alias{dbExistsTable,BigQueryConnection,AsIs-method}
\alias{dbListFields,BigQueryConnection,character-method}
\alias{dbListFields,BigQueryConnection,Id-method}
\alias{dbListFields,BigQueryConnection,AsIs-method}
\alias{dbRemoveTable,BigQueryConnection,character-method}
\alias{dbRemoveTable,BigQueryConnection,Id-method}
\alias{dbRemoveTable,BigQueryConnection,AsIs-method}
\alias{dbGetInfo,BigQueryConnection-method}
\alias{dbBegin,BigQueryConnection-method}
\alias{dbCommit,BigQueryConnection-method}
\alias{dbRollback,BigQueryConnection-method}
\alias{BigQueryResult-class}
\alias{show,BigQueryResult-method}
\alias{dbIsValid,BigQueryResult-method}
\alias{dbClearResult,BigQueryResult-method}
\alias{dbFetch,BigQueryResult-method}
\alias{dbHasCompleted,BigQueryResult-method}
\alias{dbGetStatement,BigQueryResult-method}
\alias{dbColumnInfo,BigQueryResult-method}
\alias{dbGetRowCount,BigQueryResult-method}
\alias{dbGetRowsAffected,BigQueryResult-method}
\alias{dbBind,BigQueryResult-method}
\title{DBI methods}
\usage{
\S4method{dbConnect}{bq_dataset}(drv, ...)

\S4method{show}{BigQueryDriver}(object)

\S4method{dbGetInfo}{BigQueryDriver}(dbObj, ...)

\S4method{dbIsValid}{BigQueryDriver}(dbObj, ...)

\S4method{dbDataType}{BigQueryDriver}(dbObj, obj, ...)

\S4method{show}{BigQueryConnection}(object)

\S4method{dbIsValid}{BigQueryConnection}(dbObj, ...)

\S4method{dbDisconnect}{BigQueryConnection}(conn, ...)

\S4method{dbSendQuery}{BigQueryConnection,character}(conn, statement, ..., params = NULL)

\S4method{dbExecute}{BigQueryConnection,character}(conn, statement, ..., params = NULL)

\S4method{dbQuoteString}{BigQueryConnection,character}(conn, x, ...)

\S4method{dbQuoteString}{BigQueryConnection,SQL}(conn, x, ...)

\S4method{dbQuoteIdentifier}{BigQueryConnection,character}(conn, x, ...)

\S4method{dbQuoteIdentifier}{BigQueryConnection,SQL}(conn, x, ...)

\S4method{dbQuoteLiteral}{BigQueryConnection,logical}(conn, x, ...)

\S4method{dbDataType}{BigQueryConnection}(dbObj, obj, ...)

\S4method{dbWriteTable}{BigQueryConnection,character,data.frame}(
  conn,
  name,
  value,
  ...,
  overwrite = FALSE,
  append = FALSE,
  field.types = NULL,
  temporary = FALSE,
  row.names = NA
)

\S4method{dbWriteTable}{BigQueryConnection,Id,data.frame}(
  conn,
  name,
  value,
  ...,
  overwrite = FALSE,
  append = FALSE,
  field.types = NULL,
  temporary = FALSE,
  row.names = NA
)

\S4method{dbWriteTable}{BigQueryConnection,AsIs,data.frame}(
  conn,
  name,
  value,
  ...,
  overwrite = FALSE,
  append = FALSE,
  field.types = NULL,
  temporary = FALSE,
  row.names = NA
)

\S4method{dbAppendTable}{BigQueryConnection,character,data.frame}(conn, name, value, ..., row.names = NULL)

\S4method{dbAppendTable}{BigQueryConnection,Id,data.frame}(conn, name, value, ..., row.names = NULL)

\S4method{dbAppendTable}{BigQueryConnection,AsIs,data.frame}(conn, name, value, ..., row.names = NULL)

\S4method{dbCreateTable}{BigQueryConnection}(conn, name, fields, ..., row.names = NULL, temporary = FALSE)

\S4method{dbCreateTable}{BigQueryConnection}(conn, name, fields, ..., row.names = NULL, temporary = FALSE)

\S4method{dbReadTable}{BigQueryConnection,character}(conn, name, ...)

\S4method{dbReadTable}{BigQueryConnection,Id}(conn, name, ...)

\S4method{dbReadTable}{BigQueryConnection,AsIs}(conn, name, ...)

\S4method{dbListTables}{BigQueryConnection}(conn, ...)

\S4method{dbExistsTable}{BigQueryConnection,character}(conn, name, ...)

\S4method{dbExistsTable}{BigQueryConnection,Id}(conn, name, ...)

\S4method{dbExistsTable}{BigQueryConnection,AsIs}(conn, name, ...)

\S4method{dbListFields}{BigQueryConnection,character}(conn, name, ...)

\S4method{dbListFields}{BigQueryConnection,Id}(conn, name, ...)

\S4method{dbListFields}{BigQueryConnection,AsIs}(conn, name, ...)

\S4method{dbRemoveTable}{BigQueryConnection,character}(conn, name, ...)

\S4method{dbRemoveTable}{BigQueryConnection,Id}(conn, name, ...)

\S4method{dbRemoveTable}{BigQueryConnection,AsIs}(conn, name, ...)

\S4method{dbGetInfo}{BigQueryConnection}(dbObj, ...)

\S4method{dbBegin}{BigQueryConnection}(conn, ...)

\S4method{dbCommit}{BigQueryConnection}(conn, ...)

\S4method{dbRollback}{BigQueryConnection}(conn, ...)

\S4method{show}{BigQueryResult}(object)

\S4method{dbIsValid}{BigQueryResult}(dbObj, ...)

\S4method{dbClearResult}{BigQueryResult}(res, ...)

\S4method{dbFetch}{BigQueryResult}(res, n = -1, ...)

\S4method{dbHasCompleted}{BigQueryResult}(res, ...)

\S4method{dbGetStatement}{BigQueryResult}(res, ...)

\S4method{dbColumnInfo}{BigQueryResult}(res, ...)

\S4method{dbGetRowCount}{BigQueryResult}(res, ...)

\S4method{dbGetRowsAffected}{BigQueryResult}(res, ...)

\S4method{dbBind}{BigQueryResult}(res, params, ...)
}
\arguments{
\item{...}{Other arguments to methods.}

\item{object}{Any R object}

\item{dbObj}{An object inheriting from \link[DBI:DBIObject-class]{DBI::DBIObject},
i.e. \link[DBI:DBIDriver-class]{DBI::DBIDriver}, \link[DBI:DBIConnection-class]{DBI::DBIConnection}, or a \link[DBI:DBIResult-class]{DBI::DBIResult}.}

\item{obj}{An R object whose SQL type we want to determine.}

\item{conn}{A \link[DBI:DBIConnection-class]{DBI::DBIConnection} object, as returned by
\code{\link[DBI:dbConnect]{DBI::dbConnect()}}.}

\item{statement}{a character string containing SQL.}

\item{params}{Named list of parameters match to query parameters.
Parameter \code{x} will be matched to placeholder \verb{@x}.}

\item{x}{A character vector to quote as string.}

\item{name}{The table name, passed on to \code{\link[DBI:dbQuoteIdentifier]{dbQuoteIdentifier()}}. Options are:
\itemize{
\item a character string with the unquoted DBMS table name,
e.g. \code{"table_name"},
\item a call to \code{\link[DBI:Id]{Id()}} with components to the fully qualified table name,
e.g. \code{Id(schema = "my_schema", table = "table_name")}
\item a call to \code{\link[DBI:SQL]{SQL()}} with the quoted and fully qualified table name
given verbatim, e.g. \code{SQL('"my_schema"."table_name"')}
}}

\item{value}{A \link{data.frame} (or coercible to data.frame).}

\item{overwrite}{a logical specifying whether to overwrite an existing table
or not. Its default is \code{FALSE}.}

\item{append}{a logical specifying whether to append to an existing table
in the DBMS.  Its default is \code{FALSE}.}

\item{field.types, temporary}{Ignored. Included for compatibility with
generic.}

\item{row.names}{A logical specifying whether the \code{row.names} should be
output to the output DBMS table; if \code{TRUE}, an extra field whose name
will be whatever the R identifier \code{"row.names"} maps to the DBMS (see
\code{\link[DBI:make.db.names]{DBI::make.db.names()}}). If \code{NA} will add rows names if
they are characters, otherwise will ignore.}

\item{fields}{Either a character vector or a data frame.

A named character vector: Names are column names, values are types.
Names are escaped with \code{\link[DBI:dbQuoteIdentifier]{dbQuoteIdentifier()}}.
Field types are unescaped.

A data frame: field types are generated using
\code{\link[DBI:dbDataType]{dbDataType()}}.}

\item{res}{An object inheriting from \link[DBI:DBIResult-class]{DBI::DBIResult}.}

\item{n}{maximum number of records to retrieve per fetch. Use \code{n = -1}
or \code{n = Inf}
to retrieve all pending records.  Some implementations may recognize other
special values.}
}
\description{
Implementations of pure virtual functions defined in the \code{DBI} package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-auth.R
\name{bq_auth_configure}
\alias{bq_auth_configure}
\alias{bq_oauth_client}
\title{Edit and view auth configuration}
\usage{
bq_auth_configure(client, path)

bq_oauth_client()
}
\arguments{
\item{client}{A Google OAuth client, presumably constructed via
\code{\link[gargle:gargle_oauth_client_from_json]{gargle::gargle_oauth_client_from_json()}}. Note, however, that it is
preferred to specify the client with JSON, using the \code{path} argument.}

\item{path}{JSON downloaded from \href{https://console.cloud.google.com}{Google Cloud Console}, containing a client id and
secret, in one of the forms supported for the \code{txt} argument of
\code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} (typically, a file path or JSON string).}
}
\value{
\itemize{
\item \code{bq_auth_configure()}: An object of R6 class
\link[gargle:AuthState]{gargle::AuthState}, invisibly.
\item \code{bq_oauth_client()}: the current user-configured OAuth client.
}
}
\description{
These functions give more control over and visibility into the auth
configuration than \code{\link[=bq_auth]{bq_auth()}} does. \code{bq_auth_configure()}
lets the user specify their own:
\itemize{
\item OAuth client, which is used when obtaining a user token.
}

See the \code{vignette("get-api-credentials", package = "gargle")}
for more.
If the user does not configure these settings, internal defaults
are used.

\code{bq_oauth_client()} retrieves the currently configured OAuth client.
}
\examples{
# see and store the current user-configured OAuth client (probably `NULL`)
(original_client <- bq_oauth_client())

# the preferred way to configure your own client is via a JSON file
# downloaded from Google Developers Console
# this example JSON is indicative, but fake
path_to_json <- system.file(
  "extdata", "data", "client_secret_123.googleusercontent.com.json",
  package = "bigrquery"
)
bq_auth_configure(path = path_to_json)

# confirm the changes
bq_oauth_client()

# restore original auth config
bq_auth_configure(client = original_client)
}
\seealso{
Other auth functions:
\code{\link[=bq_auth]{bq_auth()}},
\code{\link[=bq_deauth]{bq_deauth()}}
}
\concept{auth functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbi-driver.R
\name{bigquery}
\alias{bigquery}
\alias{dbi_driver}
\alias{dbConnect,BigQueryDriver-method}
\title{BigQuery DBI driver}
\usage{
\S4method{dbConnect}{BigQueryDriver}(
  drv,
  project,
  dataset = NULL,
  billing = project,
  page_size = 10000,
  quiet = NA,
  use_legacy_sql = FALSE,
  bigint = c("integer", "integer64", "numeric", "character"),
  labels = getOption("bigrquery.labels"),
  ...
)
}
\arguments{
\item{drv}{The result of \code{bigquery()}.}

\item{project, dataset}{Project and dataset identifiers}

\item{billing}{Identifier of project to bill.}

\item{page_size}{Number of items per page.}

\item{quiet}{If \code{FALSE}, displays progress bar; if \code{TRUE} is silent;
if \code{NA} picks based on whether or not you're in an interactive context.}

\item{use_legacy_sql}{If \code{TRUE} will use BigQuery's legacy SQL format.}

\item{bigint}{The R type that BigQuery's 64-bit integer types should be mapped to.
The default is \code{"integer"} which returns R's \code{integer} type but results in \code{NA} for
values above/below +/- 2147483647. \code{"integer64"} returns a \link[bit64:integer64]{bit64::integer64},
which allows the full range of 64 bit integers.}

\item{labels}{A named list of strings used to attach
\href{https://docs.cloud.google.com/bigquery/docs/labels-intro}{BigQuery labels}
to the resulting job, e.g. \code{list(env = "prod", team = "data")}. This
is most useful for cost allocation and other FinOps reporting.
Defaults to the value of \code{getOption("bigrquery.labels")}.}

\item{...}{Other arguments for compatibility with generic; currently ignored.}
}
\description{
Creates a BigQuery DBI driver for use in \code{\link[DBI:dbConnect]{DBI::dbConnect()}}.
}
\examples{
\dontshow{if (bq_testable()) withAutoprint(\{ # examplesIf}
con <- DBI::dbConnect(
  bigquery(),
  project = "publicdata",
  dataset = "samples",
  billing = bq_test_project()
)
con
DBI::dbListTables(con)
DBI::dbReadTable(con, "natality", n_max = 10)

# Create a temporary dataset to explore
ds <- bq_test_dataset()
con <- DBI::dbConnect(
  bigquery(),
  project = ds$project,
  dataset = ds$dataset
)
DBI::dbWriteTable(con, "mtcars", mtcars)
DBI::dbReadTable(con, "mtcars")[1:6, ]

DBI::dbGetQuery(con, "SELECT count(*) FROM mtcars")

res <- DBI::dbSendQuery(con, "SELECT cyl, mpg FROM mtcars")
dbColumnInfo(res)
dbFetch(res, 10)
dbFetch(res, -1)
DBI::dbHasCompleted(res)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-project.R
\name{api-project}
\alias{api-project}
\alias{bq_project_datasets}
\alias{bq_project_jobs}
\title{BigQuery project methods}
\usage{
bq_project_datasets(x, page_size = 100, max_pages = 1, warn = TRUE)

bq_project_jobs(x, page_size = 100, max_pages = 1, warn = TRUE)
}
\arguments{
\item{x}{A string giving a project name.}

\item{page_size}{Number of items per page.}

\item{max_pages}{Maximum number of pages to retrieve. Use \code{Inf} to retrieve
all pages (this may take a long time!)}

\item{warn}{If \code{TRUE}, warn when there are unretrieved pages.}
}
\value{
\itemize{
\item \code{bq_project_datasets()}: a list of \link{bq_dataset}s
\item \code{bq_project_jobs()}: a list of \link{bq_job}s.
}
}
\description{
Projects have two primary components: datasets and jobs. Unlike other
BigQuery objects, is no accompanying \code{bq_project} S3 class because a project
is a simple string.
}
\section{Google BigQuery API documentation}{

\itemize{
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/datasets/list}{datasets}
\item \href{https://docs.cloud.google.com/bigquery/docs/reference/rest/v2/jobs/list}{jobs}
}
}

\examples{
\dontshow{if (bq_testable()) withAutoprint(\{ # examplesIf}
bq_project_datasets("bigquery-public-data")
bq_project_datasets("githubarchive")

bq_project_jobs(bq_test_project(), page_size = 10)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr.R
\name{src_bigquery}
\alias{src_bigquery}
\title{A BigQuery data source for dplyr.}
\usage{
src_bigquery(project, dataset, billing = project, max_pages = 10)
}
\arguments{
\item{project}{project id or name}

\item{dataset}{dataset name}

\item{billing}{billing project, if different to \code{project}}

\item{max_pages}{(IGNORED) maximum pages returned by a query}
}
\description{
Create the connection to the database with \code{DBI::dbConnect()} then
use \code{\link[dplyr:tbl]{dplyr::tbl()}} to connect to tables within that database. Generally,
it's best to provide the fully qualified name of the table (i.e.
\code{project.dataset.table}) but if you supply a default \code{dataset} in the
connection, you can use just the table name. (This, however, will
prevent you from making joins across datasets.)
}
\examples{
\dontrun{
library(dplyr)

# To run this example, replace billing with the id of one of your projects
# set up for billing
con <- DBI::dbConnect(bigquery(), project = bq_test_project())

shakespeare <- con \%>\% tbl(I("publicdata.samples.shakespeare"))
shakespeare
shakespeare \%>\%
  group_by(word) \%>\%
  summarise(n = sum(word_count, na.rm = TRUE)) \%>\%
  arrange(desc(n))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-auth.R
\name{bq_token}
\alias{bq_token}
\title{Produce configured token}
\usage{
bq_token()
}
\value{
A \code{request} object (an S3 class provided by \link[httr:httr]{httr}).
}
\description{
For internal use or for those programming around the BigQuery API.
Returns a token pre-processed with \code{\link[httr:config]{httr::config()}}. Most users
do not need to handle tokens "by hand" or, even if they need some
control, \code{\link[=bq_auth]{bq_auth()}} is what they need. If there is no current
token, \code{\link[=bq_auth]{bq_auth()}} is called to either load from cache or
initiate OAuth2.0 flow.
If auth has been deactivated via \code{\link[=bq_deauth]{bq_deauth()}}, \code{bq_token()}
returns \code{NULL}.
}
\examples{
\dontrun{
bq_token()
}
}
\seealso{
Other low-level API functions:
\code{\link[=bq_has_token]{bq_has_token()}}
}
\concept{low-level API functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-test.R
\name{bq_test_project}
\alias{bq_test_project}
\alias{bq_test_init}
\alias{bq_test_dataset}
\alias{bq_testable}
\alias{bq_authable}
\alias{gs_test_bucket}
\alias{gs_test_object}
\title{Project to use for testing bigrquery}
\usage{
bq_test_project()

bq_test_init(name = "basedata")

bq_test_dataset(name = random_name(), location = "US")

bq_testable()

bq_authable()

gs_test_bucket()

gs_test_object(name = random_name())
}
\arguments{
\item{name}{Dataset name - used only for testing.}
}
\value{
\code{bq_test_project()} returns the name of a project suitable for use in
testing. \code{bq_test_dataset()} creates a temporary dataset whose lifetime is
tied to the lifetime of the object that it returns.
}
\description{
You'll need to set the \code{BIGQUERY_TEST_PROJECT} (name of a project) and
\code{BIGQUERY_TEST_BUCKET} (name of bucket) env vars in order to run bigrquery
tests locally. I recommend creating a new project because the tests involve
both reading and writing in BigQuery and Cloud Storage.

The \code{BIGQUERY_TEST_PROJECT} must have billing enabled for the project. While
logged in, via \code{bq_auth()}, as a user with permission to work in
\code{BIGQUERY_TEST_PROJECT}, run \code{bq_test_init()} once to perform some setup.
}
\section{Testing}{
 In tests, \code{bq_test_project()} (and hence \code{bq_test_dataset()}) will
automatically skip if auth and a test project are not available.
}

\examples{
\dontshow{if (bq_testable()) withAutoprint(\{ # examplesIf}
ds <- bq_test_dataset()
bq_mtcars <- bq_table_upload(bq_table(ds, "mtcars"), mtcars)

# dataset and table will be automatically deleted when ds is GC'd
\dontshow{\}) # examplesIf}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bq-auth.R
\name{bq_user}
\alias{bq_user}
\title{Get info on current user}
\usage{
bq_user()
}
\value{
An email address or, if no token has been loaded, \code{NULL}.
}
\description{
Reveals the email address of the user associated with the current token.
If no token has been loaded yet, this function does not initiate auth.
}
\examples{
\dontrun{
bq_user()
}
}
\seealso{
\code{\link[gargle:token_userinfo]{gargle::token_userinfo()}}, \code{\link[gargle:token_email]{gargle::token_email()}},
\code{\link[gargle:token_tokeninfo]{gargle::token_tokeninfo()}}
}
